from optparse import OptionParser
import sys, os, re, glob
from utilities.align_parsers import psl, sam
from utilities import intspan, tools
from utilities.alignment import Alignment
from utilities.assembly import Assembly
from utilities.bam import BAM
from feature import FeatureFinder
from transcript import Transcript
from cnv import CNV
from copy import deepcopy
from annotations import dbsnp, dgv
from pep_change import translate
import operator, sys

VERSION = '1.4.4'

import utilities.cfg
PACKAGE_DIR = "/".join(os.path.abspath(__file__).split("/")[:-2])
PYTHON = 'python'

class Fusion:

    def __init__(self, a1, a2):
        self.align1 = a1
        self.align2 = a2

        self.contig = a1.query
        self.contig_size = a1.query_len

        self.gene1 = self.gene2 = 'NA'
	self.sense1 = self.sense2 = 'NA'
	self.exon1 = self.exon2 = 'NA'
	self.txt1 = self.txt2 = None
	self.txt1_name = self.txt2_name = 'NA'
	self.exon_bound1 = self.exon_bound2 = 'NA'

        self.overlap_query = None
        self.overlap_query_fraction = None
        self.overlap_target = None
        self.overlap_target_fraction = None
        self.contig_cover = None
        self.contig_cover_fraction = None
	
	#read evidence
        self.num_read_pairs = 0
	self.num_flanking_pairs = 0
	self.num_breakpoint_pairs = [0,0]
        self.num_spanning_reads = 0
	
        self.mutation = 'NA'
	self.breakpoint = 'NA'
	self.lsr_region = None
	self.cnv = None
	
	self.fusion_type = 'LSR'
	
        #filtering
        self.exclude = False
	self.good_read_pair = False
	self.good_span_read = False
        self.read_support = False
        
        self.contig_local = None
        self.local = None
        self.ff = None
	self.gene_fusion = False
        self.lsr = False
	self.orients = []
	self.reciprocal = None
	self.event_id = None
	self.descriptor = 'NA'
	self.breaks = None
	self.chroms = None
	
	self.dbsnp = 'NA'
	self.dgv = 'NA'
	
	self.gene5prime = 'NA'
	self.exon5prime = 'NA'
	self.gene3prime = 'NA'
	self.exon3prime = 'NA'
	self.align5prime = self.align3prime = None
	
	self.txt5prime = self.txt3prime = None
	self.frame = 'NA'
			
    def details(self, tsv=True, id=None):
        ctg_fields = self.contig.split(" ")
        ctg = "%s" % (ctg_fields[0])
	
	align1, align2 = self.align1, self.align2
	    
        target = "%s:%s-%s,%s:%s-%s" % (align1.target, align1.tstart, align1.tend, align2.target, align2.tstart, align2.tend)
        query = "%s-%s,%s-%s" % (align1.qstart, align1.qend, align2.qstart, align2.qend)
	
	#alignment strands
        align_strands = "NA,NA"
        if align1.strand:
            align_strands = "%s,%s" % (align1.strand, align2.strand)
	    
	#defined alignment params
	align_len1 = float(align1.qend) - float(align1.qstart) + 1
        align_len2 = float(align2.qend) - float(align2.qstart) + 1
        align_frac1 = align_len1/float(align1.query_len)
        align_frac2 = align_len2/float(align2.query_len)
        params = "TO:%.2f,CO:%.2f,CC:%.2f,I1:%.1f,I2:%.1f,AF1:%.2f,AF2:%.2f" % (self.overlap_target_fraction, self.overlap_query_fraction, self.contig_cover_fraction, float(align1.identity), float(align2.identity), align_frac1, align_frac2)
	    
	#defines fusion type
	fusion_type = 'NA'
	if self.gene_fusion:
	    fusion_type = 'gene_fusion'
	elif self.lsr:
	    fusion_type = 'lsr'
	elif self.local:
	    fusion_type = 'local'
	    
	reciprocal = 'NA'
	if self.reciprocal:
	    reciprocal = self.reciprocal
	    
	if self.orients:
	    if self.orients[0] and self.orients[1]:
		orient1, orient2 = 'L', 'L'
	    elif self.orients[0] and not self.orients[1]:
		orient1, orient2 = 'L', 'R'
	    elif not self.orients[0] and self.orients[1]:
		orient1, orient2 = 'R', 'L'
	    elif not self.orients[0] and not self.orients[1]:
		orient1, orient2 = 'R', 'R'
	else:
	    orient1, orient2 = 'NA', 'NA'
	orients = '%s,%s' % (orient1, orient2)
	
	genes = '%s,%s' % (self.gene1, self.gene2)
	txt1, txt2 = self.txt1_name, self.txt2_name
	if self.txt1:
	    txt1 = self.txt1.name
	if self.txt2:
	    txt2 = self.txt2.name
	transcripts = '%s,%s' % (txt1, txt2)
	senses = '%s,%s' % (self.sense1, self.sense2)
	exons = '%s,%s' % (self.exon1, self.exon2)
	exon_bounds = '%s,%s' % (self.exon_bound1, self.exon_bound2)
	    
        data = [str(self.event_id), str(ctg), str(self.contig_size), target, query, align_strands, str(self.num_flanking_pairs), ','.join([str(self.num_breakpoint_pairs[0]), str(self.num_breakpoint_pairs[1])]), str(self.num_spanning_reads), self.mutation, self.breakpoint, str(self.size()), genes, transcripts, senses, exons, exon_bounds, reciprocal, self.descriptor, orients, self.gene5prime, self.gene3prime, self.exon5prime, self.exon3prime, self.frame, params, self.fusion_type]  
	            		
	if self.cnv:
	    data.append(self.cnv)
	    
	data.append(self.dbsnp)
	data.append(self.dgv)
            
        if tsv:
            return "\t".join(data)
		     
	
    def report_align_params(self):
	align_len1 = float(self.align1.qend) - float(self.align1.qstart) + 1
        align_len2 = float(self.align2.qend) - float(self.align2.qstart) + 1
        align_frac1 = align_len1/float(self.align1.query_len)
        align_frac2 = align_len2/float(self.align2.query_len)

        params = "TO:%.2f,CO:%.2f,CC:%.2f,I1:%.1f,I2:%.1f,AF1:%.2f,AF2:%.2f" % (self.overlap_target_fraction, self.overlap_query_fraction, self.contig_cover_fraction, float(self.align1.identity), float(self.align2.identity), align_frac1, align_frac2)
	
	return params
    
    def find_read_pairs(self, bam, genome_coord_buffer=0, breakpoint_buffer=0, keep_mito=False, reads=None):
        if not keep_mito and (self.align1.target == "chrM" or self.align2.target == "chrM" or self.align1.target == "chrMT" or self.align2.target == "chrMT"):
            return

	#determining breakpoint and passing it to determine valid read pairs
	breakpoint = None
	if self.breakpoint.upper() != 'NA':
	    chr1, coord1, chr2, coord2 = re.split('[:\|]', self.breakpoint)
	    	    
	    #match up breakpoint with alignment
	    if chr1 != chr2:
		if chr1 == self.align1.target:
		    breakpoint = [chr1, int(coord1)], [chr2, int(coord2)]
		else:
		    breakpoint = [chr2, int(coord2)], [chr1, int(coord1)]
	    else:
		if (int(coord1) == int(self.align1.tstart) or int(coord1) == int(self.align1.tend)) and int(coord1) != int(self.align2.tstart) and int(coord1) != int(self.align2.tend):
		    breakpoint = [chr1, int(coord1)], [chr2, int(coord2)]
		elif (int(coord2) == int(self.align1.tstart) or int(coord2) == int(self.align1.tend)) and int(coord2) != int(self.align2.tstart) and int(coord2) != int(self.align2.tend):
		    breakpoint = [chr2, int(coord2)], [chr1, int(coord1)]
		    
	
	    #add buffer to breakpoint
	    if breakpoint:
		if int(self.align1.tstart) == int(breakpoint[0][1]):
		    region1 = [self.align1.target, int(self.align1.tstart), max(int(self.align1.tend), breakpoint[0][1] + genome_coord_buffer)]
		elif int(self.align1.tend) == breakpoint[0][1]:
		    region1 = [self.align1.target, min(int(self.align1.tstart), breakpoint[0][1] - genome_coord_buffer), int(self.align1.tend)]
	
		if int(self.align2.tstart) == int(breakpoint[1][1]):
		    region2 = [self.align2.target, int(self.align2.tstart), max(int(self.align2.tend), breakpoint[1][1] + genome_coord_buffer)]
		elif int(self.align2.tend) == breakpoint[1][1]:
		    region2 = [self.align2.target, min(int(self.align2.tstart), breakpoint[1][1] - genome_coord_buffer), int(self.align2.tend)]
		
	    else:
		region1 = [self.align1.target, int(self.align1.tstart), int(self.align1.tend)]
		region2 = [self.align2.target, int(self.align2.tstart), int(self.align2.tend)]
	    
	else:
	    region1 = [self.align1.target, int(self.align1.tstart), int(self.align1.tend)]
	    region2 = [self.align2.target, int(self.align2.tstart), int(self.align2.tend)]
	    
	if not breakpoint:
	    return False
	    
	#make sure after buffer subtraction it's not negative
	if region1[1] < 0:
	    region1[1] = 0
	if region2[1] < 0:
	    region2[1] = 0
	
	#find full flanking pairs
	self.num_flanking_pairs = bam.confirm(None, None, feature="breakpoint", regions=[region1, region2], breakpoint=breakpoint, breakpoint_buffer=breakpoint_buffer)	
	
	#find breakpoint pairs - pairs with one mate overlapping junctions
	if reads:
	    reads1 = bam.find_mates_in_genome(region1, breakpoint[0], missing_mates=reads, breakpoint_buffer=breakpoint_buffer)
	    reads2 = bam.find_mates_in_genome(region2, breakpoint[1], missing_mates=reads, breakpoint_buffer=breakpoint_buffer)
		
	    if len(reads1) > 0:
		if not bam.is_single_stack(reads1):
		    self.num_breakpoint_pairs[0] = len(reads1)
		else:
		    self.num_breakpoint_pairs[0] = -1 * len(reads1)

	    if len(reads2) > 0:
		if not bam.is_single_stack(reads2):
		    self.num_breakpoint_pairs[1] = len(reads2)
		else:
		    self.num_breakpoint_pairs[1] = -1 * len(reads2)
		    
	self.num_read_pairs = self.num_flanking_pairs + min(self.num_breakpoint_pairs)
		
	    
    def find_spanning_reads(self, bam, lib, breakpoint_buffer=0):
        if int(self.align1.qstart) > int(self.align2.qstart):
            start = int(self.align2.qend)
            end = int(self.align1.qstart)
        else:
            start = int(self.align1.qend)
            end = int(self.align2.qstart)

        if start > end:
            tmp = start
            start = end
            end = tmp

        start -= breakpoint_buffer
        end += breakpoint_buffer

        if start <= 0:
            start = 1
        if end > int(self.align1.query_len):
            end = int(self.align1.query_len)

	contig = tools.bamfile_contig_name(self.align1.query, lib)
		    
	spanning_reads = bam.confirm(target=contig, pos=[start,end], feature="breakpoint", seq=True)
	self.num_spanning_reads = len(spanning_reads)
	
	return spanning_reads
	
	
    def is_local(self, model=None, annot_file=None, local_gap=5000):
        if self.align1.target != self.align2.target:
            self.local = False
            return self.local
        
        if self.align1.target == self.align2.target:
	    spans = [[int(self.align1.tstart), int(self.align1.tend)], [int(self.align2.tstart), int(self.align2.tend)]]
	    if int(self.align1.tstart) > int(self.align2.tstart):
		spans.reverse()
	    spans[0][1] += local_gap

	    if intspan.overlap(spans[0], spans[1]):
		self.local = True
		self.exclude = True

        if self.gene1 and self.gene2 and self.gene1.upper() != 'NA' and self.gene2.upper() != 'NA':
            if not self.local:
                if self.gene1 == self.gene2:
                    self.local = True
                    self.exclude = True

            if not self.local and model and annot_file:
                if not self.txt1:
                    txts = Transcript.find_by_name(model, annot_file, self.gene1)
                    if txts:
                        txts.sort(lambda x,y:y.length - x.length)
                        self.txt1 = txts[0]

                if not self.txt2:
                    txts = Transcript.find_by_name(model, annot_file, self.gene2)
                    if txts:
                        txts.sort(lambda x,y:y.length - x.length)
                        self.txt2 = txts[0]

                if self.txt1 and self.txt2 and self.txt1.chrom == self.txt2.chrom and intspan.overlap((self.txt1.txStart, self.txt1.txEnd), (self.txt2.txStart, self.txt2.txEnd)):
                    self.local = True
                    self.exclude = True

        return self.local
    
    def define_mutation(self, cytobands=None, genome=None):
	"""Define event type
	
	Keyword arguments:
	cytobands (optional) -- for outputing proper genetic name
	
	Algorithm:
	1. sort the target coordinates based on query strands
	2. define mutation based on whether target coordinates ascend or descend in each region
	3. define and keep breakpoints in each target region: self.breaks
	4. formulate breakpoint
	5. defines 'orientation' in each segment for finding reciprocal events
	6. gives proper genetic name of event if cytobands provided
	
	"""
		
	if self.align1.target != self.align2.target:
	    self.mutation = 'translocation'
            
            span1 = [int(self.align1.tstart), int(self.align1.tend)]
            span2 = [int(self.align2.tstart), int(self.align2.tend)]
            if self.align1.strand == '-':
                span1.reverse()
            if self.align2.strand == '-':
                span2.reverse()

            if int(self.align1.qstart) < int(self.align2.qstart):
                self.breakpoint = "%s:%s|%s:%s" % (self.align1.target, span1[1], self.align2.target, span2[0])
                self.breaks = (span1[1], span2[0])
                    
                if span1[1] == min(span1):
                    self.orients.append(False)
                else:
                    self.orients.append(True)
                if span2[0] == min(span2):
                    self.orients.append(False)
                else:
                    self.orients.append(True)
                    
            else:
                self.breakpoint = "%s:%s|%s:%s" % (self.align1.target, span1[0], self.align2.target, span2[1])
                self.breaks = (span1[0], span2[1])
                    
                if span1[0] == min(span1):
                    self.orients.append(False)
                else:
                    self.orients.append(True)
                if span2[1] == min(span2):
                    self.orients.append(False)
                else:
                    self.orients.append(True)
                    
            span1.sort(lambda x,y: x-y)
            span1.sort(lambda x,y: x-y)
            self.lsr_region = [[self.align1.target, span1[0], span1[1]], [self.align2.target, span2[0], span2[1]]]
	    	    
	else:
	    if int(self.align1.qstart) < int(self.align2.qstart):
		span1 = [int(self.align1.tstart), int(self.align1.tend)]
		span2 = [int(self.align2.tstart), int(self.align2.tend)]
		if self.align1.strand == '-':
		    span1.reverse()
		if self.align2.strand == '-':
		    span2.reverse()
	    else:
		span1 = [int(self.align2.tstart), int(self.align2.tend)]
		span2 = [int(self.align1.tstart), int(self.align1.tend)]
		if self.align2.strand == '-':
		    span1.reverse()
		if self.align1.strand == '-':
		    span2.reverse()
	    
	    ascend1 = None
	    if span1[0] < span1[1]:
		ascend1 = True
	    else:
		ascend1 = False
		    
	    ascend2 = None
	    if span2[0] < span2[1]:
		ascend2 = True
	    else:
		ascend2 = False
			
	    if span1[0] < span2[0]:
		if ascend1 and ascend2:
		    self.mutation = 'deletion'
		    self.orients = [True, False]
		elif not ascend1 and not ascend2:
		    self.mutation = 'duplication'
		    self.orients = [False, True]
		elif ascend1 and not ascend2:
		    self.mutation = 'inversion'
		    self.orients = [True, True]
		elif not ascend1 and ascend2:
		    self.mutation = 'inversion'
		    self.orients = [False, False]
		    
		self.breakpoint = "%s:%s|%s:%s" % (self.align1.target, span1[1], self.align2.target, span2[0])
		self.lsr_region = [self.align1.target, span1[1], span2[0]]
		self.breaks = (span1[1], span2[0])
		
	    elif span1[0] > span2[0]:
		if ascend1 and ascend2:
		    self.mutation = 'duplication'
		    self.orients = [False, True]
		elif not ascend1 and not ascend2:
		    self.mutation = 'deletion'
		    self.orients = [True, False]
		elif ascend2 and not ascend1:
		    self.mutation = 'inversion'
		    self.orients = [False, False]
		elif not ascend2 and ascend1:
		    self.mutation = 'inversion'
		    self.orients = [True, True]
		    		
		self.breakpoint = "%s:%s|%s:%s" % (self.align2.target, span2[0], self.align1.target, span1[1])
		self.lsr_region = [self.align1.target, span2[0], span1[1]]
		self.breaks = (span2[0], span1[1])
				
	    #probably not real, mis-assembly
	    elif span1[0] == span2[0]:
		self.mutation = 'NA'
		self.breakpoint = 'NA'
		self.lsr_region = []
	
	if cytobands:
	    self.descriptor = self.proper_name(cytobands, genome)
    

    def proper_name(self, cytobands, genome=None):
	"""Create proper genetic name for event.
	
	Keyword arguments:
	cytobands (required) -- a dictionary where key is chromosome name and value is a list of (start, end, band)
	self.mutation -- mutation type
	self.breaks -- list of break coordinates
	
	Return values:
	string containg name or 'NA'
		
	"""
	
	if self.mutation and self.breaks:
	    chrom1, chrom2 = self.align1.target, self.align2.target
	    if genome is not None:
		chrom1 = tools.proper_chrom(self.align1.target, genome)
		chrom2 = tools.proper_chrom(self.align2.target, genome)
	    
	    band1, band2 = tools.overlap_cytobands(cytobands, chrom1, self.breaks[0], chrom2, self.breaks[1])
	    
	    #inversion across centromere becomes translocation
	    if self.mutation == 'inversion' and band1 is not None and band2 is not None and ((band1[0] == 'p' and band2[0] == 'q') or (band1[0] == 'q' and band2[0] == 'p')):
		self.mutation = 'translocation'
	    
	    if band1 is not None and band2 is not None:
		return tools.formulate_event(self.mutation, self.align1.target, self.align2.target, band1, band2)
	    else:
		return 'NA'
	else:
	    return 'NA'
	
    def size(self):
	if self.align1.target == self.align2.target and self.breaks:
	    return abs(self.breaks[1] - self.breaks[0]) - 1
	    
	else:
	    return '-'
			    		    
    def order_gene(self):
	result = None
	if self.gene1.upper() != 'NA' and self.gene2.upper() != 'NA' and self.txt1 and self.txt2:
	    if self.sense1 == self.sense2:
		self.fusion_type = 'sense_fusion'
	    else:
		self.fusion_type = 'antisense_fusion'
	    
	    self.gene5prime = '-'
	    self.exon5prime = '-'
	    self.gene3prime = '-'
	    self.exon3prime = '-'
	    
	    prime1 = self.prime_gene(self.orients[0], self.txt1)
	    prime2 = self.prime_gene(self.orients[1], self.txt2)

	    if prime1 != prime2 and prime1 in (5,3) and prime2 in (5,3):
		if prime1 == 5:
		    self.gene5prime = self.gene1
		    self.gene3prime = self.gene2
		    self.txt5prime = self.txt1
		    self.txt3prime = self.txt2
		    self.align5prime = self.align1
		    self.align3prime = self.align2
		    self.sense5prime = self.sense1
		    self.sense3prime = self.sense2
		    
		    if 'exon' in self.exon1:
			self.exon5prime = self.exon1.lstrip('exon')
		    elif 'intron' in self.exon1:
			self.exon5prime = self.exon1.lstrip('intron')
		    elif 'utr' in self.exon1:
			self.exon5prime = self.exon1
			
		    if 'exon' in self.exon2:
			self.exon3prime = self.exon2.lstrip('exon')
		    elif 'intron' in self.exon2:
			self.exon3prime = str(int(self.exon2.lstrip('intron')) + 1)
		    elif 'utr' in self.exon2:
			self.exon3prime = self.exon2
						
		else:
		    self.gene5prime = self.gene2
		    self.gene3prime = self.gene1
		    self.txt5prime = self.txt2
		    self.txt3prime = self.txt1
		    self.align5prime = self.align2
		    self.align3prime = self.align1
		    self.sense5prime = self.sense2
		    self.sense3prime = self.sense1
		    
		    if 'exon' in self.exon2:
			self.exon5prime = self.exon2.lstrip('exon')
		    elif 'intron' in self.exon2:
			self.exon5prime = self.exon2.lstrip('intron')
		    elif 'utr' in self.exon2:
			self.exon5prime = self.exon2
			
		    if 'exon' in self.exon1:
			self.exon3prime = self.exon1.lstrip('exon')
		    elif 'intron' in self.exon1:
			self.exon3prime = str(int(self.exon1.lstrip('intron')) + 1)
		    elif 'utr' in self.exon1:
			self.exon3prime = self.exon1
		    
    def prime_gene(self, orient, txt):
	'''
	orient=L, txt_strand='+' => 5'
	orient=L, txt_strand='-' => 3'
	orient=R, txt_strand='+' => 3'
	orient=R, txt_strand='-' => 5'
	'''
	prime = None
	if orient:
	    if txt.strand == '+':
		prime = 5
	    elif txt.strand == '-':
		prime = 3
	else:
	    if txt.strand == '+':
		prime = 3
	    elif txt.strand == '-':
		prime = 5
		
	return prime
    
    def at_exon_bound(self):
	self.exon_bound1, self.exon_bound2 = 'NA', 'NA'
	if 'exon' in self.exon1:
	    self.exon_bound1 = 'no'

	    if self.txt1.strand == '+':
		exon = self.txt1.exons[int(self.exon1.lstrip('exon')) - 1]
	    elif self.txt1.strand == '-':
		exon = self.txt1.exons[::-1][int(self.exon1.lstrip('exon')) - 1]
		
	    if int(self.breaks[0]) == int(exon[0]) or int(self.breaks[0]) == int(exon[1]):
		self.exon_bound1 = 'yes'
    	    
	if 'exon' in self.exon2:
	    self.exon_bound2 = 'no'
	    
	    if self.txt2.strand == '+':
		exon = self.txt2.exons[int(self.exon2.lstrip('exon')) - 1]
	    elif self.txt2.strand == '-':
		exon = self.txt2.exons[::-1][int(self.exon2.lstrip('exon')) - 1]
		
	    if int(self.breaks[1]) == int(exon[0]) or int(self.breaks[1]) == int(exon[1]):
		self.exon_bound2 = 'yes'
    	    
	    
    def check_frame(self, ff, refseq, contig_seq):
	if self.sense1 == self.sense2 and self.txt5prime is not None and self.txt3prime is not None:
	    #get chromosome names for constructing cDNAs
	    if self.txt5prime.name == self.txt1.name:
		target5prime = self.align1.target
		target3prime = self.align2.target
	    else:
		target5prime = self.align2.target
		target3prime = self.align1.target
		
	    #not entire contig may be aligned, just extract aligned contig sequence
	    qcoords = [int(self.align1.qstart), int(self.align1.qend), int(self.align2.qstart), int(self.align2.qend)]
	    qcoords.sort(key=int)
	    contig_seq = contig_seq[qcoords[0]-1:qcoords[-1]]
	        
	    #extract 5' and 3' cDNAa sequence
	    cdna5prime = ff.construct_cdna([target5prime], self.txt5prime, refseq)	    
	    pep5prime = translate(cdna5prime, orient=self.txt5prime.strand, frame=0)
	    cdna3prime = ff.construct_cdna([target3prime], self.txt3prime, refseq)	    
	    pep3prime = translate(cdna3prime, orient=self.txt3prime.strand, frame=0)
	    
	    #whether contig covers transcription start determines how to compare amino acids
	    #if int(self.align5prime.tstart) < int(self.txt5prime.cdsStart):
	    if (self.txt5prime.strand == '+' and int(self.align5prime.tstart) <= int(self.txt5prime.cdsStart)) or\
	       (self.txt5prime.strand == '-' and int(self.align5prime.tstart) >= int(self.txt5prime.cdsStart)):
		contig_has_start = True
	    else:
		contig_has_start = False
			    
	    #construct all the orfs using the contig sequence
	    orfs = translate(contig_seq, all=True, full=True)	    
	    
	    #key = index of orf, value = length of match
	    orf_checks = {}
	    if pep5prime:
		for o in range(len(orfs)):
		    orf = orfs[o]
		    if (self.sense5prime == '-' and orf[2] == 1) or\
		       (self.sense5prime == '+' and orf[2] == '-'):
			continue				
		
		    last_match = None
		    if contig_has_start:
			#at least 2 matches
			for i in range(2, len(pep5prime)):
			    match = re.search(pep5prime[:i], orf[-1])
			    if not match:
				if last_match:
				    orf_checks[o] = (i - 1, last_match.end())
				break
			    else:
				last_match = match
		    else:	
			for i in range(2, len(orf[-1])):
			    match = re.search(orf[-1][:i], pep5prime)
			    if not match:
				if last_match:
				    orf_checks[o] = (i - 1, i)
				break
			    else:
				last_match = match
	    else:
		sys.stdout.write("can't determine peptide sequence of 5' gene %s %s %s %s\n" % (self.align1.query, self.txt5prime.alias, self.txt5prime.name, self.txt5prime.coding_type()))
			
	    if orfs and orf_checks:
		#sort orfs by length of match
		sorted_orfs = sorted(orf_checks.iteritems(), key=operator.itemgetter(1), reverse=True)

		#best orf = orfs[sorted_orfs[0][0]][-1]
		#get 3' remaing of best orf, start at second amino acid to allow 1 mismatch
		orf3prime = orfs[sorted_orfs[0][0]][-1][sorted_orfs[0][1][1]:][1:]
				    
		#if remaining of contig orf is part of 3' gene orf, then in-frame
		if orf3prime and pep3prime and pep3prime != '' and re.search(orf3prime, pep3prime):
		    self.frame = 'in'
		#fused to 3'UTR
		elif orfs[sorted_orfs[0][0]] and len(orfs[sorted_orfs[0][0]][-1]) > 1 and self.exon3prime == '3utr':
		    self.frame = 'in'
		else:
		    self.frame = 'out'
	    
	    
class FusionFinder:

    def __init__(self, params, debug=False, unique=False, is_multi_mapped=True):
        self.params = params
        self.fusions = []
        self.debug = debug
        self.unique = unique

	self.filtering = {'min_id':98, 'overlap':True, 'chrM':True, 'haplotype':True, 'readpairs':[0,0], 'min_span_reads':0, 'gene_fusions':False}	
	if params and params.has_key('min_identity'):
	    self.filtering['min_id'] = params['min_identity']
	
        #debug option
        if debug:
            print self.params
	    	    
	self.is_multi_mapped = is_multi_mapped
	
	self.fusion_groups = {}
	
	#FeatureFinder
	self.ff = None
	self.cytobands = None
	self.contigs_file = None	
	self.splice_motif_file = None
	self.refseq = None
	self.genome = None
	
    def prepare_annotation(self):
	package_dir = "/".join(os.path.abspath(__file__).split("/")[:-2])
	self.genome_dir = None
	if self.genome:
	    self.genome_dir = package_dir + '/annotations/' + self.genome
	    if os.path.isdir(self.genome_dir):
		self.splice_motif_file = self.genome_dir + '/splice_motifs.txt'
		
		cytoband_file = self.genome_dir + '/cytoBand.txt'
		self.cytobands = tools.extract_cytobands(cytoband_file)
		
	    self.refseq = tools.get_refseq_from_2bit(self.genome)
	    
	    if self.model:
		self.ff = FeatureFinder(self.genome, self.model)
		
    def get_split_aligns(self, aligns):
	fusions = []
	i = 0
	while i <= len(aligns) - 1:
	    start = i
	    end = i
	    j = i + 1
	    while j < len(aligns) and aligns[j].query == aligns[i].query:
		end = j 
		j += 1
		    
	    fusions.extend(self.extract_fusions(aligns[start:end + 1]))
	    
	    i = end + 1
		
	return fusions
    
    def extract_fusions(self, aligns):
	fusions = []
	
	min_extend = 30
	
	#check identity
	exclude = {}
	for i in range(len(aligns)):
	    if aligns[i].identity < self.params['min_identity']:
		exclude[i] = True
		if self.debug:
		    print 'contig:%s screened out because identity %s less than %s' % (aligns[0].query, aligns[i].identity, self.params['min_identity'])
		    
		
	#identify start
	start = None
	for i in range(len(aligns)):	    
	    if exclude.has_key(i):
		continue
	    
	    if start is None:
		start = i
		
	    elif int(aligns[i].qstart)  < int(aligns[start].qstart):
		start = i
	    	    
	    #tie in qstart
	    elif aligns[i].qstart == aligns[start].qstart and aligns[i].score > aligns[start].score:
		start = i
		    
	#identify end
	end = None
	for i in range(len(aligns)):
	    if exclude.has_key(i):
		continue
	    
	    if end is None:
		end = i
		
	    elif int(aligns[i].qend) > int(aligns[end].qend):
		end = i
	    	    
	    #tie in qend
	    elif aligns[i].qend == aligns[end].qend and aligns[i].score > aligns[end].score:
		end = i
		
	if self.debug:
	    print 'contig:%s start:%s-end:%s' % (aligns[0].query, start, end)
	    
	failed = False
	
	if start != end and start is not None and end is not None:
	    #bridge
	    current = start
	    used = [start]
	    params = {}
	    
	    count = 0
	    while used[-1] != end:
		next, params[next] = self.plank(current, aligns, used)
		if next is not None:
		    used.append(next)
		    current = next	    
		
		count += 1
		if count > 10 or (next is None and used[-1] != end):
		    failed = True
		    break
		
	    if self.debug:
		for i in used:
		    print '%d %s:%s-%s %s:%s-%s' % (i, aligns[i].query, aligns[i].qstart, aligns[i].qend, aligns[i].target, aligns[i].tstart, aligns[i].tend)
		
	    if not failed:
		passed, params2 = self.valid_path(aligns, used)
		
		if not passed:
		    failed = True
		else:
		    if params2 is not None:
			for i in used[1:]:
			    for key, value in params2.iteritems():
				params[i][key] = value
	    
	    if not failed:
		for i in range(len(used)-1):
		    chr_order = tools.compare_chr(aligns[used[i]].target, aligns[used[i+1]].target)
		    
		    align1, align2 = aligns[used[i]], aligns[used[i+1]]
		    reverse = False
		    if chr_order > 0:
			reverse = True
		    elif chr_order == 0:
			if int(aligns[used[i]].tstart) > int(aligns[used[i+1]].tstart):
			    reverse = True
			elif int(aligns[used[i]].tend) > int(aligns[used[i+1]].tend):
			    reverse = True
		    if reverse:
			align1, align2 = aligns[used[i+1]], aligns[used[i]]
			
		    fusion = Fusion(align1, align2)
		    tools.set_attrs(fusion, params[used[i+1]])
		    fusions.append(fusion)
		    		
	return fusions
	
    def plank(self, current, aligns, used):
	next = None
	min_distance = None
	max_score = None
	params_all = {}
	
	for i in range(len(aligns)):		
	    if i in used:
		continue
	    
	    #not extending
	    is_valid_pair, params = self.valid_pair(aligns[current], aligns[i])
	    if not is_valid_pair:
		continue
	    else:
		params_all[i] = params
	    
	    distance = abs(int(aligns[current].qend) - int(aligns[i].qstart))
	    	    
	    if next is None:
		next = i
		min_distance = distance
		
	    elif self.similar_extension(aligns[next], aligns[i]):
		if int(aligns[i].score) > int(aligns[next].score):
		    next = i
		    min_distance = distance
		continue
						
	    elif distance < min_distance:
		if abs(distance - min_distance) > 5:
		    next = i
		    min_distance = distance
		elif int(aligns[i].score) > int(aligns[next].score):
		    next = i
		    min_distance = distance
		
	
	if next is None:
	    return None, None
	else:
	    return next, params_all[next]
    
    def similar_extension(self, align1, align2):
	close = 10
	if abs(int(align1.qstart) - int(align2.qstart)) <= close or abs(int(align1.qend) - int(align2.qend)) <= close:
	    return True
	
	
	
    def valid_pair(self, align1, align2):
	passed = True
	params = {}
	
	#check if any member of the alignment set has a single alignment over 95% of the total length; if so, skip
	for align in (align1, align2):
	    match_len = int(align.qend) - int(align.qstart)
	    if (float(match_len) / float(align.query_len) >= self.params['max_match_fraction']):
		passed = False
		if self.debug:
		    self.output_message('Failed', align1, align2, 'max_match_fraction', float(match_len) / float(align.query_len), self.params['max_match_fraction'])
		return passed, params
	
         
	#if one of the alignments is a subset of the other
	if intspan.subsume([align1.qstart, align1.qend], [align2.qstart, align2.qend]) or intspan.subsume([align2.qstart, align2.qend], [align1.qstart, align1.qend]):
	    passed = False
	    if self.debug:
		self.output_message('Failed', align1, align2, 'query_subset')
	    return passed, params

	
	#Check how much the query alignments between the contigs overlap
	(overlap_query, overlap_query_fraction) = self.overlap([align1.qstart, align1.qend], [align2.qstart, align2.qend], align1.query_len)
	(overlap_q1, overlap_q1_fraction) = self.overlap([align1.qstart, align1.qend], [align2.qstart, align2.qend], int(align1.qend) - int(align1.qstart) + 1)
	(overlap_q2, overlap_q2_fraction) = self.overlap([align1.qstart, align1.qend], [align2.qstart, align2.qend], int(align2.qend) - int(align2.qstart) + 1)
	params['overlap_query_fraction'] = overlap_query_fraction

	#Check if the overlap query between the contigs is too high
	if not overlap_query or overlap_query_fraction <= self.params['max_query_olap'] and\
	   overlap_q1_fraction <= self.params['max_query_olap_single'] and\
	   overlap_q2_fraction <= self.params['max_query_olap_single']:

	    #Get the target query overlap
	    (overlap_target, overlap_target_fraction) = self.overlap([align1.tstart, align1.tend], [align2.tstart, align2.tend], align1.query_len)
	    params['overlap_target_fraction'] = overlap_target_fraction
                    
	    #if targets are not the same, means no overlap
	    if align1.target == align2.target:
		(overlap_target, overlap_target_fraction) = self.overlap([align1.tstart, align1.tend], [align2.tstart, align2.tend], align1.query_len)
	    else:
		(overlap_target, overlap_target_fraction) = (0.0, 0.0)

	    #Check if the target overlap between the contigs is too high
	    if overlap_target_fraction > self.params['max_target_olap']:
		#if alignments are on different strand, don't count overlap
		if not align1.strand or not align2.strand:
		    passed = False
		elif align1.strand == align2.strand:
		    passed = False
		if not passed:
		    if self.debug:
			self.output_message('Failed', align1, align2, 'max_target_overlap', overlap_target_fraction, self.params['max_target_olap'])
		    return passed, params

								
	else:
	    passed = False
	    if self.debug:
		if overlap_query_fraction > self.params['max_query_olap']:
		    self.output_message('Failed', align1, align2, 'max_query_olap', overlap_query_fraction, self.params['max_query_olap'])
		if overlap_q1_fraction > self.params['max_query_olap_single']:
		    self.output_message('Failed', align1, align2, 'max_query_olap_single', overlap_q1_fraction, self.params['max_query_olap_single'])
		if overlap_q2_fraction > self.params['max_query_olap_single']:
		    self.output_message('Failed', align1, align2, 'max_query_olap_single', overlap_q2_fraction, self.params['max_query_olap_single'])
		
	return passed, params
    
    def valid_path(self, aligns, path):
	passed = False
	params = {}
	align_paths = []
	for i in path:
	    align_paths.append(aligns[i])
	
	contig_cover, contig_cover_fraction = self.combined_aligned(align_paths)	
	params['contig_cover_fraction'] = contig_cover_fraction

	#Check if the cover enough of the contig between the two alignments is too low
	if contig_cover_fraction >= self.params['min_fusion_fraction']:
	    passed = True
	    if not passed and self.debug:
		self.output_message('Failed', align_paths[0], align_paths[-1], 'min_fusion_fraction', contig_cover_fraction, self.params['min_fusion_fraction'])
	
	return passed, params
    
    def output_message(self, status, align1, align2, metric, value=None, threshold=None):
	message_data = [status,
	                align1.query, 
	                align1.qstart, align1.qend, align1.target, align1.tstart, align1.tend,
	                align2.qstart, align2.qend, align2.target, align2.tstart, align2.tend,
	                metric]
	
	data = {'status': status,
	        'contig': align1.query,
	        'qstart1': align1.qstart,
	        'qend1': align1.qend,
	        'target1':align1.target,
	        'tstart1': align1.tstart,
	        'tend1': align1.tend,
		'qstart2': align2.qstart,
	        'qend2': align2.qend,
	        'target2': align2.target,
	        'tstart2': align2.tstart,
	        'tend2': align2.tend,
	        'metric': metric
	        }
	        
	        
	if value is None or threshold is None:
	    message = "%(status)s contig:%(contig)s aln1:%(qstart1)s-%(qend1)s %(target1)s %(tstart1)s-%(tend1)s aln2:%(qstart2)s-%(qend2)s %(target2)s %(tstart2)s-%(tend2)s metric:%(metric)s" % data
	else:
	    data['value'] = value
	    data['threshold'] = threshold
	    
	    message = "%(status)s contig:%(contig)s aln1:%(qstart1)s-%(qend1)s %(target1)s %(tstart1)s-%(tend1)s aln2:%(qstart2)s-%(qend2)s %(target2)s %(tstart2)s-%(tend2)s metric:%(metric)s value:%(value)s threshold:%(threshold)s" % data
	    	    
	print message


    def overlap(self, span1, span2, total_len):
        "calculates overlap between two ranges and gives a proportion of the total length covered"
        
        s1 = [int(span1[0]), int(span1[1])]
        s2 = [int(span2[0]), int(span2[1])]
        s1.sort(lambda x,y:x-y)
        s2.sort(lambda x,y:x-y)

        olap = intspan.overlap(s1, s2)
        olap_fraction = 0.0
        if olap:
            olap_fraction = float(olap) / float(total_len)

        return (olap, olap_fraction)

    
    def combined_aligned(self, aligns):
	aligned = []
	olaps = []
	for i in range(len(aligns) - 1):
	    align1_aligned =  int(aligns[i].qend) - int(aligns[i].qstart) + 1
	    align2_aligned =  int(aligns[i + 1].qend) - int(aligns[i + 1].qstart) + 1
	    olap = intspan.overlap([aligns[i].qstart, aligns[i].qend], [aligns[i + 1].qstart, aligns[i + 1].qend])
	    aligned.append(align1_aligned)
	    olaps.append(olap)
	    
	    if i + 1 == len(aligns) - 1:
		aligned.append(align2_aligned)
		
	total_aligned = sum(aligned) - sum(olaps)
	aligned_fraction = float(total_aligned) / float(aligns[0].query_len)
	
	return total_aligned, aligned_fraction

    def query_spanned(self, align1, align2):
        coords = [int(align1.qstart), int(align1.qend), int(align2.qstart), int(align2.qend)]
        span = max(coords) - min(coords) + 1
        fraction = float(span) / float(align1.query_len)

        return (span, fraction)
    
    def find(self, file):
        "extract fusion partners"
        
        #filter alignments
        sys.stderr.write("parsing alignments...")
        aligns = []

        filters = {'bestn':self.params['num_aligns']}	
	        
        ext = os.path.splitext(file)[1]
        if ext == '.psl':
            aligns = psl.parse(file, filters, noblocks=True, noline=True, splice_motif_file=self.splice_motif_file, refseq=self.refseq)
	elif ext == '.sam':
	    aligns = sam.parse(file, filters, noblocks=True, noline=True, splice_motif_file=self.splice_motif_file, refseq=self.refseq)
        else:
            return False
	        
        #set orientation if information is there
        #set strand of alignments
        for a in aligns:
            if a.target_strand:
                a.strand = a.target_strand
            elif a.query_strand:
                a.strand = a.query_strand
            else:
                a.strand = None
		
        sys.stderr.write("done\n")

        #identify splitpairs
        sys.stderr.write("extracting split alignments...")
        fusions = self.get_split_aligns(aligns)
        sys.stderr.write("done\n")
	#sys.exit(0)

        #only keep unique pairs and different partners
        if self.unique:
            fusion_by_contig = {}
            same_partners = {}
            for fusion in fusions:
                if fusion_by_contig.has_key(fusion.contig):
                    fusion_by_contig[fusion.contig] += 1
                else:
                    fusion_by_contig[fusion.contig] = 1

                if fusion.align1.target == fusion.align2.target:
                    same_partners[fusion.contig] = True

            unique_fusions = [fusion for fusion in fusions if fusion_by_contig[fusion.contig] == 1 and not same_partners.has_key(fusion.contig)]
            self.fusions.extend(unique_fusions)
	    
	else:
	    self.fusions.extend(fusions)
	    
	    
    @classmethod
    def compare_lsr(cls, fus1, fus2):
	#sort by read_score, then by coordinate
	if fus1.num_read_pairs >= 0 and fus1.num_spanning_reads >= 0:
	    fus1.read_score = 0.2 * fus1.num_read_pairs + 5 * fus1.num_spanning_reads
	if fus2.num_read_pairs >= 0 and fus2.num_spanning_reads >= 0:
	    fus2.read_score = 0.2 * fus2.num_read_pairs + 5 * fus2.num_spanning_reads
	
	if fus1.read_score and fus2.read_score:
	    if fus1.read_score > fus2.read_score:
		return -1
	    elif fus1.read_score < fus2.read_score:
		return 1
	    else:
		return FusionFinder.compare_coord(fus1, fus2)
	else:
	    return FusionFinder.compare_coord(fus1, fus2)
    
    @classmethod
    def compare_coord(cls, fus1, fus2):
	if fus1.align1.target < fus1.align2.target:
	    align1 = [fus1.align1, fus1.align2]
	elif fus1.align1.target > fus1.align2.target:
	    align1 = [fus1.align2, fus1.align1]
	elif int(fus1.align1.tstart) < int(fus1.align2.tstart):
	    align1 = [fus1.align1, fus1.align2]
	elif int(fus1.align2.tstart) < int(fus1.align1.tstart):
	    align1 = [fus1.align2, fus1.align1]
	else:
	    align1 = [fus1.align1, fus1.align2]
	    
	if fus2.align1.target < fus2.align2.target:
	    align2 = [fus2.align1, fus2.align2]
	elif fus2.align1.target > fus2.align2.target:
	    align2 = [fus2.align2, fus2.align1]
	elif int(fus2.align1.tstart) < int(fus2.align2.tstart):
	    align2 = [fus2.align1, fus2.align2]
	elif int(fus2.align2.tstart) < int(fus2.align1.tstart):
	    align2 = [fus2.align2, fus2.align1]
	else:
	    align2 = [fus2.align1, fus2.align2]
	    
	if align1[0].target < align2[0].target:
	    return -1
	elif align1[0].target > align2[0].target:
	    return 1
	elif align1[1].target < align2[1].target:
	    return -1
	elif align1[1].target > align2[1].target:
	    return 1
	elif int(align1[0].tstart) < int(align2[0].tstart):
	    return -1
	elif int(align1[0].tstart) > int(align2[0].tstart):
	    return 1
	elif int(align1[1].tstart) < int(align2[1].tstart):
	    return -1
	elif int(align1[1].tstart) > int(align2[1].tstart):
	    return 1
	elif int(align1[0].tend) < int(align2[0].tend):
	    return -1
	elif int(align1[0].tend) > int(align2[0].tend):
	    return 1
	elif int(align1[1].tend) < int(align2[1].tend):
	    return -1
	elif int(align1[1].tend) > int(align2[1].tend):
	    return 1
	else:
	    return 0
	
    def report(self, outdir, append=False, filtering=False, cnv_file=None):
        if append:
            mode = 'a'
        else:
            mode = 'w'

        unfiltered_outfile = outdir + "/fusions.tsv"
        filtered_outfile = outdir + "/fusions_filtered.tsv"
        local_outfile = outdir + "/local.tsv"

	column_header = ['id', 'contig', 'contig_size', 'genomic_regions', 'contig_regions', 'strands', 'flanking_pairs', 'breakpoint_pairs', 'spanning_reads', 'rearrangement', 'breakpoint', 'size', 'genes', 'transcripts', 'senses', 'exons/introns', 'exon_bounds', 'reciprocal', 'descriptor', 'orientations', "5'gene", "3'gene", "5'exon", "3'exon", 'frame', 'alignment_params', 'type', 'dbsnp', 'dgv']
	
	if cnv_file:
	    column_header.insert(-1, "cnv")
	    
	if not filtering:
            if mode == 'w' and os.path.exists(unfiltered_outfile):
                os.remove(unfiltered_outfile)
            out = open(unfiltered_outfile, mode)

	    out.write('\t'.join(column_header) + '\n')
            self.output(out, self.fusions)
	    	    	
	else:
	    remove_bp = []
	    
	    good_contigs = {}
	    reciprocals = {}
	    good_fusions = []
	    for bp, fusions in self.fusion_groups.iteritems():	
		for i in range(len(fusions)):		    
		    if not fusions[i].exclude and fusions[i].read_support:
			good_contigs[fusions[i].align1.query] = True
		    
			if fusions[i].reciprocal is not None:
			    reciprocals[fusions[i].reciprocal] = True
			    
	    #include all reciprocal partners of good contigs
	    for bp, fusions in self.fusion_groups.iteritems():
		for fusion in fusions:
		    if reciprocals.has_key(fusion.breakpoint):
			good_contigs[fusion.align1.query] = True
		    
	    for bp, fusions in self.fusion_groups.iteritems():
		for fusion in fusions:
		    if good_contigs.has_key(fusion.align1.query):
			good_fusions.append(fusion)
		    
		    elif not fusion.exclude and fusion.read_support:
			good_fusions.append(fusion)
			
	    outfile = filtered_outfile
	    if mode == 'w' and os.path.exists(outfile):
		os.remove(outfile)
	    out = open(outfile, mode)
	    out.write('\t'.join(column_header) + '\n')
	    self.output(out, good_fusions)
	    
	    #local
	    del good_fusions[:]
	    for bp, fusions in self.fusion_groups.iteritems():
		for fusion in fusions:
		    if fusion.local and fusion.read_support:
			good_fusions.append(fusion)
			
	    if mode == 'w' and os.path.exists(local_outfile):
		os.remove(local_outfile)
	    out = open(local_outfile, mode)
	    out.write('\t'.join(column_header) + '\n')
	    self.output(out, good_fusions)
		    
	    	    
    def overlap_cnvs(self, fusions, cnvs):
	for fusion in fusions:
	    if not fusion.lsr_region:
		continue
	    
	    if fusion.mutation != 'translocation':
		fusion.cnv = "%s:%s" % (fusion.lsr_region[0], cnvs.overlap(fusion.lsr_region[0], fusion.lsr_region[1], fusion.lsr_region[2])) 
	    else:
		region1 = fusion.lsr_region[0]
		region2 = fusion.lsr_region[1]
		
		cnv1 = cnvs.overlap(region1[0], region1[1], region1[2])
		cnv2 = cnvs.overlap(region2[0], region2[1], region2[2])
		
		fusion.cnv = "%s:%s;%s:%s" % (region1[0], cnv1, region2[0], cnv2)
		
	    
    def overlap_dbsnp(self, fusions_grouped_by_chrom, max_del=1000000):
	for chrom in fusions_grouped_by_chrom.keys():
	    proper_chrom = tools.proper_chrom(chrom, self.genome)
	    snp_overlap = dbsnp.prepare_overlap(self.genome, proper_chrom)
	    
	    for event, fusions in fusions_grouped_by_chrom[chrom].iteritems():
		fusion = fusions[0]
		
		if fusion.mutation == 'deletion':
		    if fusion.size() >= max_del:
			continue
		    
		    breaks = [fusion.breaks[0], fusion.breaks[1]]
		    breaks.sort(key = int)

		    known = dbsnp.find_concordants({'type':'del', 'chrom':proper_chrom, 'start':breaks[0] + 1, 'end':breaks[1] - 1, 'size': breaks[1] - breaks[0] + 1}, snp_overlap, exact=False)
		
		    if known:
			for f in fusions:
			    f.dbsnp = ','.join(known)
		    else:
			for f in fusions:
			    f.dbsnp = '-'
			    			
	    snp_overlap.finish()
	    
    def overlap_dgv(self, fusions_grouped_by_chrom):
	dgv_overlap = dgv.prepare_overlap(self.genome)
	
	for chrom in fusions_grouped_by_chrom.keys():
	    proper_chrom = tools.proper_chrom(chrom, self.genome)
	    
	    for event, fusions in fusions_grouped_by_chrom[chrom].iteritems():
		fusion = fusions[0]
	    
		if fusion.mutation in ('deletion', 'inversion'):
		    proper_chrom = tools.proper_chrom(fusion.align1.target, self.genome)
		
		    if fusion.mutation == 'deletion':
			var_type = 'del'
		    else:
			var_type = 'inv'
		    
		    known = dgv.find_concordants({'type':var_type, 'chrom':proper_chrom, 'start':fusion.breaks[0] + 1, 'end':fusion.breaks[1] - 1}, dgv_overlap)
		    
		    if known:
			for f in fusions:
			    f.dgv = ','.join(known)
		    else:
			for f in fusions:
			    f.dgv = '-'
		    
	dgv_overlap.finish()

    def output(self, out, fusions, tsv=True):
        for fusion in sorted(fusions, cmp=lambda x,y: x.output_order - y.output_order):
            out.write(fusion.details(tsv=tsv) + "\n")
	    
    def add_spanning_reads(self, groups):
	for group, fusions in groups.iteritems():
	    total = 0
	    
	    for fusion in fusions:
		total += fusion.num_spanning_reads
		
	    for fusion in fusions:
		fusion.num_spanning_reads = total
		
    
    def group_by_breakpoint(self):
	breakpoints = {}
	for fusion in self.fusions:
	    if not breakpoints.has_key(fusion.breakpoint):
		breakpoints[fusion.breakpoint] = [fusion]
	    else:
		breakpoints[fusion.breakpoint].append(fusion)
		
	reciprocal_breakpoints = [breakpoint for breakpoint in breakpoints.keys() if breakpoints[breakpoint][0].reciprocal is not None]
	
	event_id = 1
	event_ids = {}
	output_orders = {}
	order = 0
		
	for breakpoint in reciprocal_breakpoints:
	    if event_ids.has_key(breakpoint):
		if len(breakpoints[breakpoint]) == 1:
		    fusion = breakpoints[breakpoint][0]
		    fusion.event_id = event_ids[breakpoint]
		    fusion.output_order = output_orders[fusion.event_id]
		    
		else:
		    for i in range(len(breakpoints[breakpoint])):
			fusion = breakpoints[breakpoint][i]
			output_order = output_orders[event_ids[breakpoint].split('.')[0]]
			fusion.event_id = "%s.%d" % (event_ids[breakpoint], i+1)
			fusion.output_order = output_order
			output_order += 1
	    
	    else:
		if len(breakpoints[breakpoint]) == 1:
		    fusion = breakpoints[breakpoint][0]
		    fusion.event_id = "%da" % (event_id)
		    fusion.output_order = order
		    event_ids[fusion.reciprocal] = "%db" % (event_id)
		    order += 1
		    output_orders[event_ids[fusion.reciprocal]] = order
		    order += 1
		    event_id += 1
		    
		else:
		    for i in range(len(breakpoints[breakpoint])):
			fusion = breakpoints[breakpoint][i]
			fusion.event_id = "%da.%d" % (event_id, i+1)
			fusion.output_order = order
			order += 1
			
		    event_ids[breakpoints[breakpoint][0].reciprocal] = "%db" % (event_id)
		    output_orders[event_ids[breakpoints[breakpoint][0].reciprocal]] = order
		    event_id += 1
		    order += 1
		    
	
	for breakpoint in sorted(breakpoints.keys()):
	    if breakpoint in reciprocal_breakpoints:
		continue
	    if len(breakpoints[breakpoint]) == 1:
		fusion = breakpoints[breakpoint][0]
		fusion.event_id = "%d" % (event_id)
		fusion.output_order = order
		order += 1
		event_id += 1
		    
	    else:
		for i in range(len(breakpoints[breakpoint])):
		    fusion = breakpoints[breakpoint][i]
		    fusion.event_id = "%d.%d" % (event_id, i+1)
		    fusion.output_order = order
		    order += 1
		    
		event_id += 1
		    	    
	    
	self.fusion_groups = breakpoints

	
    def group_by_chrom(self, mutations=None):
	fusions_by_chrom = {}
	for breakpoint, fusions in self.fusion_groups.iteritems():
	    if fusions[0].align1.target != fusions[0].align2.target:
		continue
	    
	    chrom = fusions[0].align1.target
	    key = fusions[0].mutation + fusions[0].breakpoint
	    
	    if not fusions_by_chrom.has_key(chrom):
		fusions_by_chrom[chrom] = {}
		fusions_by_chrom[chrom][key] = []
	    elif not fusions_by_chrom[chrom].has_key(key):
		fusions_by_chrom[chrom][key] = []
	    fusions_by_chrom[chrom][key].extend(fusions)
	    
	return fusions_by_chrom
	
    def parse_fusions_dir(self, path):
	fusion_dirs = sorted(glob.glob(os.path.join(path, '*')))
		
	fusion_files = []
	num_fusion_dirs = 0
	missing_dirs = []
	for job_num in range(1, len(fusion_dirs)+1):
	    cluster_outdir = "%s/%s" % (args[0], job_num)
	    if os.path.isdir(cluster_outdir):
		num_fusion_dirs += 1
		    
		fusion_file = cluster_outdir + '/fusions.tsv'		    
		if os.path.exists(fusion_file):
		    fusion_files.append(fusion_file)
		else:
		    missing_dirs.append(str(job_num))
		    sys.stdout.write("%s does not have fusion file\n" % (cluster_outdir))
			
	sys.stdout.write("fusion dirs:%s fusion files:%s\n" % (num_fusion_dirs, len(fusion_files)))
		
	if num_fusion_dirs == len(fusion_files):
	    concat_outfile = outdir + "/fusions.tsv"
	    concat_out = open(concat_outfile, 'w')
		    
	    #just for check if header is to be written
	    count = 0
	    for fusion_file in fusion_files:
		infile = open(fusion_file, 'r')
		if count == 0:
		    concat_out.writelines(open(fusion_file, 'r').readlines())
		else:
		    concat_out.writelines(open(fusion_file, 'r').readlines()[1:])
		infile.close()
			
		print fusion_file
		self.parse_fusions(fusion_file)
		count += 1
			
	    concat_out.close()
	else:
	    sys.stdout.write("missing (%s):%s\n" % (len(missing_dirs), ','.join(missing_dirs)))
	    
	    
    def parse_fusions(self, fusions_file):
        genes = re.compile(r'(?P<gene1>.+)\((?P<orient1>.+)\)(?P<gene2>.+)\((?P<orient2>.+)\)')
        
	output_order = 0
        for line in open(fusions_file, 'r'):
            cols = line.rstrip('\n').split('\t')
	    if cols[0] == 'id':
		continue
	    
	    #empty line
	    if len(cols) < 2:
		continue
	    
            align1 = Alignment(None)
            align2 = Alignment(None)
	    
	    align1.query = align2.query = cols[1]
	    align1.query_len = align2.query_len = cols[2]

	    targets = cols[3].split(',')
            align1.target, align1.tstart, align1.tend = re.split('[:-]', targets[0])
            align2.target, align2.tstart, align2.tend = re.split('[:-]', targets[1])

            queries = cols[4].split(',')
            align1.qstart, align1.qend = re.split('-', queries[0])
            align2.qstart, align2.qend = re.split('-', queries[1])

            align1.strand, align2.strand = cols[5].split(',')

            fusion = Fusion(align1, align2)
	    fusion.event_id = cols[0]
	    fusion.output_order = output_order

            fusion.num_flanking_pairs = int(cols[6])
	    fusion.num_breakpoint_pairs = [int(n) for n in cols[7].split(',')]
	    fusion.num_read_pairs = fusion.num_flanking_pairs + min(fusion.num_breakpoint_pairs)
            fusion.num_spanning_reads = int(cols[8])

            fusion.mutation = cols[9]
	    fusion.breakpoint = cols[10]
	    
	    
	    fusion.gene1, fusion.gene2 = cols[12].split(',')
	    fusion.txt1_name, fusion.txt2_name = cols[13].split(',')
	    fusion.sense1, fusion.sense2 = cols[14].split(',')
	    fusion.exon1, fusion.exon2 = cols[15].split(',')
	    fusion.exon_bound1, fusion.exon_bound2 = cols[16].split(',')	    
	    
	    if cols[17].upper() == 'NA':
		fusion.reciprocal = None
	    else:
		fusion.reciprocal = cols[17]
	    fusion.descriptor = cols[18]
		
	    fusion.orients = [None, None]
	    orient1, orient2 = cols[19].split(',')
	    if orient1 == 'L':
		fusion.orients[0] = True
	    else:
		fusion.orients[0] = False
	    if orient2 == 'L':
		fusion.orients[1] = True
	    else:
		fusion.orients[1] = False
	    
	    fusion.gene5prime = cols[20]
	    fusion.gene3prime = cols[21]
	    fusion.exon5prime = cols[22]
	    fusion.exon3prime = cols[23]
	    fusion.frame = cols[24]
	    
	    params = cols[25].split(',')
            fusion.overlap_target_fraction = float(params[0].split(':')[1])
            fusion.overlap_query_fraction = float(params[1].split(':')[1])
            fusion.contig_cover_fraction = float(params[2].split(':')[1])
            align1.identity = float(params[3].split(':')[1])
            align2.identity = float(params[4].split(':')[1])
			
	    fusion.fusion_type = cols[26]
	    fusion.dbsnp = cols[27]
	    fusion.dgv = cols[28]
	    
            self.fusions.append(fusion)
	    
	    output_order += 1
	    
    def extract_contig_seq(self, contigs_file):
	assembly = Assembly(None, fasta=contigs_file)
	contigs = dict((fusion.align1.query, True) for fusion in self.fusions)
	self.contig_seq = dict((contig.num, contig.sequence) for contig in assembly.get_contigs(ids=contigs.keys(), sequence=True))
	    			    

    def annotate(self, genome=None, model=None, breakpoint_buffer=0):
        self.ff = FeatureFinder(genome, model)
		
	#storing unique contig ids for extracting sequences
	contigs = {}
	
	for fusion in self.fusions: 
	    txt1 = txt2 = None
	    gene1 = gene2 = 'NA'
	    if fusion.breakpoint.upper() != 'NA':
		target1, bp1, target2, bp2 = re.split('[|:]', fusion.breakpoint)
		target1 = tools.proper_chrom(target1, genome)
		target2 = tools.proper_chrom(target2, genome)
				
		feature1, txt1 = self.ff.get_feature(' '.join((target1, bp1, bp1)), gene_strand=True, txt_obj=True, strand=fusion.align1.orient)
		if txt1 is not None:
		    gene1, transcript1, exon1 = feature1.rsplit('|',1)[0].split(':')
		    strand1 = txt1.strand
		    fusion.txt1 = txt1
		    fusion.exon1 = exon1.split('(')[0]
		    
		feature2, txt2 = self.ff.get_feature(' '.join((target2, bp2, bp2)), gene_strand=True, txt_obj=True, strand=fusion.align2.orient)
		if txt2 is not None:
		    gene2, transcript2, exon2 = feature2.rsplit('|',1)[0].split(':')
		    strand2 = txt2.strand
		    strand2 = txt2.strand
		    fusion.txt2 = txt2
		    fusion.exon2 = exon2.split('(')[0]
		
		#see if transcripts overlap
		#if target1 == target2:
		    #if txt2 and gene1.upper() == 'NA' and gene2.upper() != 'NA':
			#if intspan.overlap([txt2.txStart, txt2.txEnd], [fusion.align1.tstart, fusion.align1.tend]) and \
			   #intspan.overlap([txt2.txStart, txt2.txEnd], [fusion.align2.tstart, fusion.align2.tend]):
			    #gene1 = gene2
			    
		    #if txt1 and gene2.upper() == 'NA' and gene1.upper() != 'NA':
			#if intspan.overlap([txt1.txStart, txt1.txEnd], [fusion.align1.tstart, fusion.align1.tend]) and \
			   #intspan.overlap([txt1.txStart, txt1.txEnd], [fusion.align2.tstart, fusion.align2.tend]):
			    #gene2 = gene1
			    
		
	    fusion.gene1 = gene1
            fusion.gene2 = gene2
	    

            if gene1 and txt1 and gene1.upper() != 'NA':
                if strand1 == fusion.align1.strand:
                    fusion.sense1 = '+'
                else:
                    fusion.sense1 = '-'

            if gene2 and txt2 and gene2.upper() != 'NA':
                if strand2 == fusion.align2.strand:
                    fusion.sense2 = '+'
                else:
                    fusion.sense2 = '-'
		
	    fusion.order_gene()
	    fusion.at_exon_bound()
	    fusion.check_frame(self.ff, self.refseq, self.contig_seq[fusion.align1.query])
	    
	    
    def post_filter(self):
        #fusions by contig
        fusions_by_contig = {}
        for i in range(len(self.fusions)):
            if not fusions_by_contig.has_key(self.fusions[i].contig):
                fusions_by_contig[self.fusions[i].contig] = [i]
            else:
                fusions_by_contig[self.fusions[i].contig].append(i)

        for contig, fusion_idx in fusions_by_contig.iteritems():
            contig_local = False
            for idx in fusion_idx:
                fusion = self.fusions[idx]
		
		if not fusion.local:
		    if self.ff and fusion.is_local(self.ff.model, self.ff.annot_file):
			fusion.local = True
			contig_local = True
			break
		    elif fusion.is_local():
			fusion.local = True
			contig_local = True
			break
			                
            if contig_local:
                for idx in fusion_idx:
                    self.fusions[idx].contig_local = True
                    self.fusions[idx].exclude = True

        self.post_filter_others()
	
	
    def get_bad_gene_pairs(self, infile):
	gene_pairs = {}
	
	for line in open(infile, 'r'):
	    genes = line.rstrip('\n').split()

	    if len(genes) != 2:
		continue
	    
	    if gene_pairs.has_key(genes[0]):
		gene_pairs[genes[0]][genes[1]] = True
	    else:
		gene_pairs[genes[0]] = {genes[1]:True}
		
	    if gene_pairs.has_key(genes[1]):
		gene_pairs[genes[1]][genes[0]] = True
	    else:
		gene_pairs[genes[1]] = {genes[0]:True}

	return gene_pairs
    
    def get_bad_breakpoints(self, infile):
	breakpoints = {}
	
	for line in open(infile, 'r'):
	    breakpoint = line.rstrip('\n')
	    breakpoints[breakpoint] = True

	return breakpoints
    
    def post_filter_others(self):
        exclude_indices = {}
        mappings = {}
        contig2fusions = {}
	
	bad_gene_pairs = None
	if self.filtering.has_key('bad_gene_pairs'):
	    bad_gene_pairs = self.get_bad_gene_pairs(self.filtering['bad_gene_pairs'])
	    
	bad_breakpoints = None
	if self.filtering.has_key('bad_breakpoints'):
	    bad_breakpoints = self.get_bad_breakpoints(self.filtering['bad_breakpoints'])
	    
	for i in range(len(self.fusions)):
            fusion = self.fusions[i]

            if fusion.contig_local or fusion.local:
                fusion.exclude = True
		
	    if not fusion.exclude and self.filtering.has_key('bad_gene_pairs') or self.filtering.has_key('bad_breakpoints'):
		if fusion.gene1.upper() != 'NA' and fusion.gene2.upper() != 'NA' and bad_gene_pairs.has_key(fusion.gene1) and bad_gene_pairs[fusion.gene1].has_key(fusion.gene2):
		    fusion.exclude = True
		    
		if (fusion.gene1.upper() == 'NA' or fusion.gene2.upper() == 'NA') and bad_breakpoints.has_key(fusion.breakpoint):
		    fusion.exclude = True
	    
            #basic filtering
            if not fusion.exclude and self.filtering.has_key('chrM') and self.filtering['chrM']:
                if fusion.align1.target.lower() == 'chrm' or fusion.align2.target.lower() == 'chrm' or fusion.align1.target.lower() == 'chrmt' or fusion.align2.target.lower() == 'chrmt':
                    fusion.exclude = True

            if not fusion.exclude and self.filtering.has_key('haplotype') and self.filtering['haplotype']:
                if (re.match('^chr', fusion.align1.target, re.IGNORECASE) and '_' in fusion.align1.target) or (re.match('^chr', fusion.align2.target, re.IGNORECASE) and '_' in fusion.align2.target):
                    fusion.exclude = True
		    
	    if fusion.num_read_pairs >= self.filtering['readpairs'][0] and fusion.num_read_pairs <= self.filtering['readpairs'][1]:
                fusion.good_read_pair = True			
            if fusion.num_spanning_reads >= self.filtering['min_span_reads']:
                fusion.good_span_read = True
            if fusion.good_read_pair and fusion.good_span_read:
                fusion.read_support = True
		
		            
    def post_process(self, options):
	# define mutation
	if self.genome:
	    for fusion in ff.fusions:
		fusion.define_mutation(ff.cytobands, self.genome)
	    		
	# add annotation
	if options.gene_model:
	    self.annotate(genome=self.genome, model=self.model)
	    
	# find reciprocal pairs
	ff.find_reciprocal()
	
	# group fusions by breakpoint - for proper indexing reciprocal events
	self.group_by_breakpoint()
	    
	# collect read support
	if options.genome_bam or options.contigs_bam:
	    self.read_support(genome_bamfile=options.genome_bam, contigs_bamfile=options.contigs_bam, min_mapq=options.mapq_genome, sample_type=options.sample_type, library=options.library, keep_mito=options.keep_mito, genome_coord_buffer=options.genome_coord_buffer)
	
	# add spanning reads 
	if options.add_spanning_reads:
	    self.add_spanning_reads(ff.fusion_groups)	
	
	# add CNV
	if options.cnv_file:
	    cnvs = CNV(options.cnv_file)
	    cnvs.extract()
	    self.overlap_cnvs(self.fusions, cnvs)
	    
	if options.olap_annot:
	    chrom_groups = self.group_by_chrom()
	    self.overlap_dbsnp(chrom_groups)
	    self.overlap_dgv(chrom_groups)
	    
    def read_support(self, genome_bamfile=None, contigs_bamfile=None, min_mapq=None, sample_type=None, library=None, keep_mito=False, genome_coord_buffer=None):
	sys.stderr.write("finding read support...")
	
	genome_bam = None
	contigs_bam = None
	if genome_bamfile and os.path.exists(genome_bamfile):
	    genome_bam = BAM(genome_bamfile, min_mapq=min_mapq, sample_type=sample_type)
            
	if contigs_bamfile and os.path.exists(contigs_bamfile):
	    contigs_bam = BAM(contigs_bamfile, min_mapq=options.mapq_contigs)
	
	for fusion in self.fusions:
	    spanning_reads = []
	    if contigs_bam:
		spanning_reads = fusion.find_spanning_reads(contigs_bam, library, breakpoint_buffer=self.params['min_ctg_bp_overlap'])
			    	    
	    if genome_bam:
		fusion.find_read_pairs(genome_bam, genome_coord_buffer=genome_coord_buffer, breakpoint_buffer=self.params['min_genome_bp_overlap'], keep_mito=keep_mito, reads=dict((r.qname, r) for r in spanning_reads))
	
	sys.stderr.write("done\n")
	
			
    def find_reciprocal(self, close=100):
	breakpoint_to_fusions = {}	
	for i in range(len(self.fusions)):
	    if not breakpoint_to_fusions.has_key(self.fusions[i].breakpoint):
		breakpoint_to_fusions[self.fusions[i].breakpoint] = []
	    breakpoint_to_fusions[self.fusions[i].breakpoint].append(i)
	
	pairs = {}
	for i in range(len(self.fusions) - 1):
	    if not self.fusions[i].breaks:
		continue
	    
	    for j in range(i+1, len(self.fusions)):
		if not self.fusions[j].breaks:
		    continue
		
		if i == j or self.fusions[i].mutation != self.fusions[j].mutation:
		    continue
				
		if self.fusions[i].mutation == 'translocation':
		    if self.fusions[i].align1.target != self.fusions[j].align1.target or self.fusions[i].align2.target != self.fusions[j].align2.target:
			continue		    
		    
		elif self.fusions[i].align1.target != self.fusions[j].align1.target:
		    continue
		    
		if pairs.has_key(self.fusions[i].breakpoint) and pairs[self.fusions[i].breakpoint].has_key(self.fusions[j].breakpoint):
		    continue
		    		    
		#align breakpoints to targets
		breaks1, breaks2 = self.fusions[i].breaks, self.fusions[j].breaks
		if (abs(int(breaks1[0]) - int(breaks2[0])) <= close or abs(int(breaks1[1]) - int(breaks2[1])) <= close or\
		    (self.fusions[i].gene1.upper() != 'NA' and self.fusions[i].gene1 == self.fusions[j].gene1 and self.fusions[i].gene2 == self.fusions[j].gene2)) and\
		   self.fusions[i].orients[0] != self.fusions[j].orients[0] and self.fusions[i].orients[1] != self.fusions[j].orients[1]:
		    		    
		    if not pairs.has_key(self.fusions[i].breakpoint):
			pairs[self.fusions[i].breakpoint] = {}
		    pairs[self.fusions[i].breakpoint][self.fusions[j].breakpoint] = True
		    
					
	for bp1 in pairs.keys():
	    if len(pairs[bp1].keys()) == 1:
		bp2 = pairs[bp1].keys()[0]
		
		for i in breakpoint_to_fusions[bp1]:
		    self.fusions[i].reciprocal = bp2
		for i in breakpoint_to_fusions[bp2]:
		    self.fusions[i].reciprocal = bp1
		    	
		    
def create_jobs_file(align_dir, outdir, num, params):
    cluster_dir = outdir + "/cluster"
    if not os.path.isdir(cluster_dir):
	os.mkdir(cluster_dir)
	
    jobs_file = cluster_dir + "/jobs"
    sys.stdout.write("creating job file:%s\n" % (num))
    jobs = open(jobs_file, 'w')
    
    script = os.path.abspath(sys.argv[0])
    source = PACKAGE_DIR + "/setup"

    for i in range(1,num+1):
	outdir = cluster_dir + "/" + str(i)
	if not os.path.isdir(outdir):
	    os.mkdir(outdir)
	
	align_file = "%s/seq.%s.psl" % (align_dir, i)
	line = "source %s;%s %s %s %s -C 0 -B %s -b %s -G %s %s" % (source, PYTHON, script, align_file, outdir, options.genome_bam, options.contigs_bam, options.gene_model[0], options.gene_model[1])
	if options.library:
	    line += " -l " + options.library
	    
	for param, val in params.iteritems():
	    if type(val).__name__ in ('int', 'float'):
		line += " --%s %s" % (param, val)
	    else:
		line += " --%s" % (param)
	
	jobs.write(line + '\n')
        
    jobs.close()
    
    return jobs_file


if __name__ == '__main__':
    usage = "Usage: %proj alignment_[file|dir] output_dir"

    parser = OptionParser(usage=usage, version="%prog " + VERSION)
    parser.add_option("-m", "--max_match_fraction", dest="match_fraction", help="maximum match fraction for individual alignment. Default:0.999", default=0.999, type='float')
    parser.add_option("-q", "--max_query_olap", dest="query_olap", help="maximum query overlap. Default:0.05", default=0.05, type='float')
    parser.add_option("-Q", "--max_query_olap_single", dest="query_olap_single", help="maximum single region query overlap. Default:0.3", default=0.3, type='float')
    parser.add_option("-t", "--max_target_olap", dest="target_olap", help="maximum target overlap. Default:0.0", default=0.0, type='float')
    parser.add_option("-f", "--min_fusion_fraction", dest="fusion_fraction", help="minimum fusion fraction. Default:0.90", default=0.90, type='float')
    parser.add_option("-a", "--num_aligns", dest="num_aligns", help="number alignments. Default:20", default=20, type='int')
    parser.add_option("-A", "--num_best", dest="num_best", help="number of best alignments to use. Default:5", default=5, type='int')
    parser.add_option("-i", "--min_identity", dest="min_identity", help="minimum identity. Default:98", type='float', default=98.0)
    parser.add_option("-s", "--min_span_reads", dest="min_span_reads", help="minimum spanning reads. Default:2", type='int', default=2)
    parser.add_option("-p", "--min_read_pairs", dest="min_read_pairs", help="minimum read pairs. Default:4", type='int', default=4)
    parser.add_option("-P", "--max_read_pairs", dest="max_read_pairs", help="maximum read pairs. Default:2000", type='int', default=2000)
    
    parser.add_option("-x", "--min_ctg_bp_overlap", dest="min_ctg_bp_overlap", help="minimum contig breakpoint overlap. Default:4", type='int', default=4)
    parser.add_option("-e", "--min_genome_bp_overlap", dest="min_genome_bp_overlap", help="minimum genome breakpoint overlap. Default:4", type='int', default=4)
    
    parser.add_option("-d", "--debug", dest="debug", help="debug", action="store_true", default=False)
    parser.add_option("-u", "--unique", dest="unique", help="only keep unique pairs (not same target)", action="store_true", default=False)
    parser.add_option("-c", "--genome_coord_buffer", dest="genome_coord_buffer", help="genome coordinate buffer for extracting reads. Default:2000", default=2000, type='int')
    parser.add_option("-B", "--genome_bam", dest="genome_bam", help="genomic bam file")
    parser.add_option("-b", "--contigs_bam", dest="contigs_bam", help="contigs bam file")
    parser.add_option("-F", "--fusions", dest="fusions", help="fusions input file given", action="store_true", default=False)
    parser.add_option("-l", "--library", dest="library", help="library name (required for using reads-contigs bam file)")
    parser.add_option("-M", "--keep_mito", dest="keep_mito", help="include mitochondria for read pair evidence", action="store_true", default=False)
    parser.add_option("-Y", "--mapq_genome", dest="mapq_genome", help="minimum mapping quality for reads to genome. Default:1", type='int', default=1)
    parser.add_option("-y", "--mapq_contigs", dest="mapq_contigs", help="minimum mapping quality for reads to contigs. Default:0", type='int', default=0)
    parser.add_option("-X", "--filter", dest="filter", help="post-filter", action="store_true", default=False)
    parser.add_option("-G", "--gene_model", dest="gene_model", help="gene model used for annotation e.g. hg18 k where k=known_genes, e=ensembl, r=refseq", nargs=2)
    
    parser.add_option("-k", "--cluster", dest="cluster", help="cluster")
    parser.add_option("-v", "--bad_gene_pairs", dest="bad_gene_pairs")
    parser.add_option("-w", "--bad_breakpoints", dest="bad_breakpoints")
    parser.add_option("-V", "--cnv_file", dest="cnv_file")    
    parser.add_option("-S", "--add_spanning_reads", dest="add_spanning_reads", action="store_true", default=False)
    parser.add_option("-T", "--type", dest="sample_type", help="sample type: transcriptome(default), genome, exome - can be ignored", default='transcriptome')
    parser.add_option("-O", "--olap_annot", dest="olap_annot", help="overlaps dbsnp and dgv", action="store_true", default=False)
    parser.add_option("-C", "--contigs_file", dest="contigs_file", help="contigs file")
    
    (options, args) = parser.parse_args()

    params = {}
    params['max_match_fraction'] = options.match_fraction
    params['max_query_olap'] = options.query_olap
    params['max_query_olap_single'] = options.query_olap_single
    params['max_target_olap'] = options.target_olap
    params['min_fusion_fraction'] = options.fusion_fraction
    params['min_identity'] = options.min_identity
    params['num_aligns'] = options.num_aligns
    params['num_best'] = options.num_best
    params['min_ctg_bp_overlap'] = options.min_ctg_bp_overlap
    params['min_genome_bp_overlap'] = options.min_genome_bp_overlap
        
    tools.output_log(VERSION, sys.argv, [vars(options)], args[1])
    
    if len(args) == 2:	    
        ff = FusionFinder(params, debug=options.debug, unique=options.unique)
	if options.gene_model:
	    ff.genome, ff.model = options.gene_model
	    ff.prepare_annotation()
	if options.contigs_file:
	    ff.extract_contig_seq(contigs_file=options.contigs_file)
	    
        outdir = args[1]
	       
        if options.fusions:
	    if (os.path.isdir(args[0])):
		ff.parse_fusions_dir(args[0])
	    else:
		ff.parse_fusions(args[0])
		                
        else:
            if os.path.isdir(args[0]):                
                align_files = sorted(glob.glob(os.path.join(args[0], '*.*')))
		
		if not options.cluster:
		    for i in range(len(align_files)):
			sys.stdout.write("processing alignment file #%d\n" % (i+1))
			ff.find(align_files[i])
			
		else:
		    jobs_file = create_jobs_file(args[0], args[1], len(align_files), params)
		    print jobs_file			
			
            else:
                ff.find(args[0])
		
		
	#defines breakpoint, group, read_support, cnv
	if not options.cluster:
	    ff.post_process(options)
	
	if not options.fusions and not options.cluster:
	    ff.report(outdir, filtering=False)

	#filtering
	if options.filter:
            #ff.filtering['no_multimap'] = options.no_multimap
            ff.filtering['readpairs'] = [options.min_read_pairs, options.max_read_pairs]
            ff.filtering['min_span_reads'] = options.min_span_reads
            ff.filtering['gene_fusions'] = True

	    if options.bad_gene_pairs:
		ff.filtering['bad_gene_pairs'] = options.bad_gene_pairs
	    if options.bad_breakpoints:
		ff.filtering['bad_breakpoints'] = options.bad_breakpoints
		
            ff.post_filter()

	    
	    #output
	    ff.report(outdir, filtering=options.filter, cnv_file=options.cnv_file)

            

