#!/usr/bin/env python

# written by:    Ka Ming Nip
# updated on:    September 24, 2012
# compatibility: Trans-ABySS 1.4.4

import os, sys, string, re, subprocess
from optparse import OptionParser,OptionGroup
from utilities import cfg, job_runner
from utilities.job_runner import MyJob

try:
    TRANSABYSS_PATH = os.environ['TRANSABYSS_PATH']
except KeyError:
    sys.stderr.write("ERROR: Cannot find TRANSABYSS_PATH in env!\n")
    sys.exit(1)

fileformats={}
fileformats['bam']=[]
fileformats['txt']=['gz','bz2']
fileformats['fq']=['gz']
fileformats['fastq']=['gz']

ta_setup = "source " + TRANSABYSS_PATH + "/setup"

num_reads_per_fastq_config_key = "num_reads_per_fastq"
samtofastq_config_key = "samtofastq.jar"
convertreads_config_key = "convert_reads"
convertreads_config_value_qseq2fastq = "qseq2fastq.py"
convertreads_config_value_abysstofastq = "abyss-tofastq"

#Given the path, determine the format and the type of compression
def properformat(path):
    formats = fileformats.keys()
    path_lowered = path.lower()

    for format in formats:
        compressions = fileformats[format]
        if path_lowered.endswith(format.lower()):
            return (format, None)
        elif not compressions is None and len(compressions) > 0:
            for compression in compressions:                
                if path_lowered.endswith(format.lower() + '.' + compression.lower()):
                    return (format, compression)
    return (None, None)


#Given the path, format, and compression, give the path to the file's (un)compressed equivalents
def getequivalents(path, format_compression):
    (format, compression) = format_compression
    if format is None:
        return []
    
    compressions = fileformats[format]
    equivalents = []    
    
    #default: assume no compression
    extensionstring = ""
    uncompressed_prefix = path
    
    #there was compression
    if not compression is None:
        extensionstring = '.' + compression
        uncompressed_prefix = path.rstrip(extensionstring)
        equivalents.append(uncompressed_prefix)
    
    for c in compressions:
        if c != compression:
            equivalents.append(uncompressed_prefix + '.' + c)
    
    return equivalents


def generate_commands(paths, outdir, prefix, outreadslist, sortmem, local, settings, single_end=False, project=None):
    # Determine if we want to split reads files
    num_reads_per_fastq = cfg.get_value(settings, 'commands', num_reads_per_fastq_config_key) #config_reader.getCommand(num_reads_per_fastq_config_key, config_type, project)
    split = False
    if num_reads_per_fastq != None:
        try:
            if long(num_reads_per_fastq) > 0:
                split = True
        except ValueError:
            split = False
    
    task_counter = 1
    numtasks = len(paths) * 2
    digits = len(str(numtasks))
    converter = cfg.get_value(settings, 'commands', convertreads_config_key) #config_reader.getCommand(convertreads_config_key, config_type, project)
    qseq2fastq_script = "$TRANSABYSS_PATH/utilities/qseq2fastq.py"

    if converter == None:
        sys.stderr.write("ERROR: No 'qseq-to-fastq' converter found.\n")
        sys.exit(1)
    elif converter != convertreads_config_value_qseq2fastq and converter != convertreads_config_value_abysstofastq:
        sys.stderr.write("ERROR: Unknown 'qseq-to-fastq' converter '" + converter + "'!\n")
        sys.exit(1)

    jobs = []
    files_generated = []

    for p in paths:
        (format, compression) = properformat(p)
        if format == "bam":
            bam_basename = os.path.basename(p)
            if split:
                # TASK: Split the reads in the BAM file ...
                samtofastq_job_content = ''
                task_id1 = str(task_counter).zfill(digits)
                task_counter += 1
                task_id2 = task_id1
                if not single_end:
                    task_id2 = str(task_counter).zfill(digits)
                    task_counter += 1
                
                # 1. Create the picard working dir for storing the initial FASTQ files
                picard_wd = outdir + os.sep + "picard"
                if os.path.isdir(picard_wd):
                    print "WARNING: The directory '" + picard_wd + "' already exists!"
                else:
                    os.mkdir(picard_wd)
                
                # 2. Symlink the BAM
                #linked_bam = picard_wd + os.sep + task_id1 + '.' + bam_basename
                #linked_bam_index = linked_bam + '.bai'
                #if os.path.exists(linked_bam):
                #    print "WARNING: The file '" + linked_bam + "' already exists!"
                #else:
                #    os.symlink(p, linked_bam)
                #    
                #files_generated.append(linked_bam)
                #
                # 3. Symlink the BAM index, create one if it does not exist
                #original_bam_index = p + '.bai'
                #samtofastq_job_content = ''
                #if os.path.isfile(original_bam_index):
                #    if os.path.exists(linked_bam_index):
                #        print "WARNING: The file '" + linked_bam_index + "' already exists!"
                #    else:
                #        os.symlink(original_bam_index, linked_bam_index)
                #else:
                #    samtofastq_job_content = "samtools index " + linked_bam + ";\n"
                #
                #files_generated.append(linked_bam_index)
                
                # 4. Sort the BAM file by read name
                read_sorted_bam_prefix = bam_basename + '.readsorted'
                read_sorted_bam = picard_wd + os.sep + task_id1 + '.' + read_sorted_bam_prefix + ".bam"
                if local:
                    samtofastq_job_content = samtofastq_job_content + "samtools sort -n -m " + str(sortmem) + " " + p + " " + picard_wd + os.sep + task_id1 + '.' + read_sorted_bam_prefix + ";\n"
                else:
                    samtofastq_job_content = samtofastq_job_content + "samtools sort -n -m " + str(sortmem) + " " + p + " -o $TMPDIR  > " + read_sorted_bam + ";\n"
                files_generated.append(read_sorted_bam)
                
                # 5. Use Picard's SamToFastq.jar to get 2 FASTQ files from the BAM file
                fastq1 = picard_wd + os.sep + task_id1 + '.' + bam_basename + '_1.fq'
                fastq2 = ""
                files_generated.append(fastq1)
                if not single_end:
                    fastq2 = picard_wd + os.sep + task_id2 + '.' + bam_basename + '_2.fq'
                    files_generated.append(fastq2)
                                
                samtofastq_args = cfg.get_value(settings, 'commands', samtofastq_config_key)#config_reader.getCommand(samtofastq_config_key, config_type, project)
                samtofastq_args = samtofastq_args.replace("BAM", read_sorted_bam)
                samtofastq_args = samtofastq_args.replace("FASTQ1", fastq1)
                samtofastq_args = samtofastq_args.replace("FASTQ2", fastq2)                
                #TODO: must define -Xmx, ie. -Xmx7g
                javacmd = cfg.get_value(settings, 'java', samtofastq_config_key)
                samtofastq_job_content = samtofastq_job_content + javacmd + " -jar $PICARD_DIR/SamToFastq.jar " + samtofastq_args + ";\n"
                samtofastq_job_name = prefix + "-samtofastq." + task_id1
                if not single_end:
                    samtofastq_job_name = samtofastq_job_name + '-' + task_id2
                samtofastq_job = MyJob(content=samtofastq_job_content, id=None, name=samtofastq_job_name, predecessors=None, config_key=samtofastq_config_key)
                jobs.append(samtofastq_job)
                 
                # 6. Split each FASTQ into 4*num_reads_per_fastq lines per file, then rename each piece to have a "fq" postfix
                lines_per_file = 4 * long(num_reads_per_fastq)
                
                split_prefix1 = outdir + os.sep + task_id1 + '.' + bam_basename + '_1.fq.'                
                split_fastq_1_job_content = "split -d -a 4 -l " + str(lines_per_file) + " " + fastq1 + " " + split_prefix1 + ';\n'
                split_fastq_1_job_content = split_fastq_1_job_content + "splittedfqs=`ls " + split_prefix1 + "????`; for fq in $splittedfqs;do mv $fq $fq.fq;done;\n"
                split_fastq_1_job_name = prefix + "-getfq." + task_id1
                split_fastq_1_job = MyJob(content=split_fastq_1_job_content, id=None, name=split_fastq_1_job_name, predecessors=[samtofastq_job], config_key=convertreads_config_key)
                jobs.append(split_fastq_1_job)
                files_generated.append(split_prefix1 + '????.fq')
                
                if not single_end:
                    split_prefix2 = outdir + os.sep + task_id2 + '.' + bam_basename + '_2.fq.'
                    split_fastq_2_job_content = "split -d -a 4 -l " + str(lines_per_file) + " " + fastq2 + " " + split_prefix2 + ';\n'
                    split_fastq_2_job_content = split_fastq_2_job_content + "splittedfqs=`ls " + split_prefix2 + "????`; for fq in $splittedfqs;do mv $fq $fq.fq;done;\n"
                    split_fastq_2_job_name = prefix + "-getfq." + task_id2
                    split_fastq_2_job = MyJob(content=split_fastq_2_job_content, id=None, name=split_fastq_2_job_name, predecessors=[samtofastq_job], config_key=convertreads_config_key)
                    jobs.append(split_fastq_2_job)
                    files_generated.append(split_prefix2 + '????.fq')
            else:
                # TASK: Not splitting the reads in the BAM file ...
                task_id1 = str(task_counter).zfill(digits)
                task_counter += 1
                
                # 1. Symlink the BAM
                linked_bam = outdir + os.sep + task_id1 + '.' + bam_basename
                linked_bam_index = linked_bam + '.bai'
                if os.path.exists(linked_bam):
                    print "WARNING: The file '" + linked_bam + "' already exists!"
                else:
                    os.symlink(p, linked_bam)
                files_generated.append(linked_bam)
                
                # 2. Symlink the BAM index, create one if it does not exist
                original_bam_index = p + '.bai'
                if os.path.isfile(original_bam_index):
                    if os.path.exists(linked_bam_index):
                        print "WARNING: The file '" + linked_bam_index + "' already exists!"
                    else:
                        os.symlink(original_bam_index, linked_bam_index)
                else:
                    index_job_content = "samtools index " + linked_bam + ";\n"
                    index_job_name = prefix + "-indexbam." + task_id1
                    index_job = MyJob(content=index_job_content, id=None, name=index_job_name, predecessors=None, config_key=convertreads_config_key)
                    jobs.append(index_job)
                files_generated.append(linked_bam_index)
        elif format == "fastq" or format == "fq":
            task_id = str(task_counter).zfill(digits)
            task_counter += 1
            fq_basename = os.path.basename(p)
            if split:
                # TASK: Split the FASTQ files ...
                
                # Split each FASTQ into 4*num_reads_per_fastq lines per file, then rename each piece to have a "fq" postfix
                lines_per_file = 4 * long(num_reads_per_fastq)
                
                split_fastq_job_content = ''
                
                split_prefix = outdir + os.sep + task_id + '.' + fq_basename + '.'
                if compression != None and compression == '.gz':
                    split_fastq_job_content = "gunzip -c " + p + "| split -d -a 4 -l " + str(lines_per_file) + " - " + split_prefix + ';\n'     
                else:
                    split_fastq_job_content = "split -d -a 4 -l " + str(lines_per_file) + " " + p + " " + split_prefix + ';\n'
                split_fastq_job_content = split_fastq_job_content + "splittedfqs=`ls " + split_prefix + "????`; for fq in $splittedfqs;do mv $fq $fq.fq;done;\n"
                
                split_fastq_job_name = prefix + "-getfq." + task_id
                split_fastq_job = MyJob(content=split_fastq_job_content, id=None, name=split_fastq_job_name, predecessors=None, config_key=convertreads_config_key)
                jobs.append(split_fastq_job)
                files_generated.append(split_prefix + '????.fq')
            else:
                # TASK: Not splitting the FASTQ files ...
                
                # Symlink the FASTQ file
                linked_fq = outdir + os.sep + task_id + '.' + fq_basename
                if os.path.exists(linked_fq):
                    print "WARNING: The file '" + linked_fq + "' already exists!"
                else:
                    os.symlink(p, linked_fq)                
                files_generated.append(linked_fq)
        elif format == "txt":
            task_id = str(task_counter).zfill(digits)
            task_counter += 1
            qseq_basename = os.path.basename(p)
            
            if split:                
                split_fastq_job_content = ''
                
                if converter == "qseq2fastq.py":
                    if compression != None:
                        if compression == "bz2":
                            split_fastq_job_content = "bzip2 -dck " + p + " | python " + qseq2fastq_script + " -d " + outdir + " -g -l 1 -p " + task_id + "." + qseq_basename + " -r " + num_reads_per_fastq + ";\n"
                        elif compression == "gz":
                            split_fastq_job_content = "gunzip -c " + p + " | python " + qseq2fastq_script + " -d " + outdir + " -g -l 1 -p " + task_id + "." + qseq_basename + " -r " + num_reads_per_fastq + ";\n"
                        else:
                            sys.stderr.write("ERROR: Unknown compression format '" + compression + "'!\n")
                            sys.exit(1)                        
                    else:
                        split_fastq_job_content = "python " + qseq2fastq_script + " -i " + p + " -d " + outdir + " -g -l 1 -p " + task_id + '.' + qseq_basename + " -r " + num_reads_per_fastq + ";\n"
                    files_generated.append(outdir + os.sep + task_id + "." + qseq_basename + "*")
                elif converter == "abyss-tofastq":
                    lines_per_file = 4 * long(num_reads_per_fastq)
                    
                    split_prefix = outdir + os.sep + task_id + "." + qseq_basename + '.'
                    if compression == None or compression == "bz2" or compression == "gz":
                        split_fastq_job_content = "abyss-tofastq " + p + " | split -d -a 4 -l " + str(lines_per_file) + " - " + split_prefix + ';\n'
                        split_fastq_job_content = split_fastq_job_content + "splittedfqs=`ls " + split_prefix + "????`; for fq in $splittedfqs;do mv $fq $fq.fq;done;\n"
                    else:
                        sys.stderr.write("ERROR: Unknown compression format '" + compression + "'!\n")
                        sys.exit(1)
                    files_generated.append(split_prefix + ".????.fq")
                else:
                    sys.stderr.write("ERROR: Unknown 'qseq-to-fastq' converter '" + converter + "'!\n")
                    sys.exit(1)
                
                split_fastq_job_name = prefix + "-getfq." + task_id
                split_fastq_job = MyJob(content=split_fastq_job_content, id=None, name=split_fastq_job_name, predecessors=None, config_key=convertreads_config_key)
                jobs.append(split_fastq_job)
            else:
                split_fastq_job_content = ''
                if converter == "qseq2fastq.py":
                    if compression != None:
                        if compression == "bz2":
                            convert_fastq_job_content = "bzip2 -dck " + p + " | python " + qseq2fastq_script + " -d " + outdir + " -g -l 1 -p " + task_id + "." + qseq_basename + ";\n"
                        elif compression == "gz":
                            convert_fastq_job_content = "gunzip -c " + p + " | python " + qseq2fastq_script + " -d " + outdir + " -g -l 1 -p " + task_id + "." + qseq_basename + ";\n"
                        else:
                            sys.stderr.write("ERROR: Unknown compression format '" + compression + "'!\n")
                            sys.exit(1)
                    else:
                        convert_fastq_job_content = "python " + qseq2fastq_script + " -i " + p + " -d " + outdir + " -g -l 1 -p " + task_id + '.' + qseq_basename + ";\n"
                    files_generated.append(outdir + os.sep + task_id + "." + qseq_basename + "*")                        
                elif converter == "abyss-tofastq":
                    fastq_file = outdir + os.sep + task_id + '.' + qseq_basename + ".fq.gz"
                    convert_fastq_job_content = "abyss-tofastq " + p + " | gzip >" + fastq_file + ";\n"
                    files_generated.append(fastq_file)
                else:
                    sys.stderr.write("ERROR: Unknown 'qseq-to-fastq' converter '" + converter + "'!\n")
                    sys.exit(1)
                
                convert_fastq_job_name = prefix + "-getfq." + task_id
                convert_fastq_job = MyJob(content=convert_fastq_job_content, id=None, name=convert_fastq_job_name, predecessors=None, config_key=convertreads_config_key)
                jobs.append(convert_fastq_job)
        else:
            sys.stderr.write("ERROR: Unknown format '" + format + "'!\n")
            sys.exit(1)
    
    # TASK: list (*.fastq/fq(.gz), *.bam) files, creating OUTPUT_READS_LIST
    listreads_job_name = prefix + "-listreads"
    cleanup_job_name = prefix + "-cleanup-reads"
    
    listreads_job_content = ""
    cleanup_job_content = ""
    # check the existence of output files
    for outputread in files_generated:
        listreads_job_content = listreads_job_content + "ls " + outputread + ";\n"
        cleanup_job_content = cleanup_job_content + "rm " + outputread + ";\n"
    
    listreads_job_content = listreads_job_content + "python $TRANSABYSS_PATH/utilities/listreads.py " + outdir + " " + outreadslist
    # every single job submitted earlier would be a predecessor
    listreads_job = MyJob(content=listreads_job_content, id=None, name=listreads_job_name, predecessors=list(jobs), config_key=None)
    jobs.append(listreads_job)
    
    cleanup_job = MyJob(content=cleanup_job_content, id=None, name=cleanup_job_name, config_key=None)
    
    return [jobs, cleanup_job]


def __main__():
    sys.stdout.flush() #flush print outputs to screen

    usage = "usage: %prog INPUT_READS_LIST OUTPUT_READS_LIST [options]"
    version = "1.4.4 (Build 201208141125)"
    
    parser = OptionParser(usage=usage, version=version)
    mandatory_group = OptionGroup(parser, "Mandatory Arguments")
    mandatory_group.add_option('-G', '--genome', action='store_true', default=False, dest='genome', help='use configurations from \'genome.cfg\'')
    mandatory_group.add_option('-T', '--transcriptome', action='store_true', default=False, dest='transcriptome', help='use configurations from \'transcriptome.cfg\'')
    parser.add_option_group(mandatory_group)    
    parser.add_option('-d', '--debug', dest='debug', help='run in debug mode; no jobs will be submitted', action='store_true', default=False)
    parser.add_option('-l', '--local', action='store_true', default=False, dest='local', help='run locally')
    parser.add_option('-j', '--jobdir', dest='jobdir', help='job scripts directory', metavar='DIR')
    parser.add_option('-m', '--sort-mem', type="int", default=800000000, dest='sortmem', help='maximum memory for `samtools sort` [default: %default]', metavar='INT')    
    parser.add_option('-o', '--outputdir', dest='outdir', help='output directory', metavar='DIR')
    parser.add_option('-p', '--prefix', dest='prefix', help='library name', metavar='STR')
    parser.add_option('-P', '--project', dest='project', help='project name', metavar='STR')
    parser.add_option('-r', '--relaxed', dest='lenient', help='look for (un)compressed equivalent of non-existing paths for read files', action='store_true', default=False)
    parser.add_option('-s', '--single_end', action='store_true', default=False, dest='single_end', help='input reads files are single-end')    

    (options, args) = parser.parse_args()
    
    readslist = args[0]
    outreadslist = args[1]
    if os.path.exists(outreadslist):
        print "WARNING: The file '" + outreadslist + "' already exists!"    
    
    
    # Check if the reads list text file (*.in) exists
    if not os.path.isfile(readslist):
        sys.stderr.write("ERROR: Cannot find reads list file, " + readslist  + '\n')
        sys.exit(1)
    
    reads_paths = []
    missing_reads = []
    incorrect_format = []
 
    lenient = options.lenient
    
    # Check if the input reads files ( *.bam, *.fq/fastq(.gz), *.txt(.bz2/gz) ) exist
    fh = open(readslist, 'r')
    for line in fh:
        linestripped = line.strip()
        if len(linestripped) > 0:
            (format, compression) = properformat(linestripped)
            if format is None:
                incorrect_format.append(linestripped)
            else:
                if os.path.isfile(linestripped):
                    reads_paths.append(linestripped)
                elif lenient:
                    equivalents = getequivalents(linestripped, (format, compression))                
                    for e in equivalents:
                        if os.path.isfile(e):
                            print "WARNING: Cannot find file '" + linestripped + "'"
                            print "... will use its equivalent '" + e + "'"
                            reads_paths.append(e)
                        else:
                            missing_reads.append(linestripped)
                else:
                    missing_reads.append(linestripped)    
    fh.close()
    
    erronous_reads_list = False
    
    if len(missing_reads) > 0:
        erronous_reads_list = True
        sys.stderr.write("ERROR: Cannot find the following read files:\n")
        for f in missing_reads:
            sys.stderr.write(f + '\n')
    
    if len(incorrect_format) > 0:
        erronous_reads_list = True    
        sys.stderr.write("ERROR: The following read files have incorrect format:\n")
        for f in incorrect_format:
            sys.stderr.write(f + '\n')
    
    if erronous_reads_list:
        sys.exit(1)

    outdir = options.outdir
    if outdir:
        if not os.path.isdir(outdir):
            os.makedirs(outdir)
            if not os.path.isdir(outdir):            
                sys.stderr.write("ERROR: Cannot create output directory " + outdir  + '\n')
                sys.exit(1)
    else:
        print "No output directory was specified..."
        print "Output files will be generated in the current directory..."
        outdir = os.getcwd()
        print outdir

    jobdir = options.jobdir
    if jobdir:
        if not os.path.isdir(jobdir):
            os.makedirs(jobdir)
            if not os.path.isdir(jobdir):            
                sys.stderr.write("ERROR: Cannot create cluster job working directory " + jobdir  + '\n')
                sys.exit(1)
    else:
        print "No job scripts directory was specified..."
        print "Job scripts and cluster logs will be generated in the current directory..."
        jobdir = os.getcwd()
        print jobdir

    prefix = options.prefix
    if options.prefix == None:
        prefix = os.path.basename(readslist).rstrip(".in")

    # ALL input reads files are confirmed to be found and are in the right format at this point
    # Run the commands locally or submit them as job scripts to the cluster

    try:
        sortmem = int(options.sortmem)
    except TypeError:
        sys.stderr.write("ERROR: `samtools sort` max memory specified (" + options.sortmem + ") is not an integer.")
        sys.exit(1)
        
    if sortmem < 1:
        sys.stderr.write("ERROR: `samtools sort` max memory specified (" + options.sortmem + ") is less than one.")
        sys.exit(1)

    if options.genome and options.transcriptome:
        sys.stderr.write("ERROR: Both genome and transcriptome were selected.")
        sys.exit(1)
    elif not options.genome and not options.transcriptome:
        sys.stderr.write("ERROR: Neither genome or transcriptome were selected. Please specify either '-T' or '-G'")
        sys.exit(1)

    config_type = None    
    if options.genome:
        config_type = "genome"
    elif options.transcriptome:
        config_type = "transcriptome"

    if options.debug:
        print "WARNING: Running in debug mode!"
        print "*** Only job scripts and/or symlinks will be created. ***"
        print "*** No resource-intensive commands will be run.       ***"
        print "*** No jobs will be submitted to the cluster.         ***"

    cfg_file = '%s/configs/%s.cfg' % (TRANSABYSS_PATH, config_type)
    settings = cfg.initialize_settings(cfg_file, options.project)

    # Generate the job commands
    [jobs, cleanup_job] = generate_commands(paths=reads_paths, project=options.project, outdir=outdir, prefix=prefix, single_end=options.single_end, outreadslist=outreadslist, sortmem=sortmem, local=options.local, settings=settings)

    # Write the cleanup script, but it will NOT be run.
    job_script_configs = job_runner.getJobScriptConfigs(TRANSABYSS_PATH + "/configs/job_script.cfg")
    template = TRANSABYSS_PATH + "/configs/templates/" + job_script_configs["local"]
    cleanup_script = job_runner.makeJobScript(template=template, content=cleanup_job.content, jobname=cleanup_job.name, workingdir=jobdir, setup=ta_setup)
    fh = open(jobdir + os.sep + cleanup_job.name + ".sh", 'w')
    fh.write(cleanup_script)
    fh.close()

    if options.local:
        job_runner.run_locally(jobs=jobs, prefix=prefix, jobdir=jobdir, debug=options.debug)
    else:
        job_runner.submit_jobs(jobs=jobs, jobdir=jobdir, debug=options.debug, settings=settings, head_node=None)

    sys.exit(0)

if __name__ == "__main__":
    __main__()
