#!/usr/bin/env python

# written by:    Ka Ming Nip
# updated on:    September 24, 2012
# compatibility: Trans-ABySS 1.4.4

import os, sys, string, re, subprocess
from utilities import cfg

try:
    TRANSABYSS_PATH = os.environ['TRANSABYSS_PATH']
except KeyError:
    sys.stderr.write("ERROR: Cannot find TRANSABYSS_PATH in env!\n")
    sys.exit(1)

class MyJob:    
    def __init__(self, content=None, id=None, name=None, predecessors=None, config_key=None, threads=1, first_task_id=None, last_task_id=None, queue=None):
        self.name = name
        self.id = id
        self.content = content
        self.predecessors = predecessors
        self.config_key = config_key
        self.threads = threads
        self.first_task_id = first_task_id
        self.last_task_id = last_task_id
        self.queue = queue


def getJobScriptConfigs(config_file):
    configs = {}
    fh = open(config_file, 'r')
    for line in fh:
        line_stripped = line.strip()
        if len(line_stripped) > 0:
            if ':' in line_stripped:
                items = line_stripped.split(':', 1)
                configs[items[0].strip()] = items[1].strip()
    fh.close()
    
    #some sanity checks
    for key in ["local", "cluster_basic", "cluster_parallel", "cluster_basic_array", "cluster_parallel_array", "predecessors_list_delimiter", "submit_cluster_job_return_string", "submit_cluster_array_job_return_string", "submit_cluster_job_command", "run_local_job_command"]:
        if not configs.has_key(key) or not configs[key]:
            sys.stderr.write("ERROR: \"%s\" is not defined in %s" % (key, config_file))
            sys.exit(1)
    
    if not "${JOBID}" in configs["submit_cluster_job_return_string"]:
        sys.stderr.write("ERROR: Cannot find \"${JOBID}\" in the \"submit_cluster_job_return_string\" defined in " + config_file) 
        sys.exit(1)
    
    if not "${JOBID}" in configs["submit_cluster_array_job_return_string"]:
        sys.stderr.write("ERROR: Cannot find \"${JOBID}\" in the \"submit_cluster_array_job_return_string\" defined in " + config_file) 
        sys.exit(1)
    
    return configs


def makeJobScript(template, content, jobname=None, workingdir=None, threads=1, predecessors=None, memory=None, setup=None, first_task_id=None, last_task_id=None, tmpmem=None, queue=None):
    script = ""
    if not os.path.isfile(template):
        sys.stderr.write("ERROR: Cannot find template file, " + template)
        sys.exit(1)
    fh = open(template, 'r')
    for line in fh:
        #Replace known variable names in template with the appropriate values        
        if "${JOB_NAME}" in line:
            if jobname:
                line = line.replace("${JOB_NAME}", str(jobname))
            else:                
                continue
        if "${WORKING_DIR}" in line:
            if workingdir:
                line = line.replace("${WORKING_DIR}", str(workingdir))
            else:
                continue
        if "${THREADS}" in line:
            if threads and threads > 1:
                line = line.replace("${THREADS}", str(threads))
            else:
                continue
        if "${PREDECESSORS}" in line:
            if predecessors:
                line = line.replace("${PREDECESSORS}", str(predecessors))
            else:
                continue
        if "${MEM}" in line:
            if memory:
                line = line.replace("${MEM}", str(memory))
            else:
                continue
        if "${SETUP_PATHS}" in line:
            if setup:
                line = line.replace("${SETUP_PATHS}", str(setup))
            else:
                continue
        if "${FIRST_TASK_ID}" in line:
            if first_task_id:
                line = line.replace("${FIRST_TASK_ID}", str(first_task_id))
            else:
                continue
        if "${LAST_TASK_ID}" in line:
            if last_task_id:
                line = line.replace("${LAST_TASK_ID}", str(last_task_id))
            else:
                continue
        if "${QUEUE}" in line:
            if queue:
                line = line.replace("${QUEUE}", str(queue))
            else:
                continue
        if "${CONTENT}" in line:
            if content:
                line = line.replace("${CONTENT}", str(content))
            else:
                continue
        if "${TMPMEM}" in line:
            if tmpmem:
                line = line.replace("${TMPMEM}", str(tmpmem))
            else:
                continue
        script += line
    fh.close()
    return script


def run_locally(jobs, prefix, jobdir=None, debug=False):
    job_script_config_file = '%s/configs/job_script.cfg' % TRANSABYSS_PATH
    job_script_template_dir = '%s/configs/templates' % TRANSABYSS_PATH
    job_script_configs = getJobScriptConfigs(job_script_config_file)
    
    print "Combining jobs for local run ..."
    template = '%s/%s' % (job_script_template_dir, job_script_configs["local"])
    
    content = ""
    for j in jobs:
        content += "\n"
        if j.first_task_id and j.last_task_id:
            content += "for TA_JOBID in {%s..%s};do " % (j.first_task_id, j.last_task_id) + "\n"
            content += j.content
            content += "\ndone\n"
        else:
            content += j.content
            
    script = makeJobScript(template=template, setup=cfg.get_setup_cmd(), content=content)
    
    name = "local"
    if prefix != None:
        name = prefix + ".local"
    
    scriptpath = '%s/%s.sh' % (jobdir, name)
    print "script: %s" % scriptpath
    
    # write script to file
    fh = open(scriptpath, 'w')
    fh.write(script)
    fh.close()
    
    cmd = [job_script_configs["run_local_job_command"], scriptpath]
    
    # run script locally
    if debug:
        print " ".join(cmd)
    else:
        print "Run script locally ..."
        exit_status = subprocess.call(cmd)
        if exit_status != 0:
            sys.stderr.write("ERROR: Execution of script ended with a non-zero exit-status.")
            sys.exit(1)
        print "Finished running script."


def submit_jobs(jobs, settings, head_node=None, jobdir=None, debug=False, queue=None):
    job_script_config_file = '%s/configs/job_script.cfg' % TRANSABYSS_PATH
    job_script_template_dir = '%s/configs/templates' % TRANSABYSS_PATH
    job_script_configs = getJobScriptConfigs(job_script_config_file)

    predecessors_list_delimiter = str(job_script_configs["predecessors_list_delimiter"])
    
    submit_cluster_return_string = job_script_configs["submit_cluster_job_return_string"]
    cluster_basic_template_file = job_script_template_dir + os.sep + job_script_configs["cluster_basic"]
    cluster_parallel_template_file = job_script_template_dir + os.sep + job_script_configs["cluster_parallel"]

    submit_cluster_return_string_array = job_script_configs["submit_cluster_array_job_return_string"]
    cluster_basic_template_file_array = job_script_template_dir + os.sep + job_script_configs["cluster_basic_array"]
    cluster_parallel_template_file_array = job_script_template_dir + os.sep + job_script_configs["cluster_parallel_array"]

    submit_cluster_job_command = job_script_configs["submit_cluster_job_command"]
    
    prog = re.compile("\s*" + submit_cluster_return_string.replace("${JOBID}", "(?P<jid>\S+)") + "\s*")
    prog_array = re.compile("\s*" + submit_cluster_return_string_array.replace("${JOBID}", "(?P<jid>\w+)") + "\s*")

    for job in jobs:
        predecessors_string = None
        if job.predecessors and len(job.predecessors) > 0:
            predecessor_ids = []
            for j in job.predecessors:
                if j.id:
                    predecessor_ids.append(str(j.id))
            predecessors_string = predecessors_list_delimiter.join(predecessor_ids)
                            
        memory = None
        tmpmem = None
        threads = job.threads
        if job.config_key:
            memory = cfg.get_value(settings, 'memory', job.config_key) #getMemory(job.config_key, config_type, project)
            tmpmem = cfg.get_value(settings, 'tmpmem', job.config_key)
            if memory:
                items = memory.split(',')
                num_items = len(items)
                # memory
                if num_items > 0:
                    memory_str_stripped = items[0].strip()
                    if len(memory_str_stripped) > 0:
                        memory = memory_str_stripped
                # threads
                if num_items > 1:
                    threads_str_stripped = items[1].strip()
                    if len(threads_str_stripped) > 0:
                        threads = int(threads_str_stripped)
        
        template = cluster_basic_template_file
        matcher = prog
        if threads > 1:
            template = cluster_parallel_template_file
        if job.first_task_id and job.last_task_id:
            matcher = prog_array
            template = cluster_basic_template_file_array
            if threads > 1:
                template = cluster_parallel_template_file_array
        
        #template, content, jobname=None, workingdir=None, threads=1, predecessors=None, memory=None, setup=None
        script = makeJobScript(template=template, content=job.content, jobname=job.name, workingdir=jobdir, threads=threads, predecessors=predecessors_string, first_task_id=job.first_task_id, last_task_id=job.last_task_id, memory=memory, setup=cfg.get_setup_cmd(), queue=queue, tmpmem=tmpmem)
        scriptpath = jobdir + os.sep + job.name + ".sh"
        print "job script: " + scriptpath
        fh = open(scriptpath, 'w')
        fh.write(script)
        fh.close()
        
        cmd = []
        if head_node:
            #ssh to head node then submit the job with qsub
            cmd = ["ssh", head_node, submit_cluster_job_command, scriptpath]
        else:
            #submit the job with qsub
            cmd = [submit_cluster_job_command, scriptpath]        
        print " ".join(cmd)
        
        if not debug:
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            stdout, stderr = p.communicate()
            
            jobid = None
            
            lines = stdout.split("\n")            
            for line in lines:
                print line
                m = matcher.match(line)
                if m:
                    jobid = m.group('jid')
            
            if not jobid:
                sys.stderr.write("ERROR: Cannot identify job id.\n")
                sys.exit(1) 
            
            job.id = jobid

