#!/usr/bin/env python

# written by:    Ka Ming Nip
# updated on:    September 26, 2012
# compatibility: Trans-ABySS 1.4.4

import os, sys, string, re, subprocess
from optparse import OptionParser,OptionGroup
from utilities import cfg, job_runner
from utilities.job_runner import MyJob

try:
    TRANSABYSS_PATH = os.environ['TRANSABYSS_PATH']
except KeyError:
    sys.stderr.write("ERROR: Cannot find TRANSABYSS_PATH in env!\n")
    sys.exit(1)

ta_setup = "source " + TRANSABYSS_PATH + "/setup"
#input_files_per_abyssmap_job = 20

class R2C_Factory:
    def __init__(self, contigs, reads_paths, prefix=None, outdir=None, jobdir=None, debug=False, cleanup=False, transcriptome=False, genome=False, bwa=False, abyssmap=False, bowtie2=False, threads=int(4), single_end=False, local=False, xa2multi=False, strip=False, messy=False, sortmem=int(800000000), k=int(24), project=None, cluster=None, queue=None):
        self.contigs = contigs
        self.reads_paths = reads_paths
        self.prefix = prefix
        self.outdir = outdir
        self.jobdir = jobdir
        self.debug = debug
        self.cleanup = cleanup
        self.transcriptome = transcriptome
        self.genome = genome
        self.bwa = bwa
        self.abyssmap = abyssmap
        self.bowtie2 = bowtie2
        self.threads = threads
        self.single_end = single_end
        self.local = local
        self.xa2multi = xa2multi
        self.strip = strip
        self.messy = messy
        self.sortmem = sortmem
        self.k = k
        self.project = project
        self.cluster = cluster
        self.queue = queue
        self.setup_params()
        
    def setup_params(self):
        if not os.path.isfile(self.contigs):
            sys.stderr.write("ERROR: Cannot find contigs FASTA file, " + self.contigs + '\n')
            sys.exit(1)
        
        for r in self.reads_paths:
            if not os.path.isfile(r):
                sys.stderr.write("ERROR: Cannot find reads file, " + r + '\n')
                sys.exit(1)
            

        contigs_base_name = os.path.basename(self.contigs)
        if not self.prefix:        
            if contigs_base_name.endswith(".fa"):            
                if contigs_base_name.endswith("-contigs.fa"):
                    index = contigs_base_name.rindex("-contigs.fa")
                else:
                    index = contigs_base_name.rindex(".fa")
                self.prefix = contigs_base_name[0:index]
            elif contigs_base_name.endswith(".fasta"):
                if contigs_base_name.endswith("-contigs.fasta"):
                    index = contigs_base_name.rindex("-contigs.fasta")
                else:
                    index = contigs_base_name.rindex(".fasta")
                self.prefix = contigs_base_name[0:index]
            else:
                sys.stderr.write("ERROR: Cannot determine prefix. Please specify a prefix." + '\n')
                sys.exit(1)
        
        if self.outdir:
            if not os.path.isdir(self.outdir):
                os.makedirs(self.outdir)
                if not os.path.isdir(self.outdir):
                    sys.stderr.write("ERROR: Cannot create output directory " + self.outdir  + '\n')
                    sys.exit(1)
        else:
            print "No output directory was specified..."
            print "Output files will be generated in the current directory..."    
            self.outdir = os.getcwd()
            print self.outdir

        if self.jobdir:
            if not os.path.isdir(self.jobdir):
                os.makedirs(self.jobdir)
                if not os.path.isdir(self.jobdir):
                    sys.stderr.write("ERROR: Cannot create cluster job working directory " + self.jobdir  + '\n')
                    sys.exit(1)
        else:
            print "No job scripts directory was specified..."
            print "Job scripts and cluster logs will be generated in the current directory..."
            self.jobdir = os.getcwd()
            print self.jobdir

        link_contigs = self.outdir + "/" +  contigs_base_name;
        if not os.path.isfile(link_contigs):
            os.symlink(os.path.abspath(self.contigs), link_contigs)
        else:
            print "Warning: Contigs file already exists in output directory, " + link_contigs
        
        self.contigs = link_contigs

        try:
            self.threads = int(self.threads)
        except TypeError:
            sys.stderr.write("ERROR: Number of threads specified (" + str(self.threads) + ") is not an integer." + '\n')
            sys.exit(1)
        
        if self.threads < 1:
            sys.stderr.write("ERROR: Number of threads specified (" + str(self.threads) + ") is less than one." + '\n')
            sys.exit(1)
            
        try:
            self.sortmem = int(self.sortmem)
        except TypeError:
            sys.stderr.write("ERROR: `samtools sort` max memory specified (" + str(self.sortmem) + ") is not an integer." + '\n')
            sys.exit(1)
            
        if self.sortmem < 1:
            sys.stderr.write("ERROR: `samtools sort` max memory specified (" + str(self.sortmem) + ") is less than one." + '\n')
            sys.exit(1)            
        
        aligners = 0
        if self.bwa:
            aligners += 1
        if self.bowtie2:
            aligners += 1
        if self.abyssmap:
            aligners += 1
        
        if aligners > 1:
            sys.stderr.write("ERROR: More than one aligner were selected. Only one aligner is allowed." + '\n')
            sys.exit(1)
        elif aligners <= 0:
            sys.stderr.write("ERROR: No aligner was selected." + '\n')
            sys.exit(1)    

        if self.genome and self.transcriptome:
            sys.stderr.write("ERROR: Both genome and transcriptome were selected." + '\n')
            sys.exit(1)
        elif not self.genome and not self.transcriptome:
            sys.stderr.write("ERROR: Neither genome or transcriptome were selected. Please specify either '-T' or '-G'" + '\n')
            sys.exit(1)
        
    def create_and_submit_jobs(self):
        config_type = None    
        if self.genome:
            config_type = "genome"
        elif self.transcriptome:
            config_type = "transcriptome"

        if self.debug:
            print "WARNING: Running in debug mode!"
            print "*** Only job scripts and/or symlinks will be created. ***"
            print "*** No resource-intensive commands will be run.       ***"
            print "*** No jobs will be submitted to the cluster.         ***"

        jobs = None
        cleanupjob = None
    
        if self.bwa:
            (jobs, cleanupjob) = generate_job_commands_bwa(contigs=self.contigs, reads_paths=self.reads_paths, prefix=self.prefix, single_end=self.single_end, outdir=self.outdir, threads=self.threads, local=self.local, xa2multi=self.xa2multi, strip=self.strip, messy=self.messy, sortmem=self.sortmem)
        elif self.abyssmap:
            (jobs, cleanupjob) = generate_job_commands_abyssmap(contigs=self.contigs, reads_paths=self.reads_paths, prefix=self.prefix, outdir=self.outdir, threads=self.threads, local=self.local, k=self.k, strip=self.strip, single_end=self.single_end, messy=self.messy, sortmem=self.sortmem)
        elif self.bowtie2:
            (jobs, cleanupjob) = generate_job_commands_bowtie2(contigs=self.contigs, reads_paths=self.reads_paths, prefix=self.prefix, outdir=self.outdir, threads=self.threads, local=self.local, strip=self.strip, single_end=self.single_end, messy=self.messy, sortmem=self.sortmem)
        else:
            print "ERROR: No aligner was specified."
            sys.exit(1)

        if cleanupjob:
            if not self.messy and self.cleanup:
                # cleanup script will be the last task
                jobs.append(cleanupjob)        
            else:
                print "WARNING: Clean-up script will be generated, but it will not be run. Otherwise, use the -c option next time."
                cleanupscript_name = cleanupjob.name
                if not cleanupscript_name:
                    cleanupscript_name = "cleanup-r2c"
                job_script_configs = job_runner.getJobScriptConfigs(TRANSABYSS_PATH + "/configs/job_script.cfg")
                cleanupscript_template = TRANSABYSS_PATH + "/configs/templates/" + job_script_configs["local"]
                
                #template, content, jobname=None, workingdir=None, threads=1, predecessors=None, memory=None, setup=None
                cleanupjobscript_content = job_runner.makeJobScript(template=cleanupscript_template, content=cleanupjob.content, jobname=cleanupscript_name, workingdir=self.outdir, setup=ta_setup)
                script = self.jobdir + os.sep + cleanupscript_name + ".sh"
                fh = open(script, 'w')
                fh.write(cleanupjobscript_content)
                fh.close()
                print "clean-up script: " + script
        else:
            print "No clean-up script will need to be run."
            
        if self.local:
            job_runner.run_locally(jobs=jobs, prefix=self.prefix, jobdir=self.jobdir, debug=self.debug)
        else:
            cfg_file = '%s/configs/%s.cfg' % (TRANSABYSS_PATH, config_type)
            settings = cfg.initialize_settings(cfg_file, self.project)                
            job_runner.submit_jobs(jobs=jobs, jobdir=self.jobdir, debug=self.debug, settings=settings, head_node=self.cluster, queue=self.queue)
    
        return jobs[-1] # last job


def generate_job_commands_bowtie2(contigs, reads_paths, prefix, outdir, threads, local, strip, single_end, messy, sortmem):
    finaloutputprefix = outdir + "/" + prefix + "-contigs"
    intermediates = []
    
    #1. bowtie2-build -f FASTA PREFIX
    cmd = "cd " + outdir + " && "
    cmd = cmd + "bowtie2-build -f " + contigs + " " + finaloutputprefix + ";\n"
    name = prefix + "-bowtie2-build"
    indexing_job = MyJob(content=cmd, id=None, name=name, predecessors=None, config_key="bowtie2-build")
    for indexes_extension in [".1.bt2", ".2.bt2", ".3.bt2", ".4.bt2", ".rev.1.bt2", ".rev.2.bt2"]:
        intermediates.append(finaloutputprefix + indexes_extension)      
        
    #2. bowtie2
    task_id = 1
    align_jobs = []
    sorted_bam_files = []
    
    if single_end:
        for read in reads_paths:
            outputprefix = outdir + '/' + os.path.basename(read)
        
            cmd = "cd " + outdir + " && "
            cmd = cmd + "bowtie2 -a "
            
            if is_fastq(read):
                cmd = cmd + " -q "
            elif is_txt(read):
                cmd = cmd + " --qseq "
            elif is_fasta(read):
                cmd = cmd + " -f "
            else:
                cmd = cmd + " -r "
            
            if threads > 1:
                cmd = cmd + " -p " + str(threads)            
            
            cmd = cmd + " -x " + finaloutputprefix + " -U " + read
            
            if messy: # generate intermediate files whenever possible
                cmd = cmd + " > " + outputprefix + ".sam && "                
                lastoutputfile = outputprefix + ".sam"
                intermediates.append(lastoutputfile)
                
                if strip:
                    cmd = cmd + "cat " + lastoutputfile + " | ta-sam_strip_qual.awk > " + outputprefix + ".qual-stripped.sam && "
                    lastoutputfile = outputprefix + ".qual-stripped.sam"
                    intermediates.append(lastoutputfile)
                    
                cmd = cmd + "samtools view -hSu " + lastoutputfile + " > " + outputprefix + ".unsorted.bam && "
                lastoutputfile = outputprefix + ".unsorted.bam"
                intermediates.append(lastoutputfile)
                
                if local:
                    cmd = cmd + "samtools sort -m " + str(sortmem) + " " + lastoutputfile + " " + outputprefix + " && "
                else:
                    cmd = cmd + "samtools sort -m " + str(sortmem) + " -o " + lastoutputfile + " $TMPDIR > " + outputprefix + ".bam || rm -f $TMPDIR*.bam 2> /dev/null && "
            else: # use pipes whenever possible
                if strip:
                    cmd = cmd + " | ta-sam_strip_qual.awk "
                if local:
                    cmd = cmd + " | samtools view -hSu - | samtools sort -m " + str(sortmem) + " - " + outputprefix + " && "
                else:
                    cmd = cmd + " | samtools view -hSu - | samtools sort -m " + str(sortmem) + " -o - $TMPDIR > " + outputprefix + ".bam || rm -f $TMPDIR*.bam 2> /dev/null && "
                
            outputfile = outputprefix + ".bam"
            sorted_bam_files.append(outputfile)
            cmd = cmd + "samtools index " + outputfile + ";\n"
                
            name = prefix + "-bowtie2-" + str(task_id).zfill(4)
            align_jobs.append(MyJob(content=cmd, id=None, name=name, predecessors=[indexing_job], config_key="bowtie2", threads=threads))         
            task_id += 1   
    else:
        for pair in list(chunks(reads_paths, 2)):           
            cmd = "cd " + outdir + " && "
            cmd = cmd + "bowtie2 -a --fast"
            
            read1 = pair[0]
            read2 = pair[1]
            outputprefix = outdir + '/' + os.path.basename(read1)            
            
            if is_fastq(read1) and is_fastq(read2):
                cmd = cmd + " -q "
            elif is_txt(read1) and is_txt(read2):
                cmd = cmd + " --qseq "
            elif is_fasta(read1) and is_fasta(read2):
                cmd = cmd + " -f "
            else:
                cmd = cmd + " -r "
            
            if threads > 1:
                cmd = cmd + " -p " + str(threads)            
            
            cmd = cmd + " -x " + finaloutputprefix + " -1 " + read1 + " -2 " + read2
            
            if messy: # generate intermediate files whenever possible
                cmd = cmd + " > " + outputprefix + ".sam && "                
                lastoutputfile = outputprefix + ".sam"
                intermediates.append(lastoutputfile)
                
                if strip:
                    cmd = cmd + "cat " + lastoutputfile + " | ta-sam_strip_qual.awk > " + outputprefix + ".qual-stripped.sam && "
                    lastoutputfile = outputprefix + ".qual-stripped.sam"
                    intermediates.append(lastoutputfile)
                    
                cmd = cmd + "samtools view -hSu " + lastoutputfile + " > " + outputprefix + ".unsorted.bam && "
                lastoutputfile = outputprefix + ".unsorted.bam"
                intermediates.append(lastoutputfile)
                
                if local:
                    cmd = cmd + "samtools sort -m " + str(sortmem) + " " + lastoutputfile + " " + outputprefix + " && "
                else:
                    cmd = cmd + "samtools sort -m " + str(sortmem) + " -o " + lastoutputfile + " $TMPDIR > " + outputprefix + ".bam || rm -f $TMPDIR*.bam 2> /dev/null && "
            else: # use pipes whenever possible
                if strip:
                    cmd = cmd + " | ta-sam_strip_qual.awk "
                if local:
                    cmd = cmd + " | samtools view -hSu - | samtools sort -m " + str(sortmem) + " - " + outputprefix + " && "
                else:
                    cmd = cmd + " | samtools view -hSu - | samtools sort -m " + str(sortmem) + " -o - $TMPDIR > " + outputprefix + ".bam || rm -f $TMPDIR*.bam 2> /dev/null && "
                
            outputfile = outputprefix + ".bam"
            sorted_bam_files.append(outputfile)
            cmd = cmd + "samtools index " + outputfile + ";\n"
                
            name = prefix + "-bowtie2-" + str(task_id).zfill(4)
            align_jobs.append(MyJob(content=cmd, id=None, name=name, predecessors=[indexing_job], config_key="bowtie2", threads=threads))         
            task_id += 1 
        
    #3. Symlink or merge the sorted bam files
    cmd = "cd " + outdir + " && "
    num_align_jobs = len(align_jobs)
    merge_job = None
    if num_align_jobs > 1:
        cmd = cmd + "samtools merge -f " + outdir + "/" + prefix + "-contigs.bam " + " ".join(sorted_bam_files) + " && "
        cmd = cmd + "samtools index " + outdir + "/" + prefix + "-contigs.bam;\n"
        name = prefix + "-samtools_merge"
        merge_job = MyJob(content=cmd, name=name, predecessors=align_jobs, config_key="samtools_merge")
        
        # the sorted BAMs are intermediate files because we have the merged BAM
        intermediates.extend(sorted_bam_files)
        # we want to remove the BAM indexes, too
        for bam in sorted_bam_files:
            intermediates.append(bam + ".bai")
    elif num_align_jobs == 1:
        # this sorted BAM is not an intermediate file because it is the final BAM
        cmd = cmd + "ln -sf " + sorted_bam_files[0] + " " + outdir + "/" + prefix + "-contigs.bam && "
        cmd = cmd + "ln -sf " + sorted_bam_files[0] + ".bai " + outdir + "/" + prefix + "-contigs.bam.bai;\n"
        align_jobs[0].content = align_jobs[0].content + cmd
    else:
        sys.stderr.write("ERROR: Nonsensical number of bowtie2 alignment jobs (" + num_bwa_sam_jobs + ")." + '\n')
        sys.exit(1)    

    #4. Clean up
    cmd = "cd " + outdir + ";\n"
    for ftbd in intermediates:
        cmd = cmd + "rm " + ftbd + ";\n"
    name = prefix + "-cleanup-r2c"
    
    cleanup_predecessor = merge_job
    if not cleanup_predecessor:
        cleanup_predecessor = align_jobs[0]
        
    cleanup_job = MyJob(content=cmd, name=name, predecessors=[cleanup_predecessor])

    jobs = [indexing_job]
    jobs.extend(align_jobs)
    if merge_job:
        jobs.append(merge_job)
    
    return (jobs, cleanup_job)


def is_fastq(path):
    return path.endswith(".fastq") or path.endswith(".fastq.gz") or path.endswith(".fastq.bz2") or path.endswith(".fq") or path.endswith(".fq.gz") or path.endswith(".fq.bz2")
    
def is_fasta(path):
    return path.endswith(".fasta") or path.endswith(".fasta.gz") or path.endswith(".fasta.bz2") or path.endswith(".fa") or path.endswith(".fa.gz") or path.endswith(".fa.bz2")
    
def is_txt(path):
    return path.endswith(".txt") or path.endswith(".txt.gz") or path.endswith(".txt.bz2")

def is_bam(path):
    return path.endswith(".bam")


def generate_job_commands_abyssmap(contigs, reads_paths, prefix, outdir, threads, local, k, strip, single_end, messy, sortmem):
    finaloutputprefix = outdir + "/" + prefix + "-contigs"      
    
    jobs=[]
    abyss_map_jobs = []
    merge_job = None
    cleanup_job = None
    intermediates = []
    sorted_bam_files = []

    if single_end:
        task_id = 1
        # one job for each read file
        for read in reads_paths:
            cmd = "cd " + outdir + " && "
            outputprefix = outdir + '/' + str(task_id).zfill(4) + '.' + os.path.basename(read)           
            [align_cmd, align_finaloutput, align_intermediates] = abyssmap_helper(contigs, read, outputprefix, threads, local, k, strip, messy, sortmem)
            cmd = cmd + align_cmd
            
            intermediates.extend(align_intermediates)
            name = prefix + "-abyssmap-" + str(task_id).zfill(4)
            abyss_map_jobs.append(MyJob(content=cmd, id=None, name=name, predecessors=None, config_key="abyss-map", threads=threads))
            sorted_bam_files.append(align_finaloutput)
            task_id += 1
    else:
        lastread = None
        task_id = 1
        # one job for each mate-pair
        for read in reads_paths:
            if is_bam(read):
                # BAM files should have both mates
                if lastread:
                    print "ERROR: Read files are not paired properly."
                    print lastread
                    print read
                    sys.exit(1)
                else:
                    cmd = "cd " + outdir + " && "
                    outputprefix = outdir + '/' + str(task_id).zfill(4) + '.' + os.path.basename(read)           
                    [align_cmd, align_finaloutput, align_intermediates] = abyssmap_helper(contigs, read, outputprefix, threads, local, k, strip, messy, sortmem)
                    cmd = cmd + align_cmd
                    
                    intermediates.extend(align_intermediates)                                        
                    name = prefix + "-abyssmap-" + str(task_id).zfill(4)
                    abyss_map_jobs.append(MyJob(content=cmd, id=None, name=name, predecessors=None, config_key="abyss-map", threads=threads))
                    sorted_bam_files.append(align_finaloutput)
                    task_id += 1
            else:
                if lastread:
                    cmd = "cd " + outdir + " && "
                    outputprefix = outdir + '/' + str(task_id).zfill(4) + '.' + os.path.basename(read)           
                    [align_cmd, align_finaloutput, align_intermediates] = abyssmap_helper(contigs, lastread + " " + read, outputprefix, threads, local, k, strip, messy, sortmem)
                    cmd = cmd + align_cmd
                    
                    intermediates.extend(align_intermediates)                                        
                    name = prefix + "-abyssmap-" + str(task_id).zfill(4)
                    abyss_map_jobs.append(MyJob(content=cmd, id=None, name=name, predecessors=None, config_key="abyss-map", threads=threads))
                    sorted_bam_files.append(align_finaloutput)
                    task_id += 1
                    lastread = None
                else:
                    lastread = read

    #3. Symlink or merge the sorted bam files
    num_align_jobs = len(abyss_map_jobs)
    merge_job = None
    if num_align_jobs > 1:    
        cmd = "cd " + outdir + " && "
        cmd = cmd + "samtools merge -f " + outdir + "/" + prefix + "-contigs.bam " + " ".join(sorted_bam_files) + " && "
        cmd = cmd + "samtools index " + outdir + "/" + prefix + "-contigs.bam;\n"
        name = prefix + "-samtools_merge"
        merge_job = MyJob(content=cmd, name=name, predecessors=abyss_map_jobs, config_key="samtools_merge")
        
        # the sorted BAMs are intermediate files because we have the merged BAM
        intermediates.extend(sorted_bam_files)
        # we want to remove the BAM indexes, too
        for bam in sorted_bam_files:
            intermediates.append(bam + ".bai")
    elif num_align_jobs == 1:
        cmd = "cd " + outdir + " && "
        # this sorted BAM is not an intermediate file because it is the final BAM
        cmd = cmd + "ln -sf " + sorted_bam_files[0] + " " + outdir + "/" + prefix + "-contigs.bam && "
        cmd = cmd + "ln -sf " + sorted_bam_files[0] + ".bai " + outdir + "/" + prefix + "-contigs.bam.bai;\n"
        abyss_map_jobs[0].content = abyss_map_jobs[0].content + cmd
    else:
        sys.stderr.write("ERROR: Nonsensical number of abyss-map alignment jobs (%s)." % str(num_align_jobs) + '\n')
        sys.exit(1)    

    #4. Clean up
    cmd = "cd " + outdir + ";\n"
    for ftbd in intermediates:
        cmd = cmd + "rm " + ftbd + ";\n"
    name = prefix + "-cleanup-r2c"
    
    cleanup_predecessor = merge_job
    if not cleanup_predecessor:
        cleanup_predecessor = abyss_map_jobs[0]
        
    cleanup_job = MyJob(content=cmd, name=name, predecessors=[cleanup_predecessor])

    jobs.extend(abyss_map_jobs)
    if merge_job:
        jobs.append(merge_job)
    
    return (jobs, cleanup_job)


def abyssmap_helper(contigs, reads, outputprefix, threads, local, k, strip, messy, sortmem):
    intermediates = []
    
    cmd = "abyss-map " + reads + " " + contigs + " -k " + str(k)
    if threads > 1:
        cmd = cmd + " -j " + str(threads)
    
    if messy: # generate intermediate files whenever possible
        cmd = cmd + " > " + outputprefix + ".sam && "
        lastoutputfile = outputprefix + ".sam"
        intermediates.append(lastoutputfile)
        
        if strip:
            cmd = cmd + "cat " + lastoutputfile + " | ta-sam_strip_qual.awk > " + outputprefix + ".qual-stripped.sam && "
            lastoutputfile = outputprefix + ".qual-stripped.sam"
            intermediates.append(lastoutputfile)
            
        cmd = cmd + "samtools view -hSu " + lastoutputfile + " > " + outputprefix + ".unsorted.bam && "                        
        lastoutputfile = outputprefix + ".unsorted.bam"
        intermediates.append(lastoutputfile)
        
        if local:
            cmd = cmd + "samtools sort -m " + str(sortmem) + " " + lastoutputfile + " " + outputprefix + " && "
        else:
            cmd = cmd + "samtools sort -m " + str(sortmem) + " -o " + lastoutputfile + " $TMPDIR > " + outputprefix + ".bam || rm -f $TMPDIR*.bam 2> /dev/null && "
    else: # use pipes whenever possible
        if strip:
            cmd = cmd + " | ta-sam_strip_qual.awk "
        if local:
            cmd = cmd + " | samtools view -hSu - | samtools sort -m " + str(sortmem) + " - " + outputprefix + " && "
        else:
            cmd = cmd + " | samtools view -hSu - | samtools sort -m " + str(sortmem) + " -o - $TMPDIR > " + outputprefix + ".bam || rm -f $TMPDIR*.bam 2> /dev/null && "
    
    finaloutput = outputprefix + ".bam"
    cmd = cmd + "samtools index " + finaloutput + ";\n"
    
    return [cmd, finaloutput, intermediates]


def chunks(l, n):
    """ Yield successive n-sized chunks from l.
    """
    for i in xrange(0, len(l), n):
        yield l[i:i+n]    


def generate_job_commands_bwa(contigs, reads_paths, prefix, single_end, outdir, threads, local, xa2multi, strip, messy, sortmem):
    intermediates = []

    #1. index the contigs fasta file
    cmd = "cd " + outdir + " && "
    cmd = cmd + "samtools faidx " + contigs + " && "
    cmd = cmd + "numbases=`awk '{SUM += $2} END { print SUM }' " + contigs + ".fai` && "
    cmd = cmd + "echo \"$numbases bases\";\n"
    cmd = cmd + "if [ $numbases -le 10000000 ]; then bwa index -a is " + contigs + ";else bwa index -a bwtsw " + contigs + ";fi\n"
    name = prefix + "-index"
    indexing_job = MyJob(content=cmd, id=None, name=name, predecessors=None, config_key="bwa_index")

    for indexes_extension in [".fai", ".amb", ".ann", ".bwt", ".pac", ".sa"]:
        intermediates.append(contigs + indexes_extension)  

    #2. align the reads to contigs
    bwa_aln_jobs = []
    bwa_sam_jobs = []
    sai_files = []
    sorted_bam_files = []
    if single_end:
        i = 1
        for readfile in reads_paths:
            #bwa aln ...            
            cmd = "cd " + outdir + " && "
            cmd = cmd + "time bwa aln "
            if is_bam(readfile):
                cmd = cmd + "-b "                
            if threads > 1:
                cmd = cmd + "-t " + str(threads) + " "
            sai = outdir + "/" + os.path.basename(readfile) + ".sai"
            sai_files.append(sai)
            cmd = cmd + " -f " + sai + " "
            cmd = cmd + contigs + " " + readfile + ";\n"
            name = prefix + "-bwa_aln-" + str(i).zfill(4)
            aln_job = MyJob(content=cmd, id=None, name=name, predecessors=[indexing_job], config_key="bwa_aln", threads=threads)
            bwa_aln_jobs.append(aln_job)
            
            #bwa samse ...
            cmd = "cd " + outdir + " && "
            outputprefix = outdir + '/' + os.path.basename(readfile)
            cmd = cmd + "time bwa samse -n50 -r '@RG	ID:1	LB:1	SM:1' " + contigs + " " + sai + " " + readfile
            
            if messy: # generate intermediate files whenever possible
                cmd = cmd + " > " + outputprefix + ".sam && "
                lastoutputfile = outputprefix + ".sam"
                intermediates.append(lastoutputfile)
                
                if xa2multi:
                    cmd = cmd + "cat " + lastoutputfile + " | xa2multi.pl > " + outputprefix + ".xa2multi.sam && "
                    lastoutputfile = outputprefix + ".xa2multi.sam"
                    intermediates.append(lastoutputfile)
                
                if strip:
                    cmd = cmd + "cat " + lastoutputfile + " | ta-sam_strip_qual.awk > " + outputprefix + ".qual-stripped.sam && "
                    lastoutputfile = outputprefix + ".qual-stripped.sam"
                    intermediates.append(lastoutputfile)
                
                cmd = cmd + "samtools view -hSu " + lastoutputfile + " > " + outputprefix + ".unsorted.bam && "
                lastoutputfile = outputprefix + ".unsorted.bam"
                intermediates.append(lastoutputfile)
                
                if local:
                    cmd = cmd + "samtools sort -m " + str(sortmem) + " " + lastoutputfile + " " + outputprefix + ".sorted && "
                else:
                    cmd = cmd + "samtools sort -m " + str(sortmem) + " " + lastoutputfile + " > " + outputprefix + ".sorted.bam && "
            else: # use pipes whenever possible
                if xa2multi:
                    cmd = cmd + " | xa2multi.pl "
                if strip:
                    cmd = cmd + " | ta-sam_strip_qual.awk "
                if local:
                    cmd = cmd + " | samtools view -hSu - | samtools sort -m " + str(sortmem) + " - " + outputprefix + ".sorted && "
                else:
                    cmd = cmd + " | samtools view -hSu - | samtools sort -m " + str(sortmem) + " -o - $TMPDIR > " + outputprefix + ".sorted.bam || rm -f $TMPDIR*.bam 2> /dev/null && "
            
            outputfile = outputprefix + ".sorted.bam"
            sorted_bam_files.append(outputfile)
            cmd = cmd + "samtools index " + outputfile + ";\n"
            name = prefix + "-bwa_samse-" + str(i).zfill(4)
            sam_job = MyJob(content=cmd, id=None, name=name, predecessors=[aln_job], config_key="bwa_sam")
            bwa_sam_jobs.append(sam_job)
                        
            i += 1
    else:
        i = 1
        cursor = -1
        for index, readfile in enumerate(reads_paths):
            if index <= cursor:
                continue
            
            mate_readfile = readfile
            bwa_aln_cmd1 = "time bwa aln "
            bwa_aln_cmd2 = "time bwa aln "            
            sai1 = ""
            sai2 = ""
            if is_bam(readfile):
                cursor = index
                bwa_aln_cmd1 = "time bwa aln -b1 "
                bwa_aln_cmd2 = "time bwa aln -b2 "
                sai1 = outdir + "/" + os.path.basename(readfile) + ".b1.sai"
                sai2 = outdir + "/" + os.path.basename(mate_readfile) + ".b2.sai"
            else: # fq, fq.gz, fastq, fastq.gz
                mate_readfile = reads_paths[index+1]
                cursor = index+1                 
                sai1 = outdir + "/" + os.path.basename(readfile) + ".sai"
                sai2 = outdir + "/" + os.path.basename(mate_readfile) + ".sai"
            sai_files.append(sai1)
            sai_files.append(sai2)
 
            #bwa aln for mate 1
            cmd = "cd " + outdir + " && "
            cmd = cmd + bwa_aln_cmd1
            if threads > 1:
                cmd = cmd + "-t" + str(threads) + " "
            cmd = cmd + " -f " + sai1 + " "
            cmd = cmd + contigs + " " + readfile + ";\n"
            name = prefix + "-bwa_aln-" + str(i).zfill(4)
            aln_job1 = MyJob(content=cmd, id=None, name=name, predecessors=[indexing_job], config_key="bwa_aln", threads=threads)
            bwa_aln_jobs.append(aln_job1)
            
            #bwa aln for mate 2
            cmd = "cd " + outdir + " && "
            cmd = cmd + bwa_aln_cmd2
            if threads > 1:
                cmd = cmd + "-t" + str(threads) + " "
            cmd = cmd + " -f " + sai2 + " "
            cmd = cmd + contigs + " " + mate_readfile + ";\n"
            name = prefix + "-bwa_aln-" + str(i+1).zfill(4)
            aln_job2 = MyJob(content=cmd, id=None, name=name, predecessors=[indexing_job], config_key="bwa_aln", threads=threads)
            bwa_aln_jobs.append(aln_job2)
            
            #bwa sampe
            cmd = "cd " + outdir + " && "
            outputprefix = outdir + '/' + os.path.basename(readfile)

            cmd = cmd + "time bwa sampe -n50 -s -P -r '@RG	ID:1	LB:1	SM:1' " + contigs + " " + sai1 + " " + sai2 + " " + readfile + " " + mate_readfile
            
            if messy: # generate intermediate files whenever possible
                cmd = cmd + " > " + outputprefix + ".sam && "
                lastoutputfile = outputprefix + ".sam"
                intermediates.append(lastoutputfile)
                
                if xa2multi:
                    cmd = cmd + "cat " + lastoutputfile + " | xa2multi.pl > " + outputprefix + ".xa2multi.sam &&"
                    lastoutputfile = outputprefix + ".xa2multi.sam"
                    intermediates.append(lastoutputfile)
                
                if strip:
                    cmd = cmd + "cat " + lastoutputfile + " | ta-sam_strip_qual.awk > " + outputprefix + ".qual-stripped.sam && "
                    lastoutputfile = outputprefix + ".qual-stripped.sam"
                    intermediates.append(lastoutputfile)
                
                cmd = cmd + "samtools view -hSu " + lastoutputfile + " > " + outputprefix + ".unsorted.bam && "                        
                lastoutputfile = outputprefix + ".unsorted.bam"
                intermediates.append(lastoutputfile)
                
                if local:
                    cmd = cmd + "samtools sort -m " + str(sortmem) + " " + lastoutputfile + " " + outputprefix + ".sorted && "
                else:
                    cmd = cmd + "samtools sort -m " + str(sortmem) + " -o " + lastoutputfile + " $TMPDIR > " + outputprefix + ".sorted.bam || rm -f $TMPDIR*.bam 2> /dev/null && "
            else: # use pipes whenever possible
                if xa2multi:
                    cmd = cmd + " | xa2multi.pl "
                if strip:
                    cmd = cmd + " | ta-sam_strip_qual.awk "                            
                if local:
                    cmd = cmd + " | samtools view -hSu - | samtools sort -m " + str(sortmem) + " - " + outputprefix + ".sorted && " 
                else:
                    cmd = cmd + " | samtools view -hSu - | samtools sort -m " + str(sortmem) + " -o - $TMPDIR > " + outputprefix + ".sorted.bam || rm -f $TMPDIR*.bam 2> /dev/null && "
                
            outputfile = outputprefix + ".sorted.bam"
            sorted_bam_files.append(outputfile)
            cmd = cmd + "samtools index " + outputfile + ";\n"
            name = prefix + "-bwa_sampe-" + str(i).zfill(4) + "-" + str(i+1).zfill(4)
            sam_job = MyJob(content=cmd, id=None, name=name, predecessors=[aln_job1, aln_job2], config_key="bwa_sam")
            bwa_sam_jobs.append(sam_job)    
            
            i += 2

    #Sanity check: If you have paired-end reads, you should have even number of SAI files
    num_bwa_aln_jobs = len(bwa_aln_jobs)
    if not single_end and num_bwa_aln_jobs%2 > 0:
        sys.stderr.write("ERROR: Odd number of bwa_aln jobs (" + str(num_bwa_aln_jobs) + "). It looks like reads files are not paired properly." + '\n')
        sys.exit(1)
    if num_bwa_aln_jobs <= 0:
        sys.stderr.write("ERROR: Nonsensical number of bwa_aln jobs (" + str(num_bwa_aln_jobs) + ")." + '\n')
        sys.exit(1)
    
    intermediates.extend(sai_files)

    #3. Symlink or merge the sorted bam files
    cmd = "cd " + outdir + " && "
    num_bwa_sam_jobs = len(bwa_sam_jobs)
    merge_job = None
    if num_bwa_sam_jobs > 1:
        cmd = cmd + "samtools merge -f " + outdir + "/" + prefix + "-contigs.bam " + " ".join(sorted_bam_files) + " && "
        cmd = cmd + "samtools index " + outdir + "/" + prefix + "-contigs.bam;\n"
        name = prefix + "-samtools_merge"
        merge_job = MyJob(content=cmd, name=name, predecessors=bwa_sam_jobs, config_key="samtools_merge")
        
        # the sorted BAMs are intermediate files because we have the merged BAM
        intermediates.extend(sorted_bam_files)
        # we want to remove the BAM indexes, too
        for bam in sorted_bam_files:
            intermediates.append(bam + ".bai")
    elif num_bwa_sam_jobs == 1:
        # this sorted BAM is not an intermediate file because it is the final BAM
        cmd = cmd + "ln -sf " + sorted_bam_files[0] + " " + outdir + "/" + prefix + "-contigs.bam && "
        cmd = cmd + "ln -sf " + sorted_bam_files[0] + ".bai " + outdir + "/" + prefix + "-contigs.bam.bai;\n"
        bwa_sam_jobs[0].content = bwa_sam_jobs[0].content + cmd
    else:
        sys.stderr.write("ERROR: Nonsensical number of bwa_sam jobs (" + str(num_bwa_sam_jobs) + ")." + '\n')
        sys.exit(1)    

    #4. Clean up
    cmd = "cd " + outdir + ";\n"
    for ftbd in intermediates:
        cmd = cmd + "rm " + ftbd + ";\n"
    name = prefix + "-cleanup-r2c"
    
    cleanup_predecessor = merge_job
    if not cleanup_predecessor:
        cleanup_predecessor = bwa_sam_jobs[0]
        
    cleanup_job = MyJob(content=cmd, name=name, predecessors=[cleanup_predecessor])

    jobs = [indexing_job]
    jobs.extend(bwa_aln_jobs)
    jobs.extend(bwa_sam_jobs)
    if merge_job:
        jobs.append(merge_job)
    
    return (jobs, cleanup_job)


def __main__():
    sys.stdout.flush() #flush print outputs to screen

    usage = "usage: %prog FASTA READS_LIST -a|b|B -G|T [options]"
    version = "%prog 1.4.4 (Build 201208281006)"
    
    parser = OptionParser(usage=usage, version=version)

    mandatory_group = OptionGroup(parser, "Mandatory Arguments")
    mandatory_group.add_option('-a', '--abyssmap', action='store_true', default=False, dest='abyssmap', help='use abyss-map as aligner')
    mandatory_group.add_option('-b', '--bwa', action='store_true', default=False, dest='bwa', help='use BWA as aligner')
    mandatory_group.add_option('-B', '--bowtie2', action='store_true', default=False, dest='bowtie2', help='use Bowtie 2 as aligner')
    mandatory_group.add_option('-G', '--genome', action='store_true', default=False, dest='genome', help='use configurations from \'genome.cfg\'')
    mandatory_group.add_option('-T', '--transcriptome', action='store_true', default=False, dest='transcriptome', help='use configurations from \'transcriptome.cfg\'')
    parser.add_option_group(mandatory_group)
    
    abyssmap_group = OptionGroup(parser, "abyss-map options")
    abyssmap_group.add_option('-k', '--k', type="int", default=24, dest='k', help='k-mer size for abyss-map [default: %default]', metavar='INT')
    parser.add_option_group(abyssmap_group)
     
    bwa_group = OptionGroup(parser, "bwa options")
    bwa_group.add_option('-x', '--xa2multi', action='store_true', default=False, dest='xa2multi', help='run xa2multi for BWA alignments')
    parser.add_option_group(bwa_group)    
    
    parser.add_option('-c', '--cleanup', action='store_true', default=False, dest='cleanup', help='remove intermediate files when all jobs are done')
    parser.add_option('-C', '--cluster', dest='cluster', help='cluster head node', metavar='STR')
    parser.add_option('-d', '--debug', dest='debug', help='run in debug mode; no jobs will be submitted', action='store_true', default=False)
    parser.add_option('-j', '--jobdir', dest='jobdir', help='directory for job scripts and logs', metavar='DIR')   
    parser.add_option('-l', '--local', action='store_true', default=False, dest='local', help='run aligner locally')
    parser.add_option('-m', '--sort-mem', type="int", default=800000000, dest='sortmem', help='maximum memory for `samtools sort` [default: %default]', metavar='INT')
    parser.add_option('-o', '--outdir', dest='outdir', help='output directory', metavar='DIR')
    parser.add_option('-p', '--prefix', dest='prefix', help='library name; used as prefix of output files', metavar='STR')
    parser.add_option('-P', '--project', dest='project', help='project name in configuration file (\'genome.cfg\', \'transcriptome.cfg\')', metavar='STR')
    parser.add_option('-q', '--queue', dest='queue', help='cluster queue', metavar='STR')
    parser.add_option('-s', '--single_end', action='store_true', default=False, dest='single_end', help='input reads files are single-end')
    parser.add_option('-S', '--strip', action='store_true', default=False, dest='strip', help='strip the QUAL columns in SAM records')
    parser.add_option('-t', '--threads', type="int", default=4, dest='threads', help='max number of threads to use [default: %default]', metavar='INT')
    parser.add_option('-w', '--wasteful', dest='messy', help='generate and keep all intermediate files; overrides --cleanup', action='store_true', default=False)    

    (options, args) = parser.parse_args()

    source_contigs = args[0]
    if not os.path.isfile(source_contigs):
        sys.stderr.write("ERROR: Cannot find contigs FASTA file, " + source_contigs + '\n')
        sys.exit(1)

    readslist = args[1]
    if not os.path.isfile(readslist):
        sys.stderr.write("ERROR: Cannot find reads list file, " + readslist + '\n')
        sys.exit(1)
    
    fh = open(readslist, 'r')
    reads_paths = []
    for line in fh:
        linestripped = line.strip()
        if len(linestripped) > 0:
            if os.path.isfile(linestripped):
                reads_paths.append(linestripped)
            else:
                sys.stderr.write("ERROR: Cannot find reads file, " + linestripped + '\n')
                sys.exit(1)
    fh.close()

    factory = R2C_Factory(contigs=source_contigs, reads_paths=reads_paths, prefix=options.prefix, outdir=options.outdir, jobdir=options.jobdir, debug=options.debug, cleanup=options.cleanup, transcriptome=options.transcriptome, genome=options.genome, bwa=options.bwa, abyssmap=options.abyssmap, bowtie2=options.bowtie2, threads=options.threads, single_end=options.single_end, local=options.local, xa2multi=options.xa2multi, strip=options.strip, messy=options.messy, sortmem=options.sortmem, k=options.k, project=options.project, cluster=options.cluster, queue=options.queue)
    factory.create_and_submit_jobs() 
    
    sys.exit(0)

if __name__ == "__main__":
    __main__()
