#!/usr/bin/env python

from datetime import datetime
from optparse import OptionParser, OptionGroup
import os, sys
import random, shutil, glob, grp, subprocess, math, re
from utilities import cfg, job_runner
from utilities.job_runner import MyJob
from utilities.r2c import R2C_Factory
from utilities.tools import run_commands_concurrent

try:
    PACKAGE_DIR = os.environ['TRANSABYSS_PATH']
except KeyError:
    message = "Cannot find TRANSABYSS_PATH in your environment!"
    print_message("ERROR", message, True)

try:
    VERSION = os.environ['TRANSABYSS_VERSION']
    BUILDID = VERSION + ' (build 201210111639)' # This is particularly useful for external releases
except KeyError:
    message = "Cannot find TRANSABYSS_VERSION in your environment!"
    print_message("ERROR", message, True)
    

def print_message(message_type, message, quit=False):
    sys.stderr.write(message_type + ": " + message + os.linesep)
    if quit:
        sys.exit(1)

def is_pos_int(q):
    integer = 0
    try:
        integer = int(q)
    except TypeError:
        return False
    return integer > 0
    
class Tee(object):
    def __init__(self, name, mode):
        self.file = open(name, mode)
        self.stdout = sys.stdout
        self.stderr = sys.stderr
        sys.stdout = self
        sys.stderr = self        
    def finalize(self):
        if self.stdout:        
            sys.stdout = self.stdout
        if self.stderr:
            sys.stderr = self.stderr
        if self.file and not file.closed:
            self.file.flush()
            self.file.close()
            self.file = None
        self.stdout = None
        self.stderr = None
    def write(self, data):
        self.file.write(data)
        self.stdout.write(data)
        #self.stderr.write(data)

class Pipe:
    def __init__(self, assembly_method, project, lib, readlength, assembly_dir=None, cluster=None, genome=None, queue=None, local=False, debug=False, top_dir=None):
        self.assembly_method = assembly_method
        self.project = project
        self.lib = lib
        self.version = VERSION

        self.top_dir = top_dir
        self.working_dir = None
        self.assembly_dir = assembly_dir
        self.ks = []
        
        #self.analysis = []
        self.genome = genome
        self.cluster = cluster
        self.queue = queue
        self.readlength = readlength
        self.local = local
        self.debug = debug
        
    def set_working_dir(self, cfg_file):
        self.settings = cfg.initialize_settings(cfg_file, project=self.project)
                    
        if not self.top_dir:
            try:
                self.top_dir = self.settings[self.project]['topdir']
            except KeyError:
                print_message('ERROR', 'Cannot find top directory for project "%s" in configuration file "%s"' % (self.project, cfg_file), True)
        if not os.path.isdir(self.top_dir):
                print_message('ERROR', 'Cannot find top directory "%s"' % (self.top_dir), True)                        
        self.working_dir = '%s/%s/trans-abyss-v%s' % (self.top_dir, self.lib, self.version)
        
        if not self.genome:
            try:
                self.genome = self.settings[self.project]['reference']
            except KeyError:
                print_message('WARNING', 'Cannot find reference genome for project "%s" in configuration file "%s"' % (self.project, cfg_file), False)
        if self.genome and self.genome not in self.settings['genomes']:
            print_message('WARNING', 'Reference genome "%s" for project "%s" has not been set up in the [genomes] section of configuration file "%s"' % (self.genome, self.project, cfg_file), False)
    
    def setup(self, source=None):
        if self.working_dir is None:
            print_message('ERROR', 'Cannot find working directory for project.', True)
            #return False
        
        print 'Creating directories...'
        cur_path = os.getcwd()
        if not os.path.isdir(self.working_dir):
            os.makedirs(self.working_dir)
        os.chdir(self.working_dir)
        
        if source is None:
            folders = ['assembly', 'merge', 'reads_to_contigs', 'contigs_to_genome', 'reads_to_genome']
            for folder in folders:
                if not os.path.isdir(folder):
                    os.mkdir(folder)
            
            print 'Creating symlinks to assembly directories...'        
            self.link_assemblies()
                    
        else:
            # make the top directory for each step, symlink the important files and directories
            folders = ['merge', 'reads_to_contigs', 'contigs_to_genome', 'reads_to_genome']
            for folder in folders:
                if not os.path.isdir(folder):
                    os.mkdir(folder)
            
            source = '%s/%s/trans-abyss-v%s' % (self.top_dir, self.lib, options.source.lstrip('v'))
            
            print 'Creating symlinks to previous run...'
            os.symlink('%s/assembly' % source, 'assembly')
            
            important_files = ['merge/%s-contigs.fa' % self.lib, 'reads_to_contigs/%s.in' % self.lib, 'reads_to_contigs/%s-contigs.bam' % self.lib, 'reads_to_contigs/%s-contigs.bam.bai' % self.lib]
            for f in important_files:
                sourcefile = '%s/%s' % (source, f)
                if os.path.isfile(sourcefile):
                    os.symlink(sourcefile, '%s/%s' % (self.working_dir, f) )
                # Don't do anything if the source file doesn't exist
                # because the corresponding step might not have been run, yet!
            
            if os.path.isdir('%s/contigs_to_genome/%s-contigs' % (source, self.lib)):
                contigs_to_genome_dir = '%s/contigs_to_genome/%s-contigs' % (self.working_dir, self.lib)
                os.makedirs(contigs_to_genome_dir)
                for folder in ('input', 'output'):
                    os.symlink('%s/contigs_to_genome/%s-contigs/%s' % (source, self.lib, folder), '%s/%s' % (contigs_to_genome_dir, folder) )
            
            if os.path.isdir('%s/reads_to_genome' % source):
                for b in glob.glob('%s/reads_to_genome/*.bam*' % source):
                    os.symlink(b, '%s/reads_to_genome/%s' % (self.working_dir, os.path.basename(b)) )
                   
        os.chdir(cur_path)
                
    def get_ks(self, path):
        ks = {}
        for assembly in glob.glob(os.path.join(path, 'k*')):
            k = os.path.basename(assembly)
            largest_k = k
            if ',' in k:
                largest_k = 'k' + k.split(',')[-1]
            #ks.append(k[1:])
            ks[largest_k[1:]] = k
          
        #ks.sort(key=int)
        return ks
        
    def link_assemblies(self):
        #print self.assembly_dir, self.working_dir
        
        target_assembly_dir = "%s/assembly" % (self.working_dir)
        if not os.path.isdir(target_assembly_dir):
            os.makedirs(target_assembly_dir)
            
        self.ks = self.get_ks(self.assembly_dir)
        ks = self.ks.keys()
        ks.sort(key=int)
        for k in ks:
            source_k_assembly = "%s/%s" % (self.assembly_dir, self.ks[k])
            target_k_assembly = "%s/k%s" % (target_assembly_dir, k)
            if not os.path.exists(target_k_assembly):
                os.symlink(source_k_assembly, target_k_assembly)
        
        #We look for "$LIB.in", then "in", then *.in
        target_in = "%s/in" % (target_assembly_dir)
        infile = None
        if os.path.isfile(os.path.join(self.assembly_dir, '%s.in' % self.lib)):
            infile = os.path.join(self.assembly_dir, '%s.in' % self.lib)
        elif os.path.isfile(os.path.join(self.assembly_dir, 'in')):
            infile = os.path.join(self.assembly_dir, 'in')
        else:        
            in_files = glob.glob(os.path.join(self.assembly_dir, '*.in'))
            num_ins = len(in_files)
            if num_ins == 1:
                infile = in_files[0]
            elif num_ins > 1:                
                print_message('WARNING', 'More than one "in" files found:\n%s' % str(in_files), False)
            else:
                print_message('WARNING', 'Cannot find "in" file.', False)
        
        if infile:
            if not os.path.exists(target_in):
                shutil.copyfile(infile, target_in)
            else:
                print_message('WARNING', '%s already exists.' % target_in, False)
        else:
            print_message('WARNING', '%s will not be created.' % target_in, False)
    
    def run_jobs(self, jobs, local_prefix, cluster_dir):
        if not os.path.isdir(cluster_dir):
            os.makedirs(cluster_dir)
        if self.local:
            job_runner.run_locally(jobs, local_prefix, jobdir=cluster_dir, debug=self.debug)
        else:
            job_runner.submit_jobs(jobs, self.settings, self.cluster, jobdir=cluster_dir, debug=self.debug)
                                    
    def merge(self, input_dir, depend=[]):
        task = 'abyss-rmdups-iterative'
        script = PACKAGE_DIR + "/wrappers/" + task
                
        args = {}
        args['LIB'] = self.lib
        args['INPUT_DIR'] = input_dir
        args['OUTPUT_DIR'] = '%s/merge' % (self.working_dir)    
        args['INDEL_ONLY'] = ''
        
        cluster_dir = '%s/merge/cluster' % (self.working_dir)
        if not os.path.isdir(cluster_dir):
            os.makedirs(cluster_dir)        

        job_name = '%s-merge' % self.lib
        job_content = 'time %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
        job = MyJob(content=job_content, name=job_name, config_key=task, predecessors=depend, queue=self.queue)
        self.run_jobs([job], job_name, cluster_dir)
        return job
        
    def calculate_assembly_stats(self, depend=[]):
        task = 'abyss-fac'
        cluster_dir = '%s/merge/cluster' % self.working_dir
        job_name = '%s-stats' % self.lib        
        job_content = ''
        
        if self.readlength:
            job_content = 'cd %s && time abyss-fac -j -t %s assembly/k*/*-contigs.fa merge/*-contigs.fa > merge/stats.txt' % (self.working_dir, self.readlength)
        else:
            job_content = 'cd %s && time abyss-fac -j assembly/k*/*-contigs.fa merge/*-contigs.fa > merge/stats.txt' % (self.working_dir)
        
        job = MyJob(content=job_content, name=job_name, config_key=task, queue=self.queue, predecessors=depend)
        self.run_jobs([job], job_name, cluster_dir)
        
    def process(self, options):    
        # Action Group 1
        if options.set_up:
            tee = Tee('%s/log.txt' % self.top_dir, 'a')
            print '[%s]: %s (%s)' % (str(datetime.now()), self.lib, self.assembly_method)
            print '*** Set up directories ***'
            self.setup(options.source)
            print '-' * 30
            tee.finalize()
        
        tee = None
                        
        if self.working_dir:
            if os.path.isdir(self.working_dir):
                tee = Tee('%s/log.txt' % self.working_dir, 'a')
                print '[%s]: %s (%s)' % (str(datetime.now()), self.lib, self.assembly_method)
            else:
                print_message('WARNING', 'Cannot find %s' % self.working_dir, False)
        else:
            print_message('WARNING', 'Cannot find working directory for %s' % self.lib, False)
        
        if options.fem:
            print '*** Filter, extend, and merge contigs ***'
            self.fem()            
        if options.prepare_reads:
            print '*** Prepare reads ***'
            self.prepare_reads()
        
        # Action Group 2    
        if options.align_reads:
            print '*** Align reads to merged assembly ***'
            self.align_reads()          
        if options.align_contigs:
            print '*** Align merged assembly to reference genome ***'            
            self.align_contigs(options.rerun)        
        if options.copy_genome_bam:
            print '*** Align reads to reference genome ***'
            self.copy_genome_bam()
        
        # Action Group 3
        if options.fusion:
            print '*** Call fusions and large scale rearrangement events ***'
            self.run_fusions()        
        if options.indel:
            print '*** Call indels ***'
            self.run_indels()        
        if options.splicing:
            print '*** Call novel splicing events and calculate coverage ***'
            if self.assembly_method == 'genome' or self.assembly_method == 'targeted_genome':
                print_message('ERROR', 'This step is not supported for genome libraries.', True)
            else:
                self.run_splicing()            
        if options.track:
            print '*** Create UCSC custom track ***'
            if self.assembly_method == 'genome' or self.assembly_method == 'targeted_genome':
                print_message('ERROR', 'This step is not supported for genome libraries.', True)
            else:
                self.run_track()
        
        if tee:
            print '-' * 30
            tee.finalize()
        
    
    def prepare_reads(self):
        task = 'getreads.py'
        script = PACKAGE_DIR + "/utilities/" + task
        
        reads_list = '%s/assembly/in' % (self.working_dir)
        if not os.path.isfile(reads_list):
            message = "Cannot find assembly reads list '" + reads_list + "'"
            print_message("ERROR", message, True)
        
        args = {}
        args['INPUT_READSLIST'] = reads_list
        args['OUTPUT_READSLIST'] = '%s/reads_to_contigs/%s.in' % (self.working_dir, self.lib)
        args['LIB'] = self.lib
        args['PATH'] = self.working_dir
        args['PROJECT'] = self.project
        
        cmd = None
        args_str = ''
        if self.debug:
            args_str = ' --debug '
        
        if self.local:
            args_str += ' --local '
            args_str += cfg.create_args(self.settings, 'commands', task, args)
            cmd = '%s; python %s %s' % (cfg.get_setup_cmd(), script, args_str)
        elif self.cluster:
            #ssh to head node then submit jobs with qsub
            args_str += cfg.create_args(self.settings, 'commands', task, args)
            cmd = 'ssh %s "%s; python %s %s"' % (self.cluster, cfg.get_setup_cmd(), script, args_str)    
        else:
            #submit jobs with qsub
            args_str += cfg.create_args(self.settings, 'commands', task, args)
            cmd = '%s; python %s %s' % (cfg.get_setup_cmd(), script, args_str)
        print cmd
        #os.system(cmd)
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True)
        stdout, stderr = p.communicate()
        print stdout


    def align_reads(self):
        task = 'r2c.py'
        script = PACKAGE_DIR + "/utilities/" + task
        
        fasta_file = '%s/merge/%s-contigs.fa' % (self.working_dir, self.lib)
        if not os.path.isfile(fasta_file):
            message = "Cannot find merged assembly '" + fasta_file + "'"
            print_message("ERROR", message, True)
        
        reads_list = '%s/reads_to_contigs/%s.in' % (self.working_dir, self.lib)
        if not os.path.isfile(reads_list):
            message = "Cannot find reads list '" + reads_list + "'"
            print_message("WARNING", message)
            
            reads_list = '%s/assembly/in' % (self.working_dir)
            print "Attempting to use the assembly reads list '" + reads_list + "'"
            
            if not os.path.isfile(reads_list):
                message = "Cannot find assembly reads list '" + reads_list + "'"
                print_message("ERROR", message, True)
        
        args = {}
        args['CONTIGS'] = fasta_file
        args['READSLIST'] = reads_list
        args['LIB'] = self.lib
        args['PATH'] = self.working_dir
        args['PROJECT'] = self.project
        
        min_k = None
        assembly_topdir = None
        for assembly in glob.glob(os.path.join(self.working_dir + '/assembly', 'k*')):
            if assembly_topdir is None:
                assembly_topdir = os.path.dirname(os.path.realpath(os.path.abspath(assembly)))
            k = os.path.basename(assembly).replace('k', '')
            if min_k is None or int(k) < int(min_k):
                min_k = k           
        if min_k is None:
            message = 'Cannot find minimum k value from %s/assembly' % (self.working_dir)
            print_message("WARNING", message, False)
        else:
            args['MINK'] = min_k
        
        cmd = None
        args_str = ''
        if self.debug:
            args_str = ' --debug '
        
        if self.local:
            args_str = ' --local '
            args_str += cfg.create_args(self.settings, 'commands', task, args)
            cmd = '%s; python %s %s' % (cfg.get_setup_cmd(), script, args_str)
        elif self.cluster:
            #ssh to head node then submit job with qsub
            args_str += cfg.create_args(self.settings, 'commands', task, args)
            cmd = 'ssh %s "%s; python %s %s"' % (self.cluster, cfg.get_setup_cmd(), script, args_str)    
        else:
            #submit job with qsub
            args_str += cfg.create_args(self.settings, 'commands', task, args)
            cmd = '%s; python %s %s' % (cfg.get_setup_cmd(), script, args_str)
        print cmd
        #os.system(cmd)
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True)
        stdout, stderr = p.communicate()
        print stdout
        
 
    def copy_genome_bam(self):
        #############################################################################################
        # TODO: put your code here to copy your JAGuaR BAM file into the "reads_to_genome" directory
        #############################################################################################
        message = "Please modify 'trans-abyss.py' (def copy_genome_bam) to support copying your JAGuaR BAM file into the 'reads_to_genome' directory."
        print_message("ERROR", message, quit=True)
        
            
    def run_fusions(self):
        task = 'fusion.py'
        script = PACKAGE_DIR + "/analysis/" + task.split('-')[0]
        
        args = {}
        args['PATH'] = self.working_dir
        args['LIB'] = self.lib
        args['GENOME'] = self.genome
        
        r2gdir = self.working_dir + '/reads_to_genome'
        if os.path.isdir(r2gdir):
            genome_bam = self.get_bam_file(r2gdir, genome=True)        
            if genome_bam is None:
                message = 'Cannot find BAM file (*.bam) in directory: %s' % r2gdir
                print_message("WARNING", message, False)
            elif not os.path.isfile(genome_bam + '.bai'):
                message = 'Cannot find BAM index file (*.bam.bai) in directory: %s' % r2gdir
                print_message("WARNING", message, False)
            else:
                args['GENOME_BAM'] = genome_bam
        else:
            message = 'Directory does not exists: %s' % r2gdir
            print_message("ERROR", message, True)
        
        aln_dir = '%s/contigs_to_genome/%s-contigs/output' % (self.working_dir, self.lib)
        alns = []
        if os.path.isdir(aln_dir):
            alns = glob.glob('%s/seq.*.sam' % aln_dir)
        else:
            message = "Cannot find directory, %s" % aln_dir
            print_message("ERROR", message, True)
           
        cluster_dir = '%s/fusions/cluster' % (self.working_dir)
        
        if alns:
            max_task_id = 0
            for aln in alns:
                basename = os.path.basename(aln)
                number = basename.lstrip('seq.').rstrip('.sam')
                if number.isdigit():
                    if int(number) > max_task_id:
                        max_task_id = int(number)                    
                    out_dir_num = cluster_dir + '/' + number
                    if not os.path.isdir(out_dir_num):
                        os.makedirs(out_dir_num)
                    
            args['INPUT'] = '%s/seq.$TA_JOBID.sam' % aln_dir
            args['CTG'] = '%s/contigs_to_genome/%s-contigs/input/seq.$TA_JOBID.fa' % (self.working_dir, self.lib)
            args['OUTPUT_DIR'] = '%s/$TA_JOBID' % cluster_dir
            
            job_name = '%s-fusions' % self.lib
            job_content = 'time python %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
            job = MyJob(content=job_content, name=job_name, config_key=task, first_task_id='1', last_task_id=str(max_task_id), queue=self.queue)
            
            self.run_jobs([job], job_name, cluster_dir)
                                
            #filtering job
            if job:
                self.run_fusion_filtering(depend=[job])
            #TODO: What to do otherwise?    
        else:
            message = "Cannot find alignment files (seq.*.sam) in directory, %s" % aln_dir
            print_message("ERROR", message, True)
    
            
    def run_fusion_filtering(self, depend=None):
        task = 'fusion.py-filter'
        script = PACKAGE_DIR + "/analysis/" + task.split('-')[0]
        
        args = {}
        args['PATH'] = self.working_dir

        cluster_dir = "%s/fusions/cluster" % (self.working_dir)

        job_name = '%s-fusions-filter' % self.lib
        job_content = 'time python %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
        job = MyJob(content=job_content, name=job_name, config_key=task, predecessors=depend, queue=self.queue)
        self.run_jobs([job], job_name, cluster_dir)
            
            
    def run_indels(self):
        task = 'snv_caller.py'
        script = PACKAGE_DIR + "/analysis/" + task.split('-')[0]
        
        args = {}
        args['PATH'] = self.working_dir
        args['LIB'] = self.lib
        args['GENOME'] = self.genome
        
        if self.assembly_method == 'genome' or self.assembly_method == 'targeted_genome':
            r2gdir = self.working_dir + '/reads_to_genome'
            if os.path.isdir(r2gdir):
                genome_bam = self.get_bam_file(r2gdir, genome=True)
                if genome_bam is None:
                    message = 'Cannot find BAM file (*.bam) in directory: %s' % r2gdir
                    print_message("WARNING", message, False)
                elif not os.path.isfile(genome_bam + '.bai'):
                    message = 'Cannot find BAM index file (*.bam.bai) in directory: %s' % r2gdir
                    print_message("WARNING", message, False)
                else:
                    args['GENOME_BAM'] = genome_bam
            else:
                message = 'Directory does not exists: %s' % r2gdir
                print_message("ERROR", message, True)
        
        aln_dir = '%s/contigs_to_genome/%s-contigs/output' % (self.working_dir, self.lib)
        alns = []
        if os.path.isdir(aln_dir):
            alns = glob.glob('%s/seq.*.sam' % aln_dir)
        else:
            message = "Cannot find directory, %s" % aln_dir
            print_message("ERROR", message, True)
        cluster_dir = '%s/indels/cluster' % (self.working_dir)
        
        if alns:
            max_task_id = 0
            for aln in alns:
                basename = os.path.basename(aln)
                number = basename.lstrip('seq.').rstrip('.sam')
                if number.isdigit():
                    if int(number) > max_task_id:
                        max_task_id = int(number)                    
                    out_dir_num = cluster_dir + '/' + number
                    if not os.path.isdir(out_dir_num):
                        os.makedirs(out_dir_num)
                    
            args['ALN'] = '%s/seq.$TA_JOBID.sam' % aln_dir
            args['CTG'] = '%s/contigs_to_genome/%s-contigs/input/seq.$TA_JOBID.fa' % (self.working_dir, self.lib)
            args['OUTPUT_DIR'] = '%s/$TA_JOBID' % cluster_dir
            
            job_name = '%s-indels' % self.lib
            job_content = 'time python %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
            job = MyJob(content=job_content, name=job_name, config_key=task, first_task_id='1', last_task_id=str(max_task_id), queue=self.queue)
            self.run_jobs([job], job_name, cluster_dir)
                
            #filtering job
            if job:
                self.run_indel_filtering(depend=[job])
            #TODO: What to do otherwise?    
        else:
            message = "Cannot find alignment files (seq.*.sam) in directory, %s" % aln_dir
            print_message("ERROR", message, True)

            
    def run_indel_filtering(self, depend=None):
        task = 'snv_caller.py-filter'
        script = PACKAGE_DIR + "/analysis/" + task.split('-')[0]
        
        args = {}
        args['PATH'] = self.working_dir
    
        cluster_dir = "%s/indels/cluster" % (self.working_dir)

        job_name = '%s-indels-filter' % self.lib
        job_content = 'time python %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
        job = MyJob(content=job_content, name=job_name, config_key=task, predecessors=depend, queue=self.queue)
        self.run_jobs([job], job_name, cluster_dir)
            
            
    def get_bam_file(self, from_dir, genome=None):
        bam_files = glob.glob(os.path.join(from_dir + '/*.bam'))
        if len(bam_files) == 1:
            return bam_files[0]
        elif len(bam_files) > 1 and genome:
            return from_dir
        else:
            return None
        
    def get_num_contigs(self, ff):
        num_ctgs = 0
        with open(ff) as f:
            for i, l in enumerate(f):
                if l[0] == '>':
                    num_ctgs += 1
                pass
        return num_ctgs


class PipeTranscriptome(Pipe):
    def __init__(self, assembly_method, project, lib, readlength, assembly_dir=None, cluster=None, genome=None, queue=None, local=False, debug=False, top_dir=None):
        Pipe.__init__(self, assembly_method, project, lib, readlength, assembly_dir=assembly_dir, cluster=cluster, genome=genome, queue=queue, local=local, debug=debug, top_dir=top_dir)
        #print self.assembly_method
        
        cfg_file = "%s/configs/transcriptome.cfg" % (PACKAGE_DIR)
        self.set_working_dir(cfg_file)
        #print self.working_dir
    
    def fem(self):
        filter_jobs = self.filter_extend()
        input_dir = '%s/filter' % (self.working_dir)
        if filter_jobs is not None and len(filter_jobs) > 0:
            merge_job = self.merge(input_dir, depend=filter_jobs)
            if merge_job is not None:
                self.calculate_assembly_stats(depend=[merge_job])
            else:
                message = "Cannot submit assembly-merging jobs."
                print_message("ERROR", message, True)            
        else:
            message = "Cannot submit assembly-filtering jobs."
            print_message("ERROR", message, True)

    def filter_extend(self):
        task = 'abyss-ta-filter'
        script = PACKAGE_DIR + "/wrappers/" + task
        
        args = {}
        args['LIB'] = self.lib
        args['READLENGTH'] = self.readlength
        args['INDEL_ONLY'] = ''
                
        if not self.ks:
            assembly_link_dir = "%s/assembly" % self.working_dir
            if not os.path.isdir(assembly_link_dir) and self.assembly_dir:
                self.link_assemblies()
                
            elif os.path.isdir(assembly_link_dir):
                self.ks = self.get_ks(assembly_link_dir)
        
        if self.ks:
            jobs = []
            cluster_dir = '%s/filter/cluster' % self.working_dir
            if not os.path.isdir(cluster_dir):
                os.makedirs(cluster_dir)

            for k in self.ks:
                args['INPUT_DIR'] = '%s/assembly/k%s' % (self.working_dir, k)
                args['OUTPUT_DIR'] = '%s/filter/k%s' % (self.working_dir, k)
                args['K'] = k
                if not os.path.isdir(args['OUTPUT_DIR']):
                    os.makedirs(args['OUTPUT_DIR'])

                job_name = '%s-k%s-filter' % (self.lib, k)
                job_content = 'time %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
                jobs.append(MyJob(content=job_content, name=job_name, config_key=task, queue=self.queue))
            
            self.run_jobs(jobs, '%s-ks-filter' % self.lib, cluster_dir)
                    
            return jobs
        else:
            message = "Cannot find assembly k-directories, ie. %s/assembly/k*" % self.working_dir
            print_message("ERROR", message, True)
        
    def align_contigs(self, rerun=False):
        top_dir = "%s/contigs_to_genome" % (self.working_dir)
        if not os.path.isdir(top_dir):
            os.makedirs(top_dir)
            
        job_name = "%s-contigs" % (self.lib)
        aln_dir = "%s/%s" % (top_dir, job_name)
        cluster_dir = "%s/cluster" % (aln_dir)
        output_dir = "%s/output" % (aln_dir)
        input_dir = "%s/input" % (aln_dir)
        
        if not os.path.isdir(cluster_dir):
            os.makedirs(cluster_dir)
        if not os.path.isdir(input_dir):
            os.makedirs(input_dir)
            
        if rerun and os.path.islink(output_dir):
            os.remove(output_dir)
        if not os.path.isdir(output_dir):
            os.makedirs(output_dir)
                                
        #input
        if os.listdir(input_dir) and not rerun:
            message = 'input files already exists: %s' % (input_dir)
            print_message("ERROR", message, True)
            
        num_files = 0
        jobs = []
        split_fasta_job = None
        if not rerun:
            task = 'split_fasta'
            script = PACKAGE_DIR + "/utilities/split_fasta.py"
            batch_size = cfg.get_value(self.settings, 'commands', task)
            
            job_name = '%s-split-fasta' % self.lib           
            contigs_file = '%s/merge/%s-contigs.fa' % (self.working_dir, self.lib)
            job_content = 'time python %s %s -b %s -o %s' % (script, contigs_file, batch_size, input_dir)
            
            split_fasta_job = MyJob(content=job_content, name=job_name, config_key=task, queue=self.queue)            
            jobs.append(split_fasta_job)
            
            num_contigs = self.get_num_contigs(contigs_file)
            num_files = int(math.ceil(float(num_contigs) / float(batch_size)))
        else:
            num_files = len(glob.glob(os.path.join(input_dir + '/*.fa')))
            
        #align
        script = 'gmap'
        task = 'gmap'

        args = {}
        if self.genome:
            args['GENOME'] = self.genome
        else:
            message = 'No reference genome were specified in transcriptome.cfg.'
            print_message("ERROR", message, True)
        args['QUERY'] = '%s/seq.$TA_JOBID.fa' % input_dir
        args['OUTPUT'] = '%s/seq.$TA_JOBID.sam' % output_dir
        
        job_name = '%s-gmap' % self.lib
        job_content = 'time %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
        gmap_job_predecessors = None
        if split_fasta_job:
            gmap_job_predecessors = [split_fasta_job]
        jobs.append(MyJob(content=job_content, name=job_name, config_key=task, predecessors=gmap_job_predecessors, first_task_id='1', last_task_id=str(num_files), queue=self.queue))
        self.run_jobs(jobs, job_name, cluster_dir)
                
    def run_track(self):
        task = 'align_parser.py'
        script = PACKAGE_DIR + "/utilities/" + task.split('-')[0]
        
        args = {}
        args['PATH'] = self.working_dir
        args['LIB'] = self.lib
        args['GENOME'] = self.genome

        cluster_dir = '%s/tracks/cluster' % (self.working_dir)
        if not os.path.isdir(cluster_dir):
            os.makedirs(cluster_dir)
           
        job_name = '%s-track' % self.lib
        job_content = 'time python %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
        job = MyJob(content=job_content, name=job_name, config_key=task, queue=self.queue)
        self.run_jobs([job], job_name, cluster_dir)
    
    def run_splicing(self):
        task = 'model_matcher.py'
        script = PACKAGE_DIR + "/analysis/" + task.split('-')[0]
        #args_template = cfg.get_value(self.settings, self.project, task)
        #mem = cfg.get_value(self.settings, 'memory', task)
        
        args = {}
        args['PATH'] = self.working_dir
        args['LIB'] = self.lib
        args['GENOME'] = self.genome
                
        aln_dir = '%s/contigs_to_genome/%s-contigs/output' % (self.working_dir, self.lib)
        alns = []
        if os.path.isdir(aln_dir):
            alns = sorted(glob.glob(os.path.join(aln_dir, '*')))
        cluster_dir = '%s/splicing/cluster' % (self.working_dir)
        
        if alns:
            max_task_id = 0
            for aln in alns:
                basename = os.path.basename(aln)
                number = basename.lstrip('seq.').rstrip('.sam')
                if number.isdigit():
                    if int(number) > max_task_id:
                        max_task_id = int(number)                    
                    out_dir_num = cluster_dir + '/' + number
                    if not os.path.isdir(out_dir_num):
                        os.makedirs(out_dir_num)
                    
            args['ALN'] = '%s/seq.$TA_JOBID.sam' % aln_dir
            args['CTG'] = '%s/contigs_to_genome/%s-contigs/input/seq.$TA_JOBID.fa' % (self.working_dir, self.lib)
            args['OUTPUT_DIR'] = '%s/$TA_JOBID' % cluster_dir
            job_name = '%s-splicing' % self.lib
            job_content = 'time python %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
            job = MyJob(content=job_content, name=job_name, config_key=task, first_task_id='1', last_task_id=str(max_task_id), queue=self.queue)
            self.run_jobs([job], job_name, cluster_dir)
                
            #filtering job
            if job:
                self.run_splicing_filtering(depend=[job])
                
        else:
            message = "Cannot find alignment files (seq.*.sam) in directory, %s" % aln_dir
            print_message("ERROR", message, True)
            
    
    def run_splicing_filtering(self, depend=None):
        task = 'model_matcher.py-filter'
        script = PACKAGE_DIR + "/analysis/" + task.split('-')[0]
        
        args = {}
        args['PATH'] = self.working_dir
        args['GENOME'] = self.genome
                   
        cluster_dir = "%s/splicing/cluster" % (self.working_dir)

        job_name = '%s-splicing-filter' % self.lib
        job_content = 'time python %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
        job = MyJob(content=job_content, name=job_name, config_key=task, predecessors=depend, queue=self.queue)
        self.run_jobs([job], job_name, cluster_dir)
                
    
class PipeGenome(Pipe):
    def __init__(self, assembly_method, project, lib, readlength, assembly_dir=None, cluster=None, genome=None, queue=None, local=False, debug=False, top_dir=None):
        Pipe.__init__(self, assembly_method, project, lib, readlength, assembly_dir=assembly_dir, cluster=cluster, genome=genome, queue=queue, local=local, debug=debug, top_dir=top_dir)
        #print self.assembly_method
        
        cfg_file = "%s/configs/genome.cfg" % (PACKAGE_DIR)    
        #print cfg_file
        self.set_working_dir(cfg_file)
        #print self.working_dir
    
    # for targetted genome libraries
    def concat_contigs_indels(self):
        assembly_dir = '%s/assembly' % (self.working_dir)
        target_dir = '%s/filter' % (self.working_dir)

        cluster_dir = '%s/cluster' % (target_dir)
        jobs = []
        if self.cluster and not os.path.isdir(cluster_dir):
            os.makedirs(cluster_dir)
        
        ks = self.get_ks(assembly_dir)
        for k in ks.keys():
            target_dir = '%s/filter/k%s' % (self.working_dir, k)
            if not os.path.isdir(target_dir):
                os.makedirs(target_dir)

            source_contigs = '%s/k%s/%s-contigs.fa' % (assembly_dir, k, self.lib)
            source_indels = '%s/k%s/%s-indel.fa' % (assembly_dir, k, self.lib)
            target_contigs = '%s/%s-contigs.fa' % (target_dir, self.lib)
            
            cmd = 'cat %s %s > %s' % (source_contigs, source_indels, target_contigs)
            
            job_name = '%s-combine-k%s' % (self.lib, k)
            jobs.append(MyJob(content=cmd, name=job_name, config_key=None, predecessors=None, queue=self.queue))
        
        self.run_jobs(jobs, '%s-combine-ks', cluster_dir)
        return jobs   
    
    # for full genome libraries
    def include_indel_bubbles(self):
        task = 'abyss-ta-filter'
        script = PACKAGE_DIR + "/wrappers/" + task
            
        args = {}
        args['LIB'] = self.lib
        args['READLENGTH'] = self.readlength
        args['INDEL_ONLY'] = '-b'
            
        #bubble extending jobs
        extend_jobs = []
        cluster_dir = "%s/merge/indel_bubbles/cluster" % (self.working_dir)
        if not os.path.isdir(cluster_dir):
            os.makedirs(cluster_dir)
        
        assembly_dir = self.working_dir + '/assembly'
        ks = sorted(glob.glob(os.path.join(assembly_dir, 'k*')))
        k_sizes = []
        for k in ks:            
            args['K'] = os.path.basename(k).replace('k', '')
            args['INPUT_DIR'] = k
            args['OUTPUT_DIR'] = '%s/merge/indel_bubbles/k%s' % (self.working_dir, args['K'])
            k_sizes.append(args['K'])
        
            if not os.path.isdir(args['OUTPUT_DIR']):
                os.mkdir(args['OUTPUT_DIR'])
            
            job_name = '%s-k%s-extend' % (self.lib, args['K'])
            job_content = 'time %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
            extend_jobs.append(MyJob(content=job_content, name=job_name, config_key=task, predecessors=None, queue=self.queue))
        self.run_jobs(extend_jobs, '%s-ks-extend' % self.lib, cluster_dir)
               
        #bubble merging jobs - dependent on first jobs
        task = 'abyss-rmdups-iterative'
        script = PACKAGE_DIR + "/wrappers/" + task

        args['INPUT_DIR'] = self.working_dir + '/merge/indel_bubbles'
        args['OUTPUT_DIR'] = self.working_dir + '/merge/indel_bubbles'
        args['LIB'] = self.lib

        largest_k = self.working_dir + '/assembly/k' + str(max(k_sizes))
        pe_contigs = largest_k + '/' + self.lib + '-contigs.fa'
        merged_indels = args['OUTPUT_DIR'] + '/' + self.lib + '-b.fa'
        final_merged_assembly = self.working_dir + '/merge/' + self.lib + '-contigs.fa'

        job_content = 'time %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args))
        job_content += ' && cat %s %s > %s' % (merged_indels, pe_contigs, final_merged_assembly)
        job_name = "%s-merge" % (self.lib)   
        merge_job = MyJob(content=job_content, name=job_name, config_key=task, predecessors=extend_jobs, queue=self.queue)
        self.run_jobs([merge_job], job_name, cluster_dir)
        return merge_job
    
    def fem(self):
        if self.assembly_method == 'targeted_genome':            
            input_dir = '%s/filter' % (self.working_dir)
            jobs = self.concat_contigs_indels()
            if jobs:
                merge_job = self.merge(input_dir, depend=jobs)
                if merge_job is not None:
                    self.calculate_assembly_stats(depend=[merge_job])
                else:
                    print_message("ERROR", "Cannot submit job for merging assemblies!", True)
            else:
                print_message("ERROR", "Cannot submit jobs for concatenating *-indels.fa *-contigs.fa!", True)
        elif self.assembly_method == 'genome':
            merge_job = self.include_indel_bubbles()
            if merge_job is not None:
                self.calculate_assembly_stats(depend=[merge_job])
            else:
                print_message("ERROR", "Cannot submit job for merging assemblies!", True)
        else:
            message = 'Unknown assembly method: %s' % str(self.assembly_method)
            print_message("ERROR", message, True)
    
    def align_contigs(self, rerun=False):       
        top_dir = "%s/contigs_to_genome" % (self.working_dir)
        if not os.path.isdir(top_dir):
            os.makedirs(top_dir)
            
        job_name = "%s-contigs" % (self.lib)
        aln_dir = "%s/%s" % (top_dir, job_name)
        cluster_dir = "%s/cluster" % (aln_dir)
        output_dir = "%s/output" % (aln_dir)
        input_dir = "%s/input" % (aln_dir)
        
        if not os.path.isdir(input_dir):
            os.makedirs(input_dir)
        if not os.path.isdir(cluster_dir):
            os.makedirs(cluster_dir)
        if not os.path.isdir(output_dir):
            os.makedirs(output_dir)

        if rerun and os.path.islink(output_dir):
            os.remove(output_dir)
        if not os.path.isdir(output_dir):
            os.makedirs(output_dir)

        #split contigs
        if os.listdir(input_dir) and not rerun:
            message = 'input files already exists: %s' % (input_dir)
            print_message("ERROR", message, True)
            
        num_files = 0
        jobs = []
        split_fasta_job = None
        if not rerun:
            task = 'split_fasta'
            script = PACKAGE_DIR + "/utilities/split_fasta.py"
            batch_size = cfg.get_value(self.settings, 'commands', task)
            
            job_name = '%s-split-fasta' % self.lib           
            contigs_file = '%s/merge/%s-contigs.fa' % (self.working_dir, self.lib)
            job_content = 'time python %s %s -b %s -o %s' % (script, contigs_file, batch_size, input_dir)
            
            split_fasta_job = MyJob(content=job_content, name=job_name, config_key=task, queue=self.queue)            
            jobs.append(split_fasta_job)
            
            num_contigs = self.get_num_contigs(contigs_file)
            num_files = int(math.ceil(float(num_contigs) / float(batch_size)))
        else:
            num_files = len(glob.glob(os.path.join(input_dir + '/*.fa')))

              
        #alignment            
        script = 'bwa'
        task = 'bwasw'
        args_template = cfg.get_value(self.settings, self.project, task)
        mem = cfg.get_value(self.settings, 'memory', task)

        args = {}
        if self.settings.has_key('genomes') and self.settings['genomes'].has_key(self.genome):
            args['TARGET'] = self.settings['genomes'][self.genome]
        else:
            message = 'No reference genome were specified in genome.cfg.'
            print_message("ERROR", message, True) 

        args['QUERY'] = '%s/seq.$TA_JOBID.fa' % input_dir
        args['OUTPUT'] = '%s/seq.$TA_JOBID.sam' % output_dir
        
        job_name = '%s-bwasw' % self.lib
        job_content = 'time %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
        bwasw_job_predecessors = None
        if split_fasta_job:
            bwasw_job_predecessors = [split_fasta_job]
        jobs.append(MyJob(content=job_content, name=job_name, config_key=task, predecessors=bwasw_job_predecessors, first_task_id='1', last_task_id=str(num_files), queue=self.queue))
        self.run_jobs(jobs, job_name, cluster_dir)
               

class PipeStranded(PipeTranscriptome):
    def __init__(self, assembly_method, project, meta_lib, lib_plus, lib_minus, readlength, assembly_dir_plus=None, assembly_dir_minus=None, cluster=None, genome=None, queue=None, local=False, debug=False, top_dir=None):
        PipeTranscriptome.__init__(self, assembly_method, project, meta_lib, readlength, assembly_dir=None, cluster=cluster, genome=genome, queue=queue, local=local, debug=debug, top_dir=top_dir)
        self.meta_lib = meta_lib
        self.lib_plus = lib_plus
        self.lib_minus = lib_minus
        self.assembly_dir_plus = assembly_dir_plus
        self.assembly_dir_minus = assembly_dir_minus
    
    #setup directories
    def setup(self, source=None):
        if self.working_dir is None:
            return False
                    
        cur_path = os.getcwd()
        if not os.path.isdir(self.working_dir):
            os.makedirs(self.working_dir)
        os.chdir(self.working_dir)
        
        if source is None:
            folders = ['assembly', 'assembly/plus_strand', 'assembly/minus_strand', 'filter', 'filter/plus_strand', 'filter/minus_strand', 'merge', 'merge/plus_strand', 'merge/minus_strand', 'reads_to_contigs', 'reads_to_contigs/plus_strand', 'reads_to_contigs/minus_strand', 'reads_to_contigs/unknown', 'contigs_to_genome', 'reads_to_genome']
            print 'Creating directories...'
            for folder in folders:
                if not os.path.isdir(folder):
                    os.makedirs(folder)
            
            print 'Creating symlinks to assembly directories...'
            self.link_assemblies()                    
        else:
            # make the top directory for each step, symlink the important files and directories
            folders = ['merge', 'reads_to_contigs', 'contigs_to_genome', 'reads_to_genome']
            print 'Creating directories...'
            for folder in folders:
                if not os.path.isdir(folder):
                    os.mkdir(folder)
            
            source = '%s/%s/trans-abyss-v%s' % (self.top_dir, self.lib, options.source.lstrip('v'))
            
            print 'Creating symlinks to previous run...'
            os.symlink('%s/assembly' % source, 'assembly')
            
            important_files = ['merge/%s-contigs.fa' % self.lib, 'reads_to_contigs/%s.in' % self.lib, 'reads_to_contigs/%s-contigs.bam' % self.lib, 'reads_to_contigs/%s-contigs.bam.bai' % self.lib]
            for f in important_files:
                sourcefile = '%s/%s' % (source, f)
                if os.path.isfile(sourcefile):
                    os.symlink(sourcefile, '%s/%s' % (self.working_dir, f) )
                # Don't do anything if the source file doesn't exist
                # because the corresponding step might not have been run, yet!
            
            if os.path.isdir('%s/contigs_to_genome/%s-contigs' % (source, self.lib)):
                contigs_to_genome_dir = '%s/contigs_to_genome/%s-contigs' % (self.working_dir, self.lib)
                os.makedirs(contigs_to_genome_dir)
                for folder in ('input', 'output'):
                    os.symlink('%s/contigs_to_genome/%s-contigs/%s' % (source, self.lib, folder), '%s/%s' % (contigs_to_genome_dir, folder) )
            
            if os.path.isdir('%s/reads_to_genome' % source):
                for b in glob.glob('%s/reads_to_genome/*.bam*' % source):
                    os.symlink(b, '%s/reads_to_genome/%s' % (self.working_dir, os.path.basename(b)) )
            
        os.chdir(cur_path)        
    
    #link the assemblies for both strands
    def link_assemblies(self):
        #print self.lib_plus + " " + self.assembly_dir_plus
        #print self.lib_minus + " " + self.assembly_dir_minus        
        
        target_assembly_dir = "%s/assembly" % (self.working_dir)
        target_assembly_dir_plus = "%s/plus_strand" % (target_assembly_dir)
        target_assembly_dir_minus = "%s/minus_strand" % (target_assembly_dir)
        for d in [target_assembly_dir, target_assembly_dir_plus, target_assembly_dir_minus]:        
            if not os.path.isdir(d):
                os.makedirs(d)        

        #symlinks for plus strand            
        self.ks_plus = self.get_ks(self.assembly_dir_plus)
        ks_plus = self.ks_plus.keys()
        ks_plus.sort(key=int)
        for k in ks_plus:
            source_k_assembly = "%s/%s" % (self.assembly_dir_plus, self.ks_plus[k])
            target_k_assembly = "%s/k%s" % (target_assembly_dir_plus, k)
            if not os.path.exists(target_k_assembly):
                os.symlink(source_k_assembly, target_k_assembly)
                
        #symlinks for minus strand        
        self.ks_minus = self.get_ks(self.assembly_dir_minus)
        ks_minus = self.ks_minus.keys()
        ks_minus.sort(key=int)
        for k in ks_minus:
            source_k_assembly = "%s/%s" % (self.assembly_dir_minus, self.ks_minus[k])
            target_k_assembly = "%s/k%s" % (target_assembly_dir_minus, k)
            if not os.path.exists(target_k_assembly):
                os.symlink(source_k_assembly, target_k_assembly)
        
        #TODO: We should look for "$LIB.in", then "in", then *.in
        in_files_plus = glob.glob(os.path.join(self.assembly_dir_plus, '*in'))        
        in_files_minus = glob.glob(os.path.join(self.assembly_dir_minus, '*in'))
        if len(in_files_plus) == 1 and len(in_files_minus) == 1:
            target_in = "%s/in" % (target_assembly_dir)
            
            shutil.copyfile(in_files_plus[0], target_assembly_dir_plus + os.sep + 'in')
            shutil.copyfile(in_files_minus[0], target_assembly_dir_minus + os.sep + 'in')
                        
            #cat both in files together
            #target_in_fh = open(target_in,'wb')
            #shutil.copyfileobj(open(in_files_plus[0],'rb'), target_in_fh)
            #shutil.copyfileobj(open(in_files_minus[0],'rb'), target_in_fh)
            #target_in_fh.close()
        else:
            message = "Cannot find the 'in' files from the assembly directories..." + os.linesep + str(in_files_plus) + os.linesep + str(in_files_minus)
            print_message("ERROR", message, True)
    
    #run fem
    def fem(self):
        merge_plus_job = None
        merge_minus_job = None

        # 1.a) filter,extend contigs in PLUS assemblies
        filter_plus_jobs = self.filter_extend('plus')       
        if filter_plus_jobs and len(filter_plus_jobs) > 0:
            # 2.a) merge PLUS assemblies
            merge_plus_job = self.merge(depend=filter_plus_jobs, strand='plus')
            if not merge_plus_job:
                message = "Cannot submit jobs for merging assemblies for 'plus' strand!"
                print_message("ERROR", message, True)
        else:
            message = "Cannot submit jobs for filtering and extension of contigs for 'plus' strand!"
            print_message("ERROR", message, True)
            
        # 1.b) filter,extend contigs in MINUS assemblies
        filter_minus_jobs = self.filter_extend('minus')
        if filter_minus_jobs and len(filter_minus_jobs) > 0:
            # 2.b) merge MINUS assemblies
            merge_minus_job = self.merge(depend=jids_filter_minus, strand='minus')
            if not merge_minus_job:
                message = "Cannot submit jobs for merging assemblies for 'minus' strand!"
                print_message("ERROR", message, True)
        else:
            message = "Cannot submit jobs for filtering and extension of contigs for 'minus' strand!"
            print_message("ERROR", message, True)
            
        # 3. concatenate the merged PLUS assembly and the merged MINUS assembly together
        concat_job = self.concat(depend=[merge_plus_job, merge_minus_job])
        if concat_job is not None:
            self.calculate_assembly_stats(depend=[concat_job])
        else:
            message = "Cannot submit job for concatenating the assemblies for 'plus' and 'minus' strand!"
            print_message("ERROR", message, True)   
        
    def filter_extend(self, strand=None):
        task = 'abyss-ta-filter'
        script = PACKAGE_DIR + "/wrappers/" + task
        #args_template = cfg.get_value(self.settings, self.project, task)
        #mem = cfg.get_value(self.settings, 'memory', task)
                
        assembly_link_dir = "%s/assembly" % (self.working_dir)
        if not os.path.isdir(assembly_link_dir):
            self.link_assemblies()                
        
        lib = None
        if strand == 'plus':
            lib = self.lib_plus
        elif strand == 'minus':
            lib = self.lib_minus
        else:
            message = "Unknown strand (" + str(strand) + ")"
            print_message("ERROR", message, True)
        
        ks = self.get_ks('%s/assembly/%s_strand' % (self.working_dir, strand))        
        if ks and len(ks) > 0:
            jobs = []
            cluster_dir = '%s/filter/%s_strand/cluster' % (self.working_dir, strand)
            if not os.path.isdir(cluster_dir):
                os.makedirs(cluster_dir)
                
            args = {}
            args['READLENGTH'] = self.readlength
            args['INDEL_ONLY'] = '' # = both contigs and indels
            args['LIB'] = lib 

            for k in self.ks:
                args['INPUT_DIR'] = '%s/assembly/%s_strand/k%s' % (self.working_dir, strand, k)
                args['OUTPUT_DIR'] = '%s/filter/%s_strand/k%s' % (self.working_dir, strand, k)
                args['K'] = k
                if not os.path.isdir(args['OUTPUT_DIR']):
                    os.makedirs(args['OUTPUT_DIR'])

                job_name = '%s-k%s-filter' % (self.lib, k)
                job_content = 'time %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
                jobs.append(MyJob(content=job_content, name=job_name, config_key=task, predecessors=depend, queue=self.queue))
            self.run_jobs(jobs, '%s-ks-filter' % self.lib, cluster_dir)
                    
            return jobs
        else:
            message = "Cannot find assembly k-directories, ie. %s/assembly/k*" % self.working_dir
            print_message("ERROR", message, True)
        
            
    def merge(self, input_dir, depend=[]):
        task = 'abyss-rmdups-iterative'
        script = PACKAGE_DIR + "/wrappers/" + task
                
        lib = None
        if strand == 'plus':
            lib = self.lib_plus
        elif strand == 'minus':
            lib = self.lib_minus
        else:
            message = "Unknown strand (" + str(strand) + ")"
            print_message("ERROR", message, True)
                
        args = {}
        args['LIB'] = lib
        args['INPUT_DIR'] = '%s/filter/%s_strand' % (self.working_dir, strand)
        args['OUTPUT_DIR'] = '%s/merge/%s_strand' % (self.working_dir, strand)
        args['INDEL_ONLY'] = '' # = both contigs and indels
        
        cluster_dir = '%s/merge/%s_strand/cluster' % (self.working_dir, strand)
        if not os.path.isdir(cluster_dir):
            os.makedirs(cluster_dir)       

        job_name = '%s-merge' % self.lib
        merged_fasta = '%s/merge/%s_strand/%s-contigs.fa' % (self.working_dir, strand, lib)

        # the strand must be flagged in the contig id for the 3 sets of r2c alignments to be merged        
        job_content = 'time %s %s' % (script, cfg.create_args(self.settings, 'commands', task, args) )
        job_content += ' && sed -i.bak -e "s/>/>%s_/g" %s && rm %s.bak' % (strand, merged_fasta, merged_fasta)
        job = MyJob(content=job_content, name=job_name, config_key=task, predecessors=depend, queue=self.queue)
        self.run_jobs([job], job_name, cluster_dir)
        return job
    
    def concat(self, depend=[]):
        merged_fasta_plus = '%s/merge/plus_strand/%s-contigs.fa' % (self.working_dir, self.lib_plus)
        merged_fasta_minus = '%s/merge/minus_strand/%s-contigs.fa' % (self.working_dir, self.lib_minus)
        concat_fasta = '%s/merge/%s-contigs.fa' % (self.working_dir, self.meta_lib)
        
        cluster_dir = '%s/merge/cluster' % (self.working_dir)
        if not os.path.isdir(cluster_dir):
            os.makedirs(cluster_dir)
        
        job_content = 'time cat %s %s > %s' % (merged_fasta_plus, merged_fasta_minus, concat_fasta)
        job_name = "%s-concat" % (self.lib)   
        job = MyJob(content=job_content, name=job_name, config_key=None, predecessors=depend, queue=self.queue)
        self.run_jobs([job], job_name, cluster_dir)
        return job

    def calculate_assembly_stats(self, depend=[]):
        task = 'abyss-fac'
        cluster_dir = '%s/merge/cluster' % self.working_dir
        job_name = '%s-stats' % self.lib
        job_content = 'cd %s && time abyss-fac -t %s assembly/plus_strand/k*/*-contigs.fa assembly/minus_strand/k*/*-contigs.fa merge/plus_strand/*-contigs.fa merge/minus_strand/*-contigs.fa merge/*-contigs.fa > merge/stats.txt' % (self.working_dir, self.readlength)
        job = MyJob(content=job_content, name=job_name, config_key=task, queue=self.queue, predecessors=depend)
        self.run_jobs([job], job_name, cluster_dir)

    def prepare_reads(self):
        # TODO: Support splitting/converting read files
        
        # Assumptions:
        # - input files are BAM files
        # - use BWA as aligner
    
        assembly_plus_in = '%s/assembly/plus_strand/in' % self.working_dir
        assembly_minus_in = '%s/assembly/minus_strand/in' % self.working_dir
        
        plus_reads = []
        minus_reads = []
        unknown_reads = []

        fh = open(assembly_plus_in, 'r')
        for line in fh:
            plus_reads.append(line.strip())
        fh.close()       
        
        fh = open(assembly_minus_in, 'r')
        for line in fh:
            read = line.strip()
            if read in plus_reads:
                plus_reads.remove(read)
                unknown_reads.append(read)                
            else:
                minus_reads.append(read)
        fh.close()
        
        r2c_plus_in = '%s/reads_to_contigs/plus_strand/in' % self.working_dir
        fh = open(r2c_plus_in, 'w')
        print "Read files for the plus strand:"
        for read in plus_reads:
            fh.write(read + os.linesep)
            print read
        fh.close()
        
        r2c_minus_in = '%s/reads_to_contigs/minus_strand/in' % self.working_dir
        fh = open(r2c_minus_in, 'w')
        print "Read files for the minus strand:"
        for read in minus_reads:
            fh.write(read + os.linesep)
            print read
        fh.close()
        
        r2c_unknown_in = '%s/reads_to_contigs/unknown/in' % self.working_dir
        fh = open(r2c_unknown_in, 'w')
        print "Read files for unknown strand:"
        for read in unknown_reads:
            fh.write(read + os.linesep)
            print read
        fh.close()
        
        print "DONE. All reads are prepared."

    def align_reads(self):
        merged_fasta_plus = '%s/merge/plus_strand/%s-contigs.fa' % (self.working_dir, self.lib_plus)
        merged_fasta_minus = '%s/merge/minus_strand/%s-contigs.fa' % (self.working_dir, self.lib_minus)
        concat_fasta = '%s/merge/%s-contigs.fa' % (self.working_dir, self.meta_lib)
                
        outdir_plus = '%s/reads_to_contigs/plus_strand' % self.working_dir
        outdir_minus = '%s/reads_to_contigs/minus_strand' % self.working_dir
        outdir_unknown = '%s/reads_to_contigs/unknown' % self.working_dir
        
        plus_job = self.align_reads_helper(merged_fasta_plus, outdir_plus + "/in", self.meta_lib, outdir_plus, outdir_plus + "/cluster")
        if not plus_job:
            message = "Cannot submit reads-to-contigs jobs for reads from plus strand."
            print_message("ERROR", message, True)

        minus_job = self.align_reads_helper(merged_fasta_minus, outdir_minus + "/in", self.meta_lib, outdir_minus, outdir_minus + "/cluster")
        if not minus_job:
            message = "Cannot submit reads-to-contigs jobs for reads from minus strand."
            print_message("ERROR", message, True)

        unknown_job = self.align_reads_helper(concat_fasta, outdir_unknown + "/in", self.meta_lib, outdir_unknown, outdir_unknown + "/cluster")
        if not unknown_job:
            message = "Cannot submit reads-to-contigs jobs for reads from unknown strand."
            print_message("ERROR", message, True)
       
        plus_bam = '%s/%s-contigs.bam' % (outdir_plus, self.meta_lib)
        minus_bam = '%s/%s-contigs.bam' % (outdir_minus, self.meta_lib)
        unknown_bam = '%s/%s-contigs.bam' % (outdir_unknown, self.meta_lib)
        merged_bam = '%s/reads_to_contigs/%s-contigs.bam' % (self.working_dir, self.meta_lib)

        task = 'samtools_sort'

        #TODO: Support getting options for mergesamfiles.jar from the config files
        #      use argsparse module (function parse_args)        

        job_content = '{ samtools view -h %s && samtools view %s && samtools view %s; } | samtools view -hSu - | samtools sort -m 800000000 -o - $TMPDIR > %s || (ls $TMPDIR*.bam 2> /dev/null && rm -f $TMPDIR*.bam) && samtools index %s && echo "Job completed."' % (unknown_bam, plus_bam, minus_bam, merged_bam, merged_bam)
        
        cluster_dir = '%s/reads_to_contigs/cluster' % self.working_dir
        if not os.path.isdir(cluster_dir):
            os.makedirs(cluster_dir)
            
        job_name = '%s-mergebams' % self.meta_lib
        merge_job = MyJob(content=job_content, name=job_name, config_key=task, predecessors=[plus_job, mius_job, unknown_job], queue=self.queue)
        self.run_jobs([merge_job], job_name, cluster_dir)

    def align_reads_helper(self, contigs, reads_in, prefix, outdir, jobdir):
        reads_paths = []
        
        fh = open(reads_in, 'r')
        for line in fh:
            linestripped = line.strip()
            if len(linestripped) > 0:
                reads_paths.append(linestripped)
        fh.close()
                
        #TODO: Support getting options for r2c.py from the config files
        #      use argsparse module (function parse_args)
        factory = R2C_Factory(contigs=contigs, reads_paths=reads_paths, prefix=prefix, outdir=outdir, jobdir=jobdir, transcriptome=True, bwa=True, strip=True, project=self.project, local=self.local, debug=self.debug)
        return factory.create_and_submit_jobs() 


def make_pipes_from_input_file(options):
    pipes = []
    startinglibrary = options.library
    numlibraries = options.numlibraries
    if not is_pos_int(numlibraries):
        message = 'The number of libraries specified (%s) is not a positive integer.' % str(numlibraries)
        print_message("ERROR", message, True)        
    seenstart = False        
    
    fh = open(options.input, 'r')
    for line in fh:
        cols = line.strip().split()
        num_cols = len(cols)
        if num_cols > 0: #otherwise, it's just a line of empty spaces
            
            if startinglibrary:
                if startinglibrary == cols[0]:
                    seenstart = True
                elif not seenstart:
                    continue
                elif numlibraries and int(numlibraries)<=len(pipes):
                    break            
            
            if num_cols >= 6:
                # LIBRARY ASSEMBLY PROJECT READLENGTH TYPE METALIBRARY
                # example:
                # A12345A /assembly/dir/A12345A Project_Name 75 plus_strand A12345
                # A12345B /assembly/dir/A12345B Project_Name 75 minus_strand A12345                    
                
                library, assembly_dir, project, readlength, library_type, meta_lib = cols
                
                if not is_pos_int(readlength):
                    message = 'The read length specified (%s) is not a positive integer.' % readlength
                    print_message("ERROR", message, True)
                if not os.path.isdir(assembly_dir):
                    message = 'Cannot find the assembly directory specified: %s' % assembly_dir
                    # Only a warning because users can potentially run Trans-ABySS on any FASTA file.
                    print_message("WARNING", message, False)
                
                library_type = library_type.lower()
                if library_type == 'plus_strand' or library_type == 'minus_strand':
                    line2 = fh.next()
                    cols2 = line2.strip().split()
                    if len(cols2) < 6:
                        message = "Less than 6 columns detected for stranded library..." + os.linesep + line + os.linesep + line2
                        print_message("ERROR", message, True)
                    library2, assembly_dir2, project2, library_type2, meta_lib2 = cols2
                    if project != project2:
                        message = "The projects do not agree for stranded library..." + os.linesep + line + os.linesep + line2
                        print_message("ERROR", message, True)
                    if library_type == library_type2:
                        message = "Both strands cannot be the same..." + os.linesep + line + os.linesep + line2
                        print_message("ERROR", message, True)
                    if meta_lib != meta_lib2:
                        message = "Meta-library is not the same..." + os.linesep + line + os.linesep + line2
                        print_message("ERROR", message, True)
                    
                    lib1 = None
                    lib2 = None
                    adir1 = None
                    adir2 = None
                    if library_type == 'plus_strand':
                        lib1 = library
                        adir1 = assembly_dir
                        lib2 = library2
                        adir2 = assembly_dir2
                    else:
                        lib2 = library
                        adir2 = assembly_dir
                        lib1 = library2
                        adir1 = assembly_dir2                        

                    pipes.append(PipeStranded('stranded', project, meta_lib, lib1, lib2, readlength, adir1, adir2, cluster=options.cluster, queue=options.queue, local=options.local, debug=options.debug))
                elif 'transcriptome' == library_type:
                    pipes.append(PipeTranscriptome('transcriptome', project, library, readlength, assembly_dir=assembly_dir, cluster=options.cluster, queue=options.queue, local=options.local, debug=options.debug))
                elif 'genome' == library_type:
                    pipes.append(PipeGenome('genome', project, library, readlength, assembly_dir=assembly_dir, cluster=options.cluster, queue=options.queue, local=options.local, debug=options.debug))
                elif 'targeted_genome' == library_type or 'targetted_genome' == library_type:
                    pipes.append(PipeGenome('targeted_genome', project, library, readlength, assembly_dir=assembly_dir, cluster=options.cluster, queue=options.queue, local=options.local, debug=options.debug))
                else:
                    message = "Unknown library type..." + os.linesep + line
                    print_message("ERROR", message, True)                       
            elif num_cols >= 4:
                # LIB ASSEMBLY PROJECT READLENGTH TYPE
                # OR
                # LIB ASSEMBLY PROJECT READLENGTH
                
                library, assembly_dir, project, readlength = cols[0:4]
                library_type = None
                if num_cols == 5:
                    library_type = cols[4].lower()
                else:
                    if not options.run_genome and not options.run_transcriptome:
                        message = "You must use either option -G(genome) or -T(transcriptome)." + os.linesep + "You may alternatively specify the library type in the 4th column of input file."
                        print_message("ERROR", message, True)
                    elif options.run_genome and options.run_transcriptome:
                        message = "You must use only one of -G(genome) or -T(transcriptome)"
                        print_message("ERROR", message, True)                            
                
                if not is_pos_int(readlength):
                    message = 'The read length specified (%s) is not a positive integer.' % readlength
                    print_message("ERROR", message, True)
                if not os.path.isdir(assembly_dir):
                    # Only a warning because users can potentially run Trans-ABySS on any FASTA file in the right format.
                    message = 'Cannot find the assembly directory specified: %s' % assembly_dir
                    print_message("WARNING", message, False)                  

                if 'transcriptome' == library_type or options.run_transcriptome:
                    pipes.append(PipeTranscriptome('transcriptome', project, library, readlength, assembly_dir=assembly_dir, cluster=options.cluster, queue=options.queue, local=options.local, debug=options.debug))
                elif 'genome' == library_type or (options.run_genome and not options.targeted):
                    pipes.append(PipeGenome('genome', project, library, readlength, assembly_dir=assembly_dir, cluster=options.cluster, queue=options.queue, local=options.local, debug=options.debug))
                elif 'targeted_genome' == library_type or 'targetted_genome' == library_type or (options.run_genome and options.targeted):
                    pipes.append(PipeGenome('targeted_genome', project, library, readlength, assembly_dir=assembly_dir, cluster=options.cluster, queue=options.queue, local=options.local, debug=options.debug))
                elif 'plus_strand' == library_type or 'minus_strand' == library_type:
                    message = "Meta-library name is not specified in the 5th column..." + os.linesep + line
                    print_message("ERROR", message, True)
                else:
                    message = "Unknown library type..." + os.linesep + line
                    print_message("ERROR", message, True)
            else:
                #this line has incorrect format
                message = "This line of input file has less than 3 columns..." + os.linesep + line
                print_message("ERROR", message, True)
    fh.close()
    return pipes


if __name__ == '__main__':   
    usage = "Usage: %prog [options]"
    parser = OptionParser(usage=usage, version=BUILDID)
    
    sample_types = OptionGroup(parser, "Sample Types")
    sample_types.add_option("-G", "--run_genome", dest="run_genome", help="run genome pipeline", action="store_true", default=False)
    sample_types.add_option("-t", "--targeted", dest="targeted", help="genome is targeted, used in conjunction with --run_genome", action="store_true", default=False)
    sample_types.add_option("-T", "--run_transcriptome", dest="run_transcriptome", help="run transcriptome pipeline", action="store_true", default=False)
    parser.add_option_group(sample_types)
    
    inputs = OptionGroup(parser, "Inputs")
    inputs.add_option("-p", "--project", dest="project", help="project name", metavar='STR')
    inputs.add_option("-P", "--topdir", dest="top_dir", help="project directory", metavar='DIR')
    inputs.add_option("-a", "--assembly_dir", dest="assembly_dir", help="ABySS assembly directory", metavar='DIR')
    inputs.add_option("-l", "--library", dest="library", help="library name", metavar='STR')
    inputs.add_option("-L", "--readlength", dest="readlength", help="read length", type="int", metavar='INT')
    inputs.add_option("-X", "--reference", dest="reference", help="name of reference genome", metavar='STR')    
    inputs.add_option("-n", "--input", dest="input", help="input file (library, project, etc.)", metavar='FILE')
    inputs.add_option("-N", "--numlibraries", dest="numlibraries", help="process INT libraries [%default] in input file starting from the library specified with --library, used in conjunction with --input and --library", type="int", metavar='INT', default=1)
    parser.add_option_group(inputs)

    actions = OptionGroup(parser, "Actions", "For both genome and transcriptome analsyis")
    actions.add_option("-d", "--set_up", dest="set_up", help="set up directory", action="store_true", default=False)
    actions.add_option("-0", "--fem", dest="fem", action="store_true", help="filter, extend, and merge ABySS assemblies", default=False)
    actions.add_option("-R", "--prepare_reads", dest="prepare_reads", help="prepare read file(s) for aligning reads to merged assembly", action="store_true", default=False)
    actions.add_option("-c", "--align_contigs", dest="align_contigs", help="align merged assembly to reference genome", action="store_true", default=False)
    actions.add_option("-r", "--align_reads", dest="align_reads", help="align reads to merged assembly", action="store_true", default=False)
    actions.add_option("-k", "--track", dest="track", help="generate custom UCSC track", action="store_true", default=False)
    actions.add_option("-f", "--fusion", dest="fusion", help="call fusion events", action="store_true", default=False)
    actions.add_option("-i", "--indel", dest="indel", help="call indel events", action="store_true", default=False)
    actions.add_option("-x", "--splicing", dest="splicing", help="call novel splicing events", action="store_true", default=False)    
    actions.add_option("-b", "--copy_genome_bam", dest="copy_genome_bam", help="copy reads-to-genome BAM directory (NOT SUPPORTED)", action="store_true", default=False)
    parser.add_option_group(actions)
                           
    rerun = OptionGroup(parser, "Re-run", "Options only for re-running analysis")
    rerun.add_option("-S", "--source", dest="source", help="source version for re-runing alignment and/or analysis [e.g. v1.4], used in conjuction with -d", metavar='STR')
    rerun.add_option("-u", "--rerun", dest="rerun", help="re-run contigs-to-genome alignments but do not split the merged assembly again, used in conjunction with -c", action="store_true", default=False)
    parser.add_option_group(rerun)
    
    cluster = OptionGroup(parser, "Host", "Sets which machines the jobs should run on")
    cluster.add_option("-C", "--cluster", dest="cluster", help="submit to cluster head node", metavar='STR')
    cluster.add_option("-q", "--queue", dest="queue", help="submit to cluster queue", metavar='STR')
    cluster.add_option("-m", "--local", dest="local", help="run locally", action="store_true", default=False)
    parser.add_option_group(cluster)
    
    parser.add_option("-D", "--debug", dest="debug", help="run in debug mode; no jobs will be submitted", action="store_true", default=False)
            
    (options, args) = parser.parse_args()
    
    #TODO: Validate the options/settings used. Some are mutually exclusive. Some must be used in conjunction with another. Should use argsparse instead of optionparser.
    
    if options.debug:
        message = """Running in debug mode!
*******************************************************************
* No jobs will be run locally or submitted to the cluster.        *
* Only job scripts, directories, and/or symlinks will be created. *
*******************************************************************"""
        print_message("WARNING", message, False)
     
    if options.input:
        if not os.path.isfile(options.input):
            message = "Cannot find file: %s" % options.input
            print_message("ERROR", message, True)        
        pipes = make_pipes_from_input_file(options)
        
        for pipe in pipes:
            pipe.process(options)
    else:
        #if not options.project:
        #    print_message("ERROR", "--project is not specified.", True)
        if not options.library:
            print_message("ERROR", "--library is not specified.", True)
        if options.fem and options.run_transcriptome and not options.readlength:
            print_message("ERROR", "--readlength is not specified.", True)
        if options.set_up and not options.assembly_dir:
            print_message("ERROR", "--assembly_dir is not specified.", True)
        if (not options.run_genome and not options.run_transcriptome) or (options.run_genome and options.run_transcriptome):
            print_message("ERROR", "Must use either --run_genome or --run_transcriptome", True)

        if options.run_genome:
            assembly_method = 'genome'
            if options.targeted:
                assembly_method = 'targeted_genome'
            pipe = PipeGenome(assembly_method, options.project, options.library, options.readlength, assembly_dir=options.assembly_dir, cluster=options.cluster, queue=options.queue, local=options.local, debug=options.debug, genome=options.reference, top_dir=options.top_dir)
        else:
            assembly_method = 'transcriptome'
            pipe = PipeTranscriptome(assembly_method, options.project, options.library, options.readlength, assembly_dir=options.assembly_dir, cluster=options.cluster, queue=options.queue, local=options.local, debug=options.debug, genome=options.reference, top_dir=options.top_dir)
        pipe.process(options)
    
